#!/usr/bin/python

# Copyright 2014 Keith Clark, Peter Robinson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PyQt4 import QtGui, QtCore
import sys
import pedroclient
import time
import copy

HEIGHT = 420
SLOTS_PER_TABLE = 12
NUM_SLOTS = 3*SLOTS_PER_TABLE+4

BLOCK_SIZE = 20

BLOCK_SIZE2 = BLOCK_SIZE/2
ARM_SIZE = 4
TABLE_WIDTH = 4
ARM_SIZE2 = ARM_SIZE/2
BLOCK_SIZE4 = BLOCK_SIZE/4
TABLE_POS = HEIGHT - 70 

BS32 = BLOCK_SIZE2 + BLOCK_SIZE
BS34 = (BLOCK_SIZE*2)/3

WIDTH = NUM_SLOTS*BS32

TABLE_SIZE = SLOTS_PER_TABLE*BS32

TOP = 100

def slot2pos(i):
    return BS32*(i)

def pos2slot(p):
    return int(0.5 + p/float(BS32))

def pos2slotpos(p):
    if p > TABLE_POS:
        return 0
    return int((TABLE_POS - p)/float(BLOCK_SIZE))

def slotpos2pos(sp):
    return TABLE_POS - (sp+1)*BLOCK_SIZE

TABLE_NAMES = ['table1', 'shared', 'table2']
TABLE_MAP = {'table1':0, 'shared':1, 'table2':2}

TABLE_START = [slot2pos(1) - BLOCK_SIZE4, 
               slot2pos(SLOTS_PER_TABLE+2) - BLOCK_SIZE4, 
               slot2pos(2*SLOTS_PER_TABLE+3) - BLOCK_SIZE4]

SLOT_HOLES = [0, SLOTS_PER_TABLE+1, 2*SLOTS_PER_TABLE+2, 3*SLOTS_PER_TABLE+3]

BLOCKS = [(1, 1, QtGui.QColor(200, 30, 30)),
          (3, 2, QtGui.QColor(200, 150, 30)),
          (4, 3, QtGui.QColor(200, 180, 40)),
          (6, 4, QtGui.QColor(200, 200, 40)),
          (7, 14, QtGui.QColor(200, 220, 60)),
          (SLOT_HOLES[1]+3, 5, QtGui.QColor(160, 200, 40)),
          (SLOT_HOLES[1]+5, 6, QtGui.QColor(50, 200, 40)),
          (SLOT_HOLES[1]+7, 7, QtGui.QColor(40, 200, 170)),
          (SLOT_HOLES[1]+8, 8, QtGui.QColor(40, 100, 140)),
          (SLOT_HOLES[1]+9, 9, QtGui.QColor(80, 80, 180)),
          (SLOT_HOLES[1]+10, 15, QtGui.QColor(80, 80, 240)),
          (NUM_SLOTS-7, 16, QtGui.QColor(100, 100, 255)),
          (NUM_SLOTS-8, 10, QtGui.QColor(100, 80, 240)),
          (NUM_SLOTS-5, 11, QtGui.QColor(140, 40, 200)),
          (NUM_SLOTS-3, 12, QtGui.QColor(200, 40, 200)),
          (NUM_SLOTS-2, 13, QtGui.QColor(250, 40, 200))]


TABLE_SLOTS = [range(1,SLOTS_PER_TABLE+1), 
               range(SLOTS_PER_TABLE+2, 2*SLOTS_PER_TABLE+2),
               range(2*SLOTS_PER_TABLE+3, 3*SLOTS_PER_TABLE+3)]

TABLE2ID = {}
for i,tn in enumerate(TABLE_NAMES):
    TABLE2ID[tn] = i

SLOT2TABLE = {}
for i,ts in enumerate(TABLE_SLOTS):
    for s in ts:
        SLOT2TABLE[s] = TABLE_NAMES[i]

SPEED = 16 #24

TASK_COLOURS = [QtGui.QColor(100, 100, 100),
                QtGui.QColor(255, 0, 0),
                QtGui.QColor(255, 160, 8),
                QtGui.QColor(0, 0, 255),
                QtGui.QColor(40, 200, 200),
                QtGui.QColor(200, 40, 200),
                QtGui.QColor(0, 0, 255),
                QtGui.QColor(0, 150, 150),
                QtGui.QColor(0, 150, 255),
                QtGui.QColor(100, 100, 100)]


def make_fun_0(fun):
    return lambda : fun()

def make_fun_1(fun, arg):
    return lambda : fun(arg)
 
class Pedro(QtCore.QThread):
    def __init__(self, app, client):
        QtCore.QThread.__init__(self, parent=app)
        self.app = app
        self.client = client
        self.signal = QtCore.SIGNAL("signal")
        self.running = True
        self.start()

    def run(self):
        while (self.running):
            try:
                p2pmsg = self.client.get_term()[0]
                self.emit(self.signal, p2pmsg)
            except Exception, e:
                print str(e)
                break

    def stop(self):
        self.running = False



class Block(object):
    def __init__(self, x, y, i, c):
        self.x = x
        self.y = y
        self.i = i
        self.istr = str(i)
        self.c = c

    def move(self, dx, dy):
        self.x += dx
        self.y += dy

    def move_abs(self, x, y):
        self.x = x
        self.y = y
        

class Arm(object):
    def __init__(self, i, x, y, app):
        self.x = x
        self.y = y
        self.i = i
        self.app = app
        self.closed = False
        self.holding = None
        self.xspeed = 0
        self.yspeed = 0
        self.commands = []
        self.xdest = None
        self.ydest = None
        self.active = False
        self.lifting = False
        self.colour = 0

    def __str__(self):
        return "arm {0}".format(self.i)

    def set_commands(self, commands):
        self.commands = commands
        self.active = True
        self.commands.pop(0)()

    def stop(self):
        self.lift()
        self.active = False
        self.colour = 0

    def reached_dest(self, current, dest, direction):
        if direction > 0:
            return current > dest - 5
        else:
            return current < dest + 5

    def step(self):
        if self.is_animating():
            if self.app.should_avoid(self):
                return
            self.move(self.xspeed, self.yspeed)
            dx, dy = 0, 0
            yd = self.ydest
            if self.reached_dest(self.x, self.xdest, self.xspeed):
                dx = self.xdest - self.x 
                self.xspeed = 0
            if self.ydest == TOP:
                if self.reached_dest(self.y, self.ydest, self.yspeed):
                    dy = self.ydest - self.y 
                    self.yspeed = 0
            else:
                yd = self.app.get_height(self.ydest, self.holding)
                if self.reached_dest(self.y, yd, self.yspeed):
                    dy = yd - self.y 
                    self.yspeed = 0
            self.move(dx, dy)
            if self.x == self.xdest and self.y == yd:
                print "------", len(self.commands)
                if self.commands == []:
                    self.active = False
                    self.lifting = False
                else:
                    self.commands.pop(0)()

    def is_animating(self):
        return self.active or self.lifting

    def lift(self):
        print 'lift', self
        self.active = True
        self.lifting = True
        self.yspeed = -SPEED
        self.xspeed = 0
        self.xdest = slot2pos(pos2slot(self.x))
        self.ydest = TOP

    def drop(self, y):
        print 'drop', y, self
        self.active = True
        self.yspeed = SPEED
        self.xspeed = 0
        self.xdest = slot2pos(pos2slot(self.x))
        self.ydest = y


    def move(self, dx, dy):
        self.x += dx
        self.y += dy
        if self.holding is not None:
            self.holding.move(dx, dy)

    def move_to(self, s):
        print 'move_to', s, self
        self.xdest = slot2pos(s)
        self.ydest = TOP
        self.yspeed = 0
        if self.x < self.xdest:
            self.xspeed = SPEED
        else:
            self.xspeed = -SPEED

    def at_slot(self, s):
        pos = slot2pos(s)
        return abs(self.x - pos) < 4

    def at_top(self):
        return abs(self.y - TOP < 4)

    def is_at(self, b):
        return abs(self.x - b.x) < 4 and abs(self.y - b.y) < 4

    def is_over_home_table(self):
        slot = pos2slot(self.x)
        return slot in TABLE_SLOTS[0] or slot in TABLE_SLOTS[-1]

    def hold(self, slot):
        #assert(self.is_at(b))
        print 'hold', self
        self.active = True
        self.flip_gripper()
        self.holding = self.app.slots[slot].pop()
        if self.app.slots[slot] == []:
            on = SLOT2TABLE[slot]
        else:
            on = str(self.app.slots[slot][-1].i)
        self.app.send_msg('[r_(holding(arm{0}, {1})), f_(on({1}, {2}))]'.format(self.i, self.holding.i, on))
        

    def is_holding(self, bid):
        return self.holding is not None and self.holding.i == bid

    def release(self, slot):
        print 'release', self
        self.active = True
        if self.app.slots[slot] == []:
            on = SLOT2TABLE[slot]
        else:
            on = str(self.app.slots[slot][-1].i)
        self.app.slots[slot].append(self.holding)
        self.app.send_msg('[f_(holding(arm{0}, {1})), r_(on({1}, {2}))]'.format(self.i, self.holding.i, on))
        self.holding = None
        self.flip_gripper()

    def flip_gripper(self):
        self.closed = not self.closed

    def rectangles(self):
        yield (self.x + BLOCK_SIZE2 - ARM_SIZE2, 0, ARM_SIZE, self.y)
        yield (self.x - ARM_SIZE, self.y - ARM_SIZE, 
               BLOCK_SIZE + 2*ARM_SIZE, ARM_SIZE)

        if self.closed:
            offset = ARM_SIZE
        else:
            offset = 2*ARM_SIZE
        yield (self.x - offset, self.y - ARM_SIZE, ARM_SIZE, BLOCK_SIZE)
        yield (self.x + BLOCK_SIZE + offset - ARM_SIZE, 
               self.y - ARM_SIZE, ARM_SIZE, BLOCK_SIZE)

class BlocksApp(QtGui.QWidget):

    def __init__(self):
        super(BlocksApp, self).__init__()
        self.resize(WIDTH, HEIGHT)
        self.black = QtGui.QColor(0, 0, 0)
        self.white = QtGui.QColor(0, 0, 0)
        self.grey = QtGui.QColor(100, 100, 100)
        self.setWindowTitle("TR - Multi Tower Block World")
        self.timer = QtCore.QTimer()
        QtCore.QObject.connect(self.timer,QtCore.SIGNAL("timeout()"), self.animate)
        self.x = 0
        self.y = 0
        self.slots = [[] for _ in range(NUM_SLOTS)]
        self.slots[SLOT_HOLES[0]] = None
        self.slots[SLOT_HOLES[1]] = None
        self.slots[SLOT_HOLES[2]] = None
        self.slots[SLOT_HOLES[3]] = None
        self.blocks = []
        for s, i, c in BLOCKS:
            b = Block(slot2pos(s), TABLE_POS-BLOCK_SIZE, i, c)
            self.blocks.append(b)
            self.slots[s].append(b)
        self.arms = [Arm(1, slot2pos(SLOT_HOLES[1]-1), TOP, self), 
                     Arm(2, slot2pos(SLOT_HOLES[2]+1), TOP, self)]
        self.client = pedroclient.PedroClient()
        self.client.register('env2arm')
        self.pedro = Pedro(self, self.client)
        self.connect(self.pedro, self.pedro.signal, self.process_message)
        self.animating = False
        self.interacting = False
        self.tasks = {}
        self.table_colours = [0,0,0]
        #self.qp = QtGui.QPainter()
        self.show()
        self.percepts_handle = None
        self.slots_copy= []

    def closeEvent(self, event):
        self.client.p2p('messages:towers', 'quiting')
        time.sleep(1)

    def should_avoid(self, arm):
        otherarm = self.arms[(arm.i % 2)]
        xdiff = arm.x - otherarm.x
        if abs(xdiff) < 10:
            return xdiff < 0 and arm.xspeed > 0 or xdiff > 0 and arm.xspeed < 0

    def keyPressEvent(self, event):
        if self.interacting:
            self.interacting = False
            self.repaint()
            self.send_state()
            self.start_animate()

    def mousePressEvent(self, event):
        if not self.interacting:
            self.slots_copy = copy.deepcopy(self.slots)
        self.interacting = True
        self.stop_animate()
        self.repaint()
        sid = pos2slot(event.x())
        #print 'press', sid
        s = self.slots[sid]
        self.moving_block = None
        if s is None or s == []:
            return
        pos =  pos2slotpos(event.y())
        if len(s) <= pos:
            return
        self.moving_block = s[pos]
        blocks = []
        shuffling = False
        for i, b in enumerate(self.slots[sid]):
            if i == pos:
                shuffling = True
            else:
                blocks.append(b)
                if shuffling:
                    b.move(0, BLOCK_SIZE)
        self.slots[sid] = blocks
        self.repaint()
        self.save_x = event.x()
        self.save_y = event.y()
        


    def mouseReleaseEvent(self, event):
        #print 'release'
        if self.moving_block is None:
            return
        sid = pos2slot(event.x())
        if sid == 0:
            sid = 1
        elif sid in SLOT_HOLES:
            sid -= 1
        if self.arms[0].at_slot(sid) and not self.arms[0].at_top() :
            if sid == 1:
                sid = 2
            elif sid == SLOT_HOLES[1]+1:
                sid += 1
            else:
                sid -= 1
        if self.arms[1].at_slot(sid) and not self.arms[1].at_top() :
            if sid == SLOT_HOLES[3]-1:
                sid -= 1
            elif sid == SLOT_HOLES[2]-1:
                sid -= 1
            else:
                sid += 1
            
        pos =  pos2slotpos(event.y())
        if pos > len(self.slots[sid]):
            pos = len(self.slots[sid]) 
        blocks = []
        shuffling = False
        for i, b in enumerate(self.slots[sid]):
            if i == pos:
                self.moving_block.move_abs(slot2pos(sid), slotpos2pos(pos))
                blocks.append(self.moving_block)
                shuffling = True
            if shuffling:
                b.move(0, -BLOCK_SIZE)
            blocks.append(b)
        if not shuffling:
            self.moving_block.move_abs(slot2pos(sid), slotpos2pos(pos))
            blocks.append(self.moving_block)
        self.moving_block = None
        self.slots[sid] = blocks
        self.repaint()

    def mouseMoveEvent(self, event):
        if self.moving_block is None:
            return
        x,y = event.x(), event.y()
        self.moving_block.move(x - self.save_x, y - self.save_y)
        self.repaint()
        self.save_x, self.save_y = x,y
 
    def send_state(self):
        messages = []
        if self.arms[0].holding:
            messages.append('r_(holding(arm{0}, {1}))'.format(self.arms[0].i, self.arms[0].holding.i))
        else:
            messages.append('f_(holding(arm{0},1))'.format(self.arms[0].i))
        if self.arms[1].holding:
            messages.append('r_(holding(arm{0}, {1}))'.format(self.arms[1].i, self.arms[1].holding.i))
        else:
            messages.append('f_(holding(arm{0},1))'.format(self.arms[1].i))
        for i,s in enumerate(self.slots_copy):
            if s is not None and s != []:
                for j,b in enumerate(s):
                    if j == 0:
                        messages.append('f_(on({0}, {1}))'.format(s[0].i, SLOT2TABLE[i]))
                    else:
                        messages.append('f_(on({0}, {1}))'.format(s[j].i, s[j-1].i))
            
        for i,s in enumerate(self.slots):
            if s is not None and s != []:
                for j,b in enumerate(s):
                    if j == 0:
                        messages.append('r_(on({0}, {1}))'.format(s[0].i, SLOT2TABLE[i]))
                    else:
                        messages.append('r_(on({0}, {1}))'.format(s[j].i, s[j-1].i))
        #print messages
        self.send_msg('['+','.join(messages)+']')
        
    def set_resource_colours(self, lst):
        for a in self.arms:
            a.colour = 0
        self.table_colours = [0,0,0]
        if lst.get_type() == pedroclient.PObject.listtype:
            task_res = lst.toList()
            for tr in task_res:
                task = tr.args[0].val
                if task in  self.tasks:
                    res = tr.args[1].toList()
                    for r in res:
                        t = TABLE_MAP.get(r.val)
                        if t is not None:
                            self.table_colours[t] = self.tasks[task][2]
                        else:
                            arm = self.arms[int((r.val)[-1])-1]
                            arm.colour = self.tasks[task][2]

    def set_waiting_info(self, lst):
        for t in self.tasks:
            self.tasks[t][3] = ''
        if lst.get_type() == pedroclient.PObject.listtype:
            task_res = lst.toList()
            for tr in task_res:
                task = tr.args[0].val
                res = str(tr.args[1])
                self.tasks[task][3] = res
 
    def process_message(self, pmsg):
        msg = pmsg
        sender = msg.args[1].args[0].args[0].val
        msg = msg.args[2]
        #print msg
        if  msg.get_type() == pedroclient.PObject.atomtype and \
                msg.val == 'initialise_':
            self.percepts_handle = pmsg.args[1]
            self.send_state()
            return
        if  msg.get_type() == pedroclient.PObject.atomtype and \
                msg.val == 'finish':
            return
        if msg.get_type() == pedroclient.PObject.structtype and \
           msg.functor.val ==  'locked_resources':
            self.set_resource_colours(msg.args[0])
            return
        elif msg.get_type() == pedroclient.PObject.structtype and \
             msg.functor.val ==  'waiting_resources':
            self.set_waiting_info(msg.args[0])
            return
        elif  msg.get_type() == pedroclient.PObject.structtype and \
                msg.functor.val ==  'task_data':
            used_labels = [self.tasks[k][2] for k in self.tasks]
            for i in range(1,10):
                if i not in used_labels:
                    break
            self.tasks[msg.args[0].val] = ['{0} : {1}'.format(msg.args[0].val,
                                                              str(msg.args[1])),
                                           msg.args[2].val, 
                                           i, '']
            self.repaint()
            return
        elif  msg.get_type() == pedroclient.PObject.structtype and \
                msg.functor.val ==  'remove_task':
            self.tasks.pop(msg.args[0].val)
            self.repaint()
            return
        elif  msg.get_type() == pedroclient.PObject.structtype and \
                msg.functor.val ==  'controls':
            actions = msg.args[0]
        
            if actions.get_type() != pedroclient.PObject.listtype:
                print "COMMAND: ", str(pmsg)
                assert False
            for a in actions.toList():
                self.process_action(a, sender)
            return
        else:
            print "COMMAND: ", str(msg)
            assert False
  

    def process_action(self, msg, sender):  
        assert msg.get_type() == pedroclient.PObject.structtype
        functor = msg.functor
        assert functor.get_type() == pedroclient.PObject.atomtype
        cmd_type = functor.val
        cmd = msg.args[0]
        if cmd_type == 'stop_':
            assert cmd.get_type() == pedroclient.PObject.structtype
            cmd_functor = cmd.functor.val
            #if cmd_functor in ['pickup', 'putdown']:
            self.stop_arm(cmd.args[0].val)
            #else:
            #    self.stop_arm(cmd.args[0].val)
    
        elif cmd_type in ['start_', 'mod_']:
            self.start_animate()
            assert cmd.get_type() == pedroclient.PObject.structtype
            cmd_functor = cmd.functor.val
            if cmd_functor == 'pickup':
                self.pickup(cmd.args[0].val, cmd.args[1].val, sender)
            elif cmd_functor == 'putdown':
                if cmd.args[1].get_type() == pedroclient.PObject.inttype:
                    self.putdown_on_block(cmd.args[0].val, cmd.args[1].val, 
                                          sender)
                else:
                    self.putdown_on_table(cmd.args[0].val, cmd.args[1].val, 
                                          sender)
            elif cmd_functor == 'go_home':
                 self.go_home(cmd.args[0].val)
  
                
    def get_height(self, sid, holding):
        slot = self.slots[sid]
        if len(slot) == 0:
            return TABLE_POS - BLOCK_SIZE
        elif holding:
            return slot[-1].y - BLOCK_SIZE
        else:
            return slot[-1].y

    def stop_arm(self, arm):
        arm = self.arms[int(arm[-1])-1]
        arm.stop()

    def get_slot(self, b):
        for i, s in enumerate(self.slots):
            if s is not None:
                for b1 in s:
                    if b1.i == b:
                        return i
        return None

    def free_slot(self, table, arm):
        if table == 0:
            arm_slot = pos2slot(arm.x)
            d = 1
            while True:
                pos = arm_slot + d
                if pos in TABLE_SLOTS[table] and self.slots[pos] == []:
                    return pos
                if d > 0:
                    d = -d
                else:
                    d = 1-d
        elif table == 2:
            arm_slot = pos2slot(arm.x)
            d = -1
            while True:
                pos = arm_slot + d
                if pos in TABLE_SLOTS[table] and self.slots[pos] == []:
                    return pos
                if d < 0:
                    d = -d
                else:
                    d = -1-d
        elif arm.i == 1:
            for i in TABLE_SLOTS[table]:
                s = self.slots[i]
                if s == []:
                    return i
            return None
        else:
            slots = list(TABLE_SLOTS[table])
            slots.reverse()
            for i in slots:
                s = self.slots[i]
                if s == []:
                    return i
            return None

    def go_home(self, arm):
        armid = int(arm[-1])
        arm = self.arms[armid-1]
        if arm.is_over_home_table():
            return
        if armid == 1:
            sid = SLOT_HOLES[1] - 1
            #print 'go_home',armid,sid
        else:
            sid = SLOT_HOLES[2] + 1
            #print 'go_home',armid,sid
        if not arm.at_top():
            arm.set_commands([make_fun_0(arm.lift),
                              make_fun_1(arm.move_to, sid)])
        else:
            arm.set_commands([make_fun_1(arm.move_to, sid)])

    def pickup(self, arm, b, task):
        arm = self.arms[int(arm[-1])-1]
        b_slot = self.get_slot(b)
        block = self.slots[b_slot][-1]
        b_y = block.y
        assert not arm.is_holding(b)
        if arm.at_slot(b_slot):
            arm.set_commands([make_fun_1(arm.drop, b_slot), 
                              make_fun_1(arm.hold, b_slot),
                              make_fun_0(arm.lift)])
        elif not arm.at_top():
            arm.set_commands([make_fun_0(arm.lift),
                              make_fun_1(arm.move_to, b_slot), 
                              make_fun_1(arm.drop, b_slot), 
                              make_fun_1(arm.hold, b_slot),
                              make_fun_0(arm.lift)])
        else:
            arm.set_commands([make_fun_1(arm.move_to, b_slot), 
                              make_fun_1(arm.drop, b_slot), 
                              make_fun_1(arm.hold, b_slot),
                              make_fun_0(arm.lift)])

    def putdown_on_table(self, arm, table, task):
        table = TABLE2ID[table]
        arm = self.arms[int(arm[-1])-1]
        assert arm.holding
        table_slot = self.free_slot(table, arm)
        self.move_block(arm, table_slot)

    def putdown_on_block(self, arm, block, task):
        arm = self.arms[int(arm[-1])-1]
        assert arm.holding
        b_slot = self.get_slot(block)
        b_y = self.slots[b_slot][-1].y - BLOCK_SIZE
        self.move_block(arm, b_slot)
            

    def move_block(self, arm, b2_slot):
        if arm.at_slot(b2_slot):
            arm.set_commands([make_fun_1(arm.drop, b2_slot), 
                              make_fun_1(arm.release, b2_slot),
                              make_fun_0(arm.lift)])
        elif not arm.at_top():
            arm.set_commands([make_fun_0(arm.lift),
                              make_fun_1(arm.move_to, b2_slot), 
                              make_fun_1(arm.drop, b2_slot), 
                              make_fun_1(arm.release, b2_slot),
                              make_fun_0(arm.lift)])
        else:
            arm.set_commands([make_fun_1(arm.move_to, b2_slot), 
                              make_fun_1(arm.drop, b2_slot), 
                              make_fun_1(arm.release, b2_slot),
                              make_fun_0(arm.lift)])
            
    def send_msg(self, msg):
        #print "sending:", self.percepts_handle, msg
        self.client.p2p(self.percepts_handle, msg)

    def paintEvent(self, event):
        self.qp = QtGui.QPainter()
        self.qp.begin(self)
        self.qp.setRenderHint(QtGui.QPainter.Antialiasing)
        self.draw(event)
        self.qp.end()

    def draw_block(self, b):
        self.qp.save()
        metrics = self.qp.fontMetrics()
        height = metrics.height()
        w = metrics.width(b.istr)
        self.qp.setBrush(b.c)
        self.qp.setPen(b.c)
        self.qp.translate(b.x, b.y)
        self.qp.drawRect(0, 0, BLOCK_SIZE, BLOCK_SIZE)
        self.qp.setPen(self.black)
        self.qp.drawText((BLOCK_SIZE-w)/2, (BLOCK_SIZE + height/2)/2, b.istr)
        self.qp.restore()

    def draw_blocks(self):
        for b in self.blocks:
            self.draw_block(b)

    def draw_tables(self):
        colour = TASK_COLOURS[self.table_colours[0]]
        self.qp.setBrush(colour)
        self.qp.setPen(colour)
        self.qp.drawRect(TABLE_START[0], TABLE_POS, TABLE_SIZE, TABLE_WIDTH)
        self.qp.setPen(self.black)
        self.qp.drawText(TABLE_START[0]+TABLE_SIZE/2-20,367, "Table 1")
        colour = TASK_COLOURS[self.table_colours[1]]
        self.qp.setBrush(colour)
        self.qp.setPen(colour)
        self.qp.drawRect(TABLE_START[1], TABLE_POS, TABLE_SIZE, TABLE_WIDTH)
        self.qp.setPen(self.black)
        self.qp.drawText(TABLE_START[1]+TABLE_SIZE/2-20,367, "Shared")
        colour = TASK_COLOURS[self.table_colours[2]]
        self.qp.setBrush(colour)
        self.qp.setPen(colour)
        self.qp.drawRect(TABLE_START[2], TABLE_POS, TABLE_SIZE, TABLE_WIDTH)
        self.qp.setPen(self.black)
        self.qp.drawText(TABLE_START[2]+TABLE_SIZE/2-20,367, "Table 2")

    def draw_arm(self, arm):
        colour = TASK_COLOURS[arm.colour]
        self.qp.setBrush(colour)
        self.qp.setPen(colour)
        for rect in arm.rectangles():
            self.qp.drawRect(*rect)
        holding = arm.holding
        if holding is not None:
            self.draw_block(holding)

    def draw_arms(self):
        self.draw_arm(self.arms[0])
        self.draw_arm(self.arms[1])

    def draw_tasks(self):
        tab1 = (task for task in self.tasks.itervalues() if task[1] == 'table1')
        tab2 = (task for task in self.tasks.itervalues() if task[1] == 'table2')
        metrics = self.qp.fontMetrics()
        pos = TABLE_START[0]
        for t in tab1:
            if t[3]:
                self.qp.setPen(self.black)
                self.qp.drawText(pos, 390, t[3])
            self.qp.setPen(TASK_COLOURS[t[2]])
            self.qp.drawText(pos, 410, t[0])
            pos += 40+metrics.width(t[0])
                
        pos = TABLE_START[2]
        for t in tab2:
            if t[3]:
                self.qp.setPen(self.black)
                self.qp.drawText(pos, 390, t[3])
            self.qp.setPen(TASK_COLOURS[t[2]])
            self.qp.drawText(pos, 410, t[0])
            pos += 40+metrics.width(t[0])


    def draw(self, event):
        #self.qp.save()
        #self.qp.setBrush(self.black)
        #self.qp.drawRect(self.x, self.y, 90, 60)
        if self.interacting:
            self.qp.drawText(TABLE_START[1],390, "Press any key to continue")
        self.draw_blocks()
        self.draw_tables()
        self.draw_arms()
        self.draw_tasks()
        #self.qp.restore()

    def start_animate(self):
        if not self.animating:
            self.timer.start(10)
            self.animating = True

    def stop_animate(self):
        if self.animating:
            #print "stopping"
            self.timer.stop()
            self.animating = False

    def animate(self):
        if not self.arms[0].is_animating() and not self.arms[1].is_animating():
            self.stop_animate()
            return
        self.arms[0].step()
        self.arms[1].step()
        self.repaint()
        

if __name__ == "__main__":
    app = QtGui.QApplication(sys.argv)
    r = BlocksApp()
    sys.exit(app.exec_())

