%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.

%% Some operator declarations to make code easier read
?- compile_time_only(consult(op_decls)).
%% for storing away errors for later reporting
?- dynamic(type_error_remember/1).


%% temporary storage while processing program
?- dynamic('$tr_prog'/2).       %% remember TR programs
?- dynamic('$runtime_goal'/1).  %% remember goals to exec at runtime
?- dynamic('$goal'/1).    %% remember goals to run at consult time
%% assertions for TR programs
?- dynamic('$task_atomic'/2).
?- dynamic('$task_start'/1).
?-dynamic('$resource_info'/3).
?-dynamic('$resource'/1).
?-dynamic('$durative'/1).
?-dynamic('$discrete'/1).
?-dynamic('$resource_type'/1).
?-dynamic('$consulted_timestamp'/3).
?-dynamic('$saved_consulted'/1).
?-dynamic('$consulted'/1).
?-dynamic('$file_stream'/2).

%?-dynamic('$procedure'/1).
%?-dynamic('$dynamicP'/1).  %% dynamic percept facts
%?-dynamic('$dynamicB'/1).  %% dynamic belief facts

%?- dynamic('$expand_map'/3).
%?- dynamic('$expand_fun_map'/2).

%?- dynamic('$decl_kind'/3).
?- dynamic('$tmp_clause'/1). % temp storage of definitions
%?- dynamic('$trans_clause'/2).
%?- dynamic('$tmp_apply'/2).
?- dynamic('$error'/1).
%%for listing, clause etc so that the same var names are used
%% the last arg is the name of the file the defs/decls came from
?- dynamic('$saved_input_with_vars'/3).
%% user type info:
%% '$user_type_info'(Functor, Kind, Types, Term, VarNames)
?- dynamic('$user_type_info'/5).
?- dynamic('$belief'/1).   %% beliefs are special relations
?- dynamic('$percept'/1).  %% ditto percepts
?- dynamic('$state_belief'/1).  %% ditto state beliefs

%% When declared percepts have a relation definition then
%% should_be_remembered_(HeadPattern) is asserted and
%% can_infer_now_(Head) :- Body
%% is also asserted
?- dynamic(should_be_remembered_/1).
?- dynamic(can_infer_now_/1).

%% Both below used so that reconsulting a file will remove old defs
%% the file currently being consulted - used to remember what
%% was loaded by this consult
?- dynamic('$current_consulted_file'/1).
%% the info about what was consulted from this file
?- dynamic('$asserted_file_info'/2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Temp - for debugging
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dump_state :-
    errornl('%%%%%%%%%%%% dumping state %%%%%%%%%%%%%%%%'),
    fail.
dump_state :-
    '$durative'(X),
    errornl('$durative'(X)),
    fail.
dump_state :-
    '$discrete'(X),
    errornl('$discrete'(X)),
    fail.
dump_state :-
    '$resource'(Xs),
    errornl('$resource'(Xs)),
    fail.
dump_state :-
    '$task_start'(TR),
    errornl(task_start(TR)),
    fail.
dump_state :-
    '$task_atomic'(Term, Named),
    errornl('$task_atomic'(Term, Named)),
    fail.
dump_state :-
    '$resource_info'(Term, A2, P),
    errornl('$resource_info'(Term, A2, P)),
    fail.
/*
  dump_state :-
    '$defined_type'(A,B),
    errornl('$defined_type'(A,B)),
    fail.
*/
dump_state :-
    '$resource_type'(A),
    errornl('$resource_type'(A)),
    fail.
dump_state :-
    '$user_type_info'(A,B,C,D,E),
    errornl('$user_type_info'(A,B,C,D,E)),
    fail.
/*
  dump_state :-
    '$procedure'(P),
    errornl('$procedure'(P)),
    fail.
*/
dump_state :-
    errornl('%%%%%%%%%%%%%%%%%%%%%%%%%%%%').




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Top-level of consult
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

?- dynamic('$syntax_errors'/0).

%% Consult a .qlg file. Remove and code from an earlier consult
%% of this file (a reconsult) 

top_level_qconsult(File) :-
    assert('$top_level_consulted_file'(File)),
    writeL(["Consulting ", File, "...",nl_]),
    catch('$qconsult_file'(File), consult_error(E),
          '$failure_cleanup_consult'(E)).

qconsult(File) :-
    writeL(["Consulting ", File, "...",nl_]),
    '$qconsult_file'(File).

'$qconsult_file'(File) :-
    catch(open_qulog_file(File, read, Stream), exception(E),
          '$handle_consult_exception'(E)),
    assert('$file_stream'(File, Stream)),
    '$delete_all_code'(File),
    %% asserta is used so the first fact found will be for the
    %% "inner consult"
    asserta('$current_consulted_file'(File)),
    retractall('$syntax_errors'),
    %% set the error count to 0 - this is updated each time an error is found
    global_state_set(num_errors, 0),
    %% attempt to read in and process terms of the file
    qconsult_terms(Stream, File),
    (
      retract('$syntax_errors')
    ->
      writeL(stderr, [nl_, "Syntax errors found - file not consulted",
                      nl_, nl_]),
      throw(consult_error(syntax))
    ;
      retract('$file_stream'(File, Stream1)),
      close(Stream1)
    ),
    %% type declarations can have missing arrows and so are initially set
    %% to have an unknown kind. This looks at the defns and updates the
    %% type info
    '$update_unknown_kinds',
    %%'$check_single_enum',
    %% In a type decl like t1 ::= t2 the parser converts the RHS to be
    %% a macro type with a single element. If, however, t2 is not
    %% a type then it is converted to an enumeration with a single element
    '$update_macros',
    %% check that in t1(X) ::= t2, the X occurs on RHS
    '$check_defined_type_vars',
    %% If a type declaration is left off for a definition then
    %% the system will give it a "general type". An error is produced
    %% if this is a redefinition of a system defn.
    '$check_if_clauses_have_types',
    %% if modes are left off relation or action type declarations
    %% then add default modes
    '$add_default_modes',
    %% check the types on RHS of type definition or types used
    %% in declarations are valid
    '$check_valid_types',
    %% check that the definitions of relations/functions/actions/TR programs
    %% are valid defns with correct types and modes.
    '$check_definitions',
    %% Are there any errors?
    global_state_lookup(num_errors, N1),
    (
      N1 = 1
    ->
      writeL(stderr, [nl_, nl_,"1 error detected - ", File, " not consulted",
                      nl_,nl_]),
      throw(consult_error(type))
    ;
     N1 > 1
    ->
      writeL(stderr, [nl_, nl_, N1,
                      " errors detected - ", File, " not consulted",
                      nl_,nl_]),
      throw(consult_error(type))
    ;
      true
    ),
    %% trans_fun looks to see if a variable functor comes from a TR
    %% program and if so translates the function differently
    %% this needs to be set to empty before non-TR defns are translated
    ip_set('$tr_prog_vars', []),
    %% translate all function/relation/action clauses to prolog code
    forall(retract('$tmp_clause'(clause(Clause))),
           (freeze_term(Clause), '$translate_clause'(Clause))),
    %% the remainder is about TR programs
    %% assert the resource info for each TR program
    '$add_tr_resource_info',
    %% check that if resources are required then they are correctly defined
    %% and used
    '$check_resource_uses',
    %% if a TR program is defined then we need to know how TR guards
    %% depend on percepts and beliefs. This generates the dependency info
    %% for all relations in the program
    '$tr_build_clause_info',
    %% translate the TR programs to prolog clauses
    forall(retract('$tr_prog'(Head, Body)),
           (freeze_term(Clause), '$expand_a_TR'(Head, Body))),
    %% pop consulted file info
    retract('$current_consulted_file'(File)),
    writeL(["... ", File, " consulted ", nl_]).

%% If the consult fails because of a syntax or other error then
%% all the code related to the file needs to be cleaned up
%% Given that a file can be consulted within another file
%% the '$current_consulted_file'/1 implements a stack of
%% files currently being consulted.
%% If a consult fails inside a top-level consult then we 
%% throw an exception to be picked up at the interpreter
%% level consult which should respond by removing all
%% consulted info and fail.
%% If the consult at the top-level fails because of, for example,
%% a type error in the top-level file then only the data
%% for that file should be cleaned up.
'$failure_cleanup_consult'(consults_diff) :-
    !,
    errornl(diffs____),
    '$delete_code_from_all',
    '$top_level_consulted_file'(File),
    top_level_qconsult(File).
'$failure_cleanup_consult'(_E) :-
    once('$current_consulted_file'(File)),
    '$top_level_consulted_file'(File), !,
    %% The failure is in the top-level file
     (
       retract('$file_stream'(File, Stream1))
     ->
       close(Stream1)
     ;
       true
     ),
    '$delete_all_code'(File),
    retractall(type_error_remember(_)),
    retractall('$tmp_clause'(_)),
    retractall('$tr_clause'(_, _)),
    retractall('$seen_clause'(_, _, _)),
    retractall('$saved_input_with_vars'(_, _, _)),
    retract('$current_consulted_file'(File)),
    retractall('$asserted_file_info'(File, _)),
    retractall('$consulted_timestamp'(File, File, _)),
    retractall('$saved_consulted'(_)),
    retractall('$goal'(_)),
    writeL(["Error consulting ", File, nl_]),
    retractall('$top_level_consulted_file'(_)),
    !,
    fail.
'$failure_cleanup_consult'(_E) :-
    once('$current_consulted_file'(ErrorFile)),
    '$top_level_consulted_file'(File),
    '$delete_code_from_all',
    writeL(["Error in file ",  ErrorFile, " while consulting ", File, nl_,
            "All user definitions and declarations removed.", nl_, nl_]), !,
    fail.


'$delete_code_from_all' :-
    forall(retract('$file_stream'(_, Stream)), close(Stream)),
    findall(F, '$asserted_file_info'(F, _), Fs),
    sort(Fs, SFs),
    retractall('$current_consulted_file'(_)),
    %% remove all consulted code
    forall(member(File, SFs), '$delete_all_code'(File)),
    %% remove all tmp stuff
    retractall('$tmp_clause'(_)),
    retractall('$tr_clause'(_, _)),
    retractall('$seen_clause'(_, _, _)),
    retractall('$saved_input_with_vars'(_, _, _)),
    retractall('$asserted_file_info'(_, _)),
    retractall('$consulted_timestamp'(_, _, _)),
    retractall('$consulted'(_)),
    retractall('$saved_consulted'(_)),
    retractall('$goal'(_)),
    retract('$top_level_consulted_file'(_)).


'$handle_consult_exception'(permission_error(_, qconsult(File), _)) :-
   writeL(["Cannot read ", File, nl_]),
   writeL(["Error consulting ", File, nl_]),
   throw(consult_error(read_error)).

%% read the terms of the files
qconsult_terms(Stream, File) :-
    repeat,
    %% read the string representing the next term to be processed
    read_term_string(Stream, String, StartLine, EndLine),
    (
      string_length(String, 0)
    ->
      %% EOF
      !
    ;
      retractall('$parser_info'(_, _, _, _)),
      stream_property(Stream, file_name(FileName)),
      %% store information about string for later
      %% processing of syntax errors
      assert('$parser_info'(String, StartLine, EndLine, FileName)),
      %% Get the tokens and process the variables in the term
      string2tokens(String, Tokens, VarList),
      '$collect_var_tables'(VarList, _Variables, VarNames, Singles, []),
      '$unify_on_names'(VarList),
      Tokens \= [],
      %% At this point Tokens is the list of tokens for the current term
      %% being processed
      %% The following parses the tokens
      catch(
            (
              phrase(file_term(T), Tokens, [])
            ->
              '$warn_of_single_vars'(T, Singles, VarNames),
              %% save away for displaying or error messages
              %% using the original term and var names
              assert('$saved_input_with_vars'(T, VarNames, File)),
              %% process the term
              '$process_input_top'(T, VarNames, _)
            ;
              %% Unable to parse
              nl,
              report_error('Can\'t parse', Tokens),
              nl
            ),
            %% syntax errors are caught here
            syntax_error,
            assert('$syntax_errors')
           ),
      fail
    ).

%% Whenever something from the file is added to the system
%% we keep track of it so that if there is an error or
%% the file is reconsulted we can remove all the relevant information
'$add_file_info'(Info) :-
    once('$current_consulted_file'(File)),
    (
     '$asserted_file_info'(File, Info)
    ->
     true
    ;
     assert('$asserted_file_info'(File, Info))
    ).

%% Delete all declarations/definitions/code that have been added
%% by consult(File)
'$delete_all_code'(File) :-
    retract('$asserted_file_info'(File, Info)),
    retractall(Info),
    fail.
'$delete_all_code'(File) :-
    retractall('$tr_clause'(_, _)),
    retractall('$seen_clause'(_, _, _)),
    retractall('$saved_input_with_vars'(_, _, File)).


%% '$process_input'(Term) takes a parsed Term
%% and decides what to do with it.
%%'$process_input_top'(G, Vars, Flag) :- errornl('$process_input_top'(G, Vars, Flag)),fail.

'$process_input_top'(G, _Vars, Flag) :-
    G = (?- _),
    nonvar(Flag), !,
    writeL(stderr, ["Error: ", G, " occurs after a non-command", nl_]),
    throw(consult_error(consult)).
'$process_input_top'((?- G), _, _) :-
    list(G), !,
    forall(member(C, G), '$process_input_top'('?-'(consult(C)), _, _)).
'$process_input_top'((?- runtime(Pred)), _, _) :-
    !,
    '$add_file_info'('$runtime_goal'(_)),
    assert('$runtime_goal'(Pred)).
'$process_input_top'('?-'(consult(G)), _, _) :- !,
    '$process_input_top'('?-'(qconsult(G)), _, _).
'$process_input_top'('?-'(qconsult(G)), _, _) :- !,
    concat_atom([G, '.qlg'], Gqlg),
    stat(Gqlg, stat(Time, _)),
    once('$current_consulted_file'(File)),
    (
      '$top_level_consulted_file'(File)
    ->
      assert('$consulted'(G))
    ;
      true
    ),
    (
      '$consulted_timestamp'(G, _, Time)
    ->
      true
    ;
      assert('$consulted_timestamp'(G, File, Time)),
      '$add_file_info'('$goal'(_)),
      assert('$goal'(qconsult(G))),
      (qconsult(G)-> true ; throw(consult_error(consult)))
    ).
'$process_input_top'('?-'(pconsult(G)), _, _) :- !,
    '$add_file_info'('$goal'(_)),
    assert('$goal'(consult(G))),
    consult(G).
'$process_input_top'('?-'(G), _, _) :- !,
    '$add_file_info'('$goal'(_)),
    assert('$goal'(G)),
    call(G).
'$process_input_top'(T, VarNames, Flag) :-
    (
      var(Flag)
    ->
      (
        once('$current_consulted_file'(File)), '$top_level_consulted_file'(File)
      ->
        once('$check_asserted_files')
      ;
        true
      ),
      Flag = set
    ;
      true
    ),
    '$process_input'(T, VarNames).

'$process_input'((resources(Xs)), _) :-
    !,
    '$process_input_term'(resources(Xs)).
'$process_input'(durative(Lst), _) :-
    !,
    '$process_input_term'(durative(Lst)).
'$process_input'(discrete(Lst), _) :-
    !,
    '$process_input_term'(discrete(Lst)).
'$process_input'((A ::= AT), VarNames) :-
    !,
    '$process_input_term'(defined_type(A, AT, VarNames)).
%% Declaration of what TR programs can start a task
'$process_input'(task_start(((A1,A2):T)), VarNames) :-
    !,
    '$tuple2list'((A1,A2), As),
    forall(member(A, As),
           ('$process_input_term'(task_start(A)),
	    '$process_input_term'(tr_prog_type_declaration(A,T, VarNames)))).
'$process_input'((task_start(A:T)), VarNames) :-
    !,
    '$process_input_term'(task_start(A)),
    '$process_input_term'(tr_prog_type_declaration(A,T, VarNames)).

%% The task_atomic declarations declare TR programs that grab resources
'$process_input'(task_atomic(((A1,A2):T)), VarNames) :-
    !,
    '$tuple2list'((A1,A2), As),
    forall(member(A, As),
           ('$process_input_term'(task_atomic(A)),
	    '$process_input_term'(tr_prog_type_declaration(A,T, VarNames)))).
'$process_input'((task_atomic(A:T)), VarNames) :-
    !,
    '$process_input_term'(task_atomic(A)),
    '$process_input_term'(tr_prog_type_declaration(A,T, VarNames)).

%% declare type
'$process_input'(belief(B), _) :-
    !,
    '$process_input_term'(belief_type_declaration(B)).
'$process_input'(percept(B), _) :-
    !,
    '$process_input_term'(percept_type_declaration(B)).
'$process_input'(global_int_decl(A,B), _) :-
    !,
    '$process_input_term'(global_int_decl(A,B)).
'$process_input'(global_num_decl(A,B), _) :-
    !,
    '$process_input_term'(global_num_decl(A,B)).
'$process_input'(((A1,A2):(B -> C)), VarNames) :-
    !,
    '$tuple2list'((A1,A2), As),
    forall(member(A, As),
           '$process_input_term'(function_type_declaration(A, B, C, VarNames))).
'$process_input'((A:(B -> C)), VarNames) :-
    !,
    '$process_input_term'(function_type_declaration(A,B,C,VarNames)).
'$process_input'(((A1,A2):(B <=)), VarNames) :-
    !,
    '$tuple2list'((A1,A2), As),
    forall(member(A, As),
           '$process_input_term'(relation_type_declaration(A,B, VarNames))).
'$process_input'(((A1,A2):(~>>(B))), VarNames) :-
    !,
    '$tuple2list'((A1,A2), As),
    forall(member(A, As),
           '$process_input_term'(action_type_declaration(A,B, VarNames))).
'$process_input'((A:(B <=)), VarNames) :-
    !,
    '$process_input_term'(relation_type_declaration(A,B, VarNames)).
'$process_input'((A:(~>>(B))), VarNames) :-
    !,
    '$process_input_term'(action_type_declaration(A,B, VarNames)).
'$process_input'((A:(B ~>)), VarNames) :-
    !,
    '$process_input_term'(tr_prog_type_declaration(A,B, VarNames)).
'$process_input'(((A1,A2):B), VarNames) :-
    !,
    '$tuple2list'((A1,A2), As),
    forall(member(A, As),
           '$process_input_term'(unknown_type_declaration(A,B, VarNames))).
'$process_input'((A:(B<=)), VarNames) :-
    !,
    '$process_input_term'(relation_type_declaration(A,B, VarNames)).
'$process_input'((A:B), VarNames) :-
    !,
    '$process_input_term'(unknown_type_declaration(A,B, VarNames)).
%% Save TR programs
'$process_input'(('$tr_prog'(TRHead, TRBody)), _) :-
    !,
    assert('$tr_prog'(TRHead, TRBody)).

%% clause for definitions for functions/relations/actions
'$process_input'((Head -> Body), _) :-
    !,
    '$process_input_term'(clause((Head -> Body))).
'$process_input'((Head ~>> Body), _) :-
    !,
    '$process_input_term'(clause((Head ~>> Body))).
'$process_input'('~>>'(Head), _) :-
    !,
    '$process_input_term'(clause('~>>'(Head))).
'$process_input'(Clause, _) :-
    !,
    '$process_input_term'(clause(Clause)).

'$check_asserted_files' :-
    '$saved_consulted'(_), !,
    '$top_level_consulted_file'(File),
    findall(F, ('$saved_consulted'(F), F \= File), OldConsulted),
    %once('$current_consulted_file'(File)),
    findall(File1, '$consulted'(File1), NewConsulted),
    retractall('$saved_consulted'(_)),
    (
      OldConsulted = NewConsulted
    ->
      true
    ;
      writeL(stderr, [nl_, "The consulted files have changed - restarting consult", nl_]),
      throw(consult_error(consults_diff))
    ), !.
'$check_asserted_files'.

%% Sometimes a definition will have a variable representing a function
%% applied to arguments ( like using call in Prolog) - e.g.
%% mapF(F, [H|T]) -> [F(H)|mapF(F, T)]
%% This is translated to use apply(F, H, Result)
%% and to see determine what apply does to a given function we generate
%% auxilary clauses - e.g. for the defn of factorial we would generate
%% 'factorial$apply'(fact, 0, A) :- factorial(0, A).
%% 'factorial$apply'(fact, A, B) :- factorial(A, B).
%% where factorial/2 is the translation of the factorial function
%% The definition of apply constructs the appropriate call on the
%% above auxilary predicate.
%% For higer-order predicates like curry we have definitions for
%% when all the possible number of args are already applied :
%% 'curry$apply'(curry(A)(B), C, D) :- %do the evaluation - all args given
%% 'curry$apply'(curry(A), B, curry(A)(B)). % build a structure
%% 'curry$apply'(curry, A, curry(A)).
%% This technique is also used for higher-order relations.
'$process_apply_clause'(Head, Body) :-
    functor(Head, F, N),
    functor(Template, F, N),
    '$add_file_info'(Template),
    Head = Apply(Fun, _Args, _Ret),
    (
      current_predicate(Apply/3)
    ->
      true
    ;
      '$process_apply_clause_aux'(Apply, Fun, _)
    ),
    assert((Head :- Body)).

'$process_apply_clause_aux'(_, Fun, Fun) :-
    atom(Fun), !.
'$process_apply_clause_aux'(Apply, Fun, TemplateFun) :-
    Fun =.. [F|Args],
    '$process_apply_clause_aux'(Apply, F, TemplateF),
    length(Args, N),
    length(TempArgs, N),
    '$list2tuple'(TempArgs, Tuple),
    TemplateFun =.. [TemplateF|TempArgs],
    assert(Apply(TemplateF, Tuple, TemplateFun)).


%% If a type declaration is left off for a definition then
%% the system will give it a "general type". An error is produced
%% if this is a redefinition of a system defn.
'$check_if_clauses_have_types' :-
    '$tmp_clause'(clause(Clause)),
    freeze_term(Clause),
    '$check_clause_head_for_type'(Clause),
    fail.
'$check_if_clauses_have_types'.


 %% check that in t1(X) ::= t2, the X occurs on RHS
'$check_defined_type_vars' :-
    once('$current_consulted_file'(File)),
    '$user_type_info'(_P, defined_type, TName, T, VarNames),
    '$saved_input_with_vars'((TName ::= T), _, File),
    collect_vars(TName, HeadVars),
    collect_vars(T, BodyVars),
    (
      diff_list(HeadVars, BodyVars, [])
    ->
      true
    ;
      '$display_error'(invalid_type_unmatched_vars((TName ::= T), VarNames))
    ),
    fail.
'$check_defined_type_vars'.


%% Look to see if a relation/action/function/TR has an associated
%% consistent type - if not add a general type involving term as the type
'$check_clause_head_for_type'('?-'(_)) :- !.
'$check_clause_head_for_type'((A :: _B)) :-
    !,
    '$check_clause_head_for_type_aux'(A, relation).
'$check_clause_head_for_type'((A :: _B -> _C)) :-
    !,
    '$check_clause_head_for_type_aux'(A, function).
'$check_clause_head_for_type'((A -> _C)) :-
    !,
    '$check_clause_head_for_type_aux'(A, function).
'$check_clause_head_for_type'((A :: _B <= _C)) :-
    !,
    '$check_clause_head_for_type_aux'(A, relation).
'$check_clause_head_for_type'((A <= _C)) :-
    !,
    '$check_clause_head_for_type_aux'(A, relation).
'$check_clause_head_for_type'((A :: _B ~>> _C)) :-
    !,
    '$check_clause_head_for_type_aux'(A, action).
'$check_clause_head_for_type'((A ~>> _C)) :-
    !,
    '$check_clause_head_for_type_aux'(A, action).
'$check_clause_head_for_type'('$tr_prog'(A, _)) :-
    !,
    '$check_clause_head_for_type_aux'(A, tr_prog).
'$check_clause_head_for_type'(A) :-
    !,
    '$check_clause_head_for_type_aux'(A, relation).

%% The user as made a type declaration - nothing to do
'$check_clause_head_for_type_aux'(A, _) :-
    '$get_ultimate_functor'(A, F),
    %functor(A, F, _),
    '$user_type_info'(F, _, _, _, _), !.
%% The user is defining something that has a builtin type
%% but is intended for the user to define  - nothing to do
'$check_clause_head_for_type_aux'(A, _) :-
    '$get_ultimate_functor'(A, F),
    '$dynamic_builtin_type_info'(F, _, _, _, _), !.
%% The user is defining something that has a builtin type
%% that is not supposed to be defined by the user - error
'$check_clause_head_for_type_aux'(A, _K) :-
    '$get_ultimate_functor'(A, F),
    %functor(A, F, _),
    '$builtin_type_info'(F, Kind, _, _, _), !,
    (
      '$system_user_type'(F)
    ->
      true
    ;
    '$display_error'(builtin_redefinition(F, Kind))
    ).
%% Otherwise a generic type declaration is added
'$check_clause_head_for_type_aux'(A, K) :-
    functor(A, F, N),
    '$add_file_info'('$user_type_info'(F, _, _, _, _)),
    findall(term, between(1, N, _), Types),
    %% Types is a list whose elements are all 'term'
    (
      K = function
    ->
      Decl = ('$tuple_type'(Types) -> term),
      assert('$user_type_info'(F, function, Decl, '$function', []))
     ;
      K = relation
    ->
      Decl = '<='('$tuple_type'(Types)),
      functor(T, F, N),
      assert('$user_type_info'(F, relation, Types, T, []))
    ;
      K = action
    ->
      functor(T, F, N),
      assert('$user_type_info'(F, action, Types, T, [])),
      Decl = '~>>'('$tuple_type'(Types))
    ;
      functor(T, F, N),
      assert('$user_type_info'(F, tr_prog, Types, T, [])),
      Decl = '~>'('$tuple_type'(Types))
    ),
    writeL(stderr, ["Warning: adding the type declaration ", F:Decl, nl_]).

%% process all the terms that come from the parser (through process_input)

%%'$process_input_term'(A) :- errornl('$process_input_term'(A)),fail.

%% A declaration has been made but alternative types are used
%% so the kind (relation/action/function/TR) needs to be determined
'$process_input_term'(unknown_type_declaration(A, alt_types(B), VarNames)) :-
    !,
    (
      '$type_info'(A, _, _, Type, _)
    ->
      %% there is already a type for A
      '$real_kind'(A, Kind),
      '$vars2names'((A:B), VarNames),
      '$display_error'(re_declaration_alts(A, (A:alt_types(B)), Kind, Type))
    ;
      freeze_term(B), B = [T1|Ts], '$consistent_alt_types'(T1, Ts), thaw_term(B)
    ->
      %%the alternative types are consistent
      '$assert_alt_types'(B,A,VarNames)
    ;
      '$vars2names'((A:B), VarNames),
      '$display_error'(inconsistent_declarations(A, B))
    ).
%% When the relation/action/TR is missing from a declaration
%% then it becomes of unknown type and is subsequently processed
%% when clauses have been added
'$process_input_term'(unknown_type_declaration(A, B, VarNames)) :-
    !,
    (
      '$type_info'(A, _, _, _, _)
    ->
      '$real_kind'(A, _Kind),
      '$vars2names'((A:B), VarNames),
      '$display_error'(re_declaration(A, B))
    ;
      (
        B = '$tuple_type'(Types)
      ->
        true
      ;
        Types = [B]
      ),
      length(Types, N),
      functor(RC, A, N),
     '$add_file_info'('$user_type_info'(A, _, _, _, _)),
      assert('$user_type_info'(A, unknown, Types, RC, VarNames))
    ).
'$process_input_term'(relation_type_declaration(A, B, VarNames)) :-
    !,
    (
      '$type_info'(A, _, _, _, _)
    ->
      '$real_kind'(A, Kind),
      '$vars2names'((A:B), VarNames),
      '$display_error'(re_declaration_kind(A, (A:'<='(B)), Kind))
    ;
      (
        B = '$tuple_type'(Types)
      ->
        true
      ;
        Types = [B]
      ),
      length(Types, N),
      length(Args, N),
      '@..'(A, Args, RC),
      %functor(RC, A, N),
     '$add_file_info'('$user_type_info'(A, _, _, _, _)),
      assert('$user_type_info'(A, relation, Types, RC, VarNames))
    ).
'$process_input_term'(action_type_declaration(A, B, VarNames)) :-
    !,
    (
      '$type_info'(A, _, _, _, _)
    ->
      '$real_kind'(A, Kind),
      '$vars2names'((A:B), VarNames),
      '$display_error'(re_declaration_kind(A, (A:'~>>'(B)), Kind))
    ;
      (
        B = '$tuple_type'(Types)
      ->
        true
      ;
        Types = [B]
      ),
      length(Types, N),
      length(Args, N),
      '@..'(A, Args, RC),
      %functor(RC, A, N),
     '$add_file_info'('$user_type_info'(A, _, _, _, _)),
      assert('$user_type_info'(A, action, Types, RC, VarNames))
    ).
'$process_input_term'(function_type_declaration(FName, HeadTypes, ResultType,
                                          VarNames)) :-
    !,
    (
      '$type_info'(FName, _, _, _, _)
    ->
      '$real_kind'(FName, Kind),
      Term = (FName:(HeadTypes -> ResultType)),
      '$vars2names'(Term, VarNames),
      '$display_error'(re_declaration_kind(FName, Term, Kind))
    ;
      '$get_full_goal_term'(FName, (HeadTypes -> ResultType), GT),
      '$add_file_info'('$user_type_info'(FName, _, _, _, _)),
      assert('$user_type_info'(FName, function, (HeadTypes -> ResultType),
                               GT, VarNames))
    ).

%% declarations for beliefs and percepts create relation type facts,
%% belief or percept facts, and make the predicates dynamic 
'$process_input_term'(belief_type_declaration(Xs)) :-
    closed_list(Xs),!,
    (
     forall(member(X, Xs), '$process_input_term'(belief_type_declaration(X)))
    ->
     true
    ;
     true
    ).
'$process_input_term'(belief_type_declaration(B)) :-
    !,
    functor(B, FName, _),
    (
      '$type_info'(FName, _ , _, _, _)
    ->
      '$real_kind'(FName, Kind),
      B =.. [_|Args],
      TArgs = '$tuple'(Args),
      '$display_error'(re_declaration_kind(FName, belief(FName:TArgs), Kind))
    ;
      atom(B)
    ->
     '$add_file_info'('$user_type_info'(B, _, _, _, _)),
      assert('$user_type_info'(B, relation, [], B, [])),
      assert('$belief'(B)),
      dynamic(FName/0)
    ;
      B =.. [F|Types],
      length(Types, N),
      functor(RC, F, N),
      dynamic(F/N),
     '$add_file_info'('$user_type_info'(F, _, _, _, _)),
      assert('$user_type_info'(F, relation, Types, RC, [])),
      assert('$belief'(RC))
    ).
'$process_input_term'(percept_type_declaration(Xs)) :-
    closed_list(Xs),!,
    (
     forall(member(X, Xs), '$process_input_term'(percept_type_declaration(X)))
    ->
     true
    ;
     true
    ).
'$process_input_term'(percept_type_declaration(B)) :-
    !,
    functor(B, FName, _),
    (
      '$type_info'(FName, _ , _, _, _)
    ->
      '$real_kind'(FName, Kind),
      B =.. [_|Args],
      TArgs = '$tuple'(Args),
      '$display_error'(re_declaration_kind(FName, percept(FName:TArgs), Kind))
    ;
      atom(B)
    ->
     '$add_file_info'('$user_type_info'(B, _, _, _, _)),
      assert('$user_type_info'(B, relation, [], B, [])),
      assert('$percept'(B)),
      dynamic(FName/0)
    ;
      B =.. [F|Types],
      length(Types, N),
      functor(RC, F, N),
      dynamic(F/N),
     '$add_file_info'('$user_type_info'(F, _, _, _, _)),
      assert('$user_type_info'(F, relation, Types, RC, [])),
      assert('$percept'(RC))
    ).

%% ditto for state beliefs as well as a state_belief fact added
'$process_input_term'(global_int_decl(A,B)) :-
     !,
    (
      '$type_info'(A, _ , _, _, _)
    ->
      '$real_kind'(A, Kind),
      '$display_error'(re_declaration_kind(A, global_int_decl(A, B), Kind))
    ;
     dynamic(A/1),
     '$add_file_info'('$user_type_info'(A, _, _, _, _)),
     assert('$user_type_info'(A, relation, [int], A(_), [])),
     '$add_file_info'('$belief'(A(_))),
     '$add_file_info'('$state_belief'(A)),
     assert('$belief'(A(_))),
     assert('$state_belief'(A)),
     assert('$tmp_clause'(clause(A(B))))
    ).
'$process_input_term'(global_num_decl(A,B)) :-
     !,
    (
      '$type_info'(A, _ , _, _, _)
    ->
      '$real_kind'(A, Kind),
      '$display_error'(re_declaration_kind(A, global_num_decl(A, B), Kind))
    ;
      dynamic(A/1),
     '$add_file_info'('$user_type_info'(A, _, _, _, _)),
      assert('$user_type_info'(A, relation, [num], A(_), [])),
     '$add_file_info'('$belief'(A(_))),
     '$add_file_info'('$state_belief'(A(_))),
      assert('$belief'(A(_))),
      assert('$state_belief'(A)),
      assert('$tmp_clause'(clause(A(B))))
    ).
%% Anything that is a clause defn is saved away
'$process_input_term'(clause(A)) :-
    !,
    assert('$tmp_clause'(clause(A))).

%% process TR type declarations
'$process_input_term'(tr_prog_type_declaration(A, B, VarNames)) :-
    !,
    (
      '$type_info'(A, _, _, _, _)
    ->
      '$real_kind'(A, Kind),
      '$vars2names'((A, B), VarNames),
      '$display_error'(re_declaration_kind(A, B, Kind))
    ;
      B = '$tuple_type'(Types),
      length(Types, N),
      functor(RC, A, N),
     '$add_file_info'('$user_type_info'(A, _, _, _, _)),
      assert('$user_type_info'(A, tr_prog, Types, RC, VarNames))
    ;
      functor(RC, A, 1),
     '$add_file_info'('$user_type_info'(A, _, _, _, _)),
      assert('$user_type_info'(A, tr_prog, [B], RC, VarNames))
    ).

%% process defined types
'$process_input_term'(defined_type(A, AT, VarNames)) :-
    functor(A, P, _N),
    '$type_info'(A, Kind, _, _, _), !,
    %% A already has type declaration
    '$vars2names'((A, AT), VarNames),
    '$display_error'(re_declaration_kind(P, (A ::= AT), Kind)).
 
'$process_input_term'(defined_type(A, AT, VarNames)) :-
    !,
    (
      AT = '$enum_type'(_)
    ->
      RealAT = AT
    ;
      AT = '$constr_enum_type'(_)
    ->
      RealAT = AT
    ;
      AT = (_ .. _)
    ->
      RealAT = AT
    ;
      AT = '$set_type'(_)
    ->
      RealAT = AT
    ;
      AT = '$list_type'(_)
    ->
      RealAT = AT
    ;
      AT = '$union_type'(_)
    ->
      RealAT = AT
    ;
      %% Anything else is treated as a type macro
      RealAT = '$type_macro'(AT)
    ),
    functor(A, P, _N),
    '$add_file_info'('$user_type_info'(P, _, _, _, _)),
    assert('$user_type_info'(P, defined_type, A, RealAT, VarNames)).
%% processing resource declarations
'$process_input_term'(resources(R)) :-
    '$resource'(_), !,
    writeL(stderr, ["Warning: there is already a resource declaration.",
			     nl_, "The declaration ", resources(R),
			     " is ignored.", nl_]).
'$process_input_term'(resources(Rs)) :-
    !,
    forall(member(R, Rs), ('$add_file_info'('$resource'(R)),
			   assert('$resource'(R)))).

%% process task_atomic
'$process_input_term'(task_atomic(D)) :-
    !,
    '$assert_task_atomic'(D).

%% process task_start    
'$process_input_term'(task_start(D)) :-
    !,
    (
      closed_list(D)
    ->
      forall(member(D1, D), ('$add_file_info'('$task_start'(D1)),
			     assert('$task_start'(D1))))
    ;
     '$add_file_info'('$task_start'(D)),
     assert('$task_start'(D))
    ).

%% process durative
'$process_input_term'(durative(Xs)) :-
    closed_list(Xs),!,
    (
     forall(member(X, Xs), '$process_input_term'(durative(X)))
    ->
     true
    ;
     true
    ).
'$process_input_term'(durative(D)) :-
    !,
    (
     '$is_durative_decl'(D)
    ->
     functor(D, P, N),
     (
      '$type_info'(_, Kind, _ , D, _)
     ->
      '$display_error'(re_declaration_kind(P, durative(D), Kind))
     ;
      D =.. [_|Types],
      functor(DC, P, N),
     '$add_file_info'('$user_type_info'(P, _, _, _, _)),
      assert('$user_type_info'(P, durative, Types, DC, [])),
      '$assert_resource_info'(P)
     )
    ;
      '$display_error'(invalid_declaration(durative(D)))
    ).
%% proces discrete
'$process_input_term'(discrete(Xs)) :-
    closed_list(Xs),!,
    (
     forall(member(X, Xs), '$process_input_term'(discrete(X)))
    ->
     true
    ;
     true
    ).
'$process_input_term'(discrete(D)) :-
    !,
    (
     '$is_discrete_decl'(D)
    ->
     functor(D, P, N),
     (
      '$type_info'(_, Kind, _ , D, _)
     ->
      '$display_error'(re_declaration_kind(P, discrete(D), Kind))
     ;
      D =.. [_|Types],
      functor(DC, P, N),
     '$add_file_info'('$user_type_info'(P, _, _, _, _)),
      assert('$user_type_info'(P, discrete, Types, DC, [])),
      '$assert_resource_info'(P)
     )
    ;
      '$display_error'(invalid_declaration(discrete(D)))
    ).
%% Should never get here
'$process_input_term'(D) :-
    writeL(stderr, ["Error: unprocessed term ", D, nl_]).

%% all the alternative types in Ts are consistent with T1    
'$consistent_alt_types'(T1, Ts) :-
    '$consistent_alt_types_pattern'(T1, P),
    '$consistent_alt_types_aux'(Ts, P).

'$consistent_alt_types_aux'([], _).
'$consistent_alt_types_aux'([T|Ts], P) :-
    '$consistent_alt_types_pattern'(T, P1),
    P = P1,
    '$consistent_alt_types_aux'(Ts, P).

'$consistent_alt_types_pattern'('<='(T1), P) :-
    !,
    '$consistent_alt_types_pattern'(T1, P1),
    P = '<='(P1).
'$consistent_alt_types_pattern'('~>>'(T1), P) :-
    !,
    '$consistent_alt_types_pattern'(T1, P1),
    P = '~>>'(P1).
'$consistent_alt_types_pattern'('~>'(T1), P) :-
    !,
    '$consistent_alt_types_pattern'(T1, P1),
    P = '~>'(P1).
'$consistent_alt_types_pattern'((T1 -> T2), P) :-
    !,
    '$consistent_alt_types_pattern'(T1, P1),
    '$consistent_alt_types_pattern'(T2, P2),
    P = (P1 -> P2).
'$consistent_alt_types_pattern'('$tuple_type'(T), P) :-
    !,
    length(T, N),
    length(P, N).
'$consistent_alt_types_pattern'(_, atom).

%% if modes are left off relation or action type declarations
%% then add default modes to the types
'$add_default_modes' :-
    '$user_type_info'(F, Kind, Types, Term, Vars),
    ( Kind = relation ; Kind = action ),
    (
      ( '$percept_type_info'(_, _, Term) ; '$belief_type_info'(_, _, Term) )
    ->
      Ann = '?'
    ;
      Ann = '!'
    ),
    '$add_annotations'(Types, ATypes, Ann),
    (
      Types == ATypes
    ->
      true
    ;
      retract('$user_type_info'(F, Kind, Types, Term, Vars)),
      assert('$user_type_info'(F, Kind, ATypes, Term, Vars))
    ),
    fail.
'$add_default_modes'.
      
%% When there are alternative types for a function/relation/action/TR
%% then assert a '$user_type_info' fact for each alternative
'$assert_alt_types'([],_,_VarNames).
'$assert_alt_types'([B|Bs],A,VarNames) :-
    '$add_file_info'('$user_type_info'(A, _, _, _, _)),
    (
      B = (HeadTypes -> ResultType)
    ->
      '$get_full_goal_term'(A, B, GT),
      assert('$user_type_info'(A, function, (HeadTypes -> ResultType),
                               GT, VarNames))
    ;
      B = '<='(B1)
    ->
      (
        B1 = '$tuple_type'(Types)
      ->
        true
      ;
        Types = [B1]
      ),
      length(Types, N),
      functor(RC, A, N),
      assert('$user_type_info'(A, relation, Types, RC, VarNames))
    ;
      B = '~>>'(B1)
    ->
      (
        B1 = '$tuple_type'(Types)
      ->
        true
      ;
        Types = [B1]
      ),
      length(Types, N),
      functor(RC, A, N),
      assert('$user_type_info'(A, action, Types, RC, VarNames))
    ;
      B = '~>'(B1)
    ->
      (
        B1 = '$tuple_type'(Types)
      ->
        true
      ;
        Types = [B1]
      ),
      length(Types, N),
      functor(RC, A, N),
      assert('$user_type_info'(A, tr_prog, Types, RC, VarNames))
    ;
      (
        B = '$tuple_type'(Types)
      ->
        true
      ;
        Types = [B]
      ),
      length(Types, N),
      functor(RC, A, N),
      assert('$user_type_info'(A, unknown, Types, RC, VarNames))
    ),
    '$assert_alt_types'(Bs,A,VarNames).

%%'$assert_task_atomic'(A+B) :-
%%    !,
%%    '$add_file_info'('$task_atomic'(A, _)),
%%    assert('$task_atomic'(A,B)).
'$assert_task_atomic'(A) :-
    '$add_file_info'('$task_atomic'(A, _)),
    assert('$task_atomic'(A,[])).

%% Translate clauses (rules) into Prolog clauses
%%'$translate_clause'(X) :- errornl('$translate_clause'(X)),fail.
'$translate_clause'((Head :: Test -> Body)) :-
    !,
    '$translate_function'(Head, Test, Body).
'$translate_clause'((Head -> Body)) :-
    !,
    '$translate_function'(Head, true, Body).
'$translate_clause'((Head :: Test <= Body)) :-
    !,
    '$translate_body'(((Test & !) & Body), TransBody),
    '$assert_the_relation_clause'(Head, TransBody).
'$translate_clause'((Head <= Body)) :-
    !,
    '$translate_body'((Body), TransBody),
    '$assert_the_relation_clause'(Head, TransBody).
'$translate_clause'((Head :: Test ~>> Body)) :-
    !,
    '$translate_body'((Test & ! & Body), TransBody),
    '$assert_the_relation_clause'(Head, TransBody).
%% Action rules without a commit have an implicit ::true  commit
'$translate_clause'((Head ~>> Body)) :-
    !,
    '$translate_body'((!, Body), TransBody),
    '$assert_the_relation_clause'(Head, TransBody).
'$translate_clause'((Head :: Test)) :-
    !,
    '$translate_body'((Test & !), TransBody),
    '$assert_the_relation_clause'(Head, TransBody).
'$translate_clause'(Head) :-
    !,
    (
     '$type_info'(_, action, _, Head, _)
    ->
      %% implicit :: true commit
     '$assert_the_relation_clause'(Head, !)
    ;
     '$assert_the_relation_clause'(Head, true)
    ).

%% Add the translated rule to the DB

%% Special case when we have a higher order relation
%% These are functions that, given all but the last collection
%% of args returns a relation 
'$assert_the_relation_clause'(Head, TransBody) :-
    compound(Head), functor(Head, F, _), compound(F), !,
    Head =.. [F|Args],
    '$list2tuple'(Args, ArgsTuple),
    '$generate_apply_name'(F, ApplyName),
    A1 = ApplyName(F, ArgsTuple, true),
    freeze_term((A1, TransBody)),
    '$process_apply_clause'(A1, TransBody).
%% clauses about percepts are processed so that
%% forward inference can be done when percepts change
%% should_be_remembered_/1 is used to remove
%% infered percept facts that can no longer be infered
%% can_infer_now_/1 are used to infer percept facts
'$assert_the_relation_clause'(Head, true) :-
    '$percept'(Head),
    should_be_remembered_(Head), !,
    functor(Head, F, N),
    functor(Template, F, N),
    '$add_file_info'(can_infer_now_(Template)),
    assert(can_infer_now_(Head)).
'$assert_the_relation_clause'(Head, true) :-
    '$percept'(Head),
    functor(Head, F, N),
    functor(Template, F, N),
    '$add_file_info'(should_be_remembered_(Template)),
    '$add_file_info'(can_infer_now_(Template)),
    assert(should_be_remembered_(Template)),
    assert(can_infer_now_(Head)).
'$assert_the_relation_clause'(Head, true) :-
    !,
    functor(Head, F, N),
    functor(Template, F, N),
    '$add_file_info'(Template),
    assert(Head).
'$assert_the_relation_clause'(Head, TransBody) :-
    '$percept'(Head),
    should_be_remembered_(Head), !,
    functor(Head, F, N),
    functor(Template, F, N),
    '$add_file_info'(can_infer_now_(Template)),
    assert((can_infer_now_(Head) :- TransBody)).
'$assert_the_relation_clause'(Head, TransBody) :-
    '$percept'(Head), !, 
    functor(Head, F, N),
    functor(Template, F, N),
    '$add_file_info'(should_be_remembered_(Template)),
    '$add_file_info'(can_infer_now_(Template)),
    assert(should_be_remembered_(Template)),
    assert((can_infer_now_(Head) :- TransBody)).
'$assert_the_relation_clause'(Head, TransBody) :-
    functor(Head, F, N),
    functor(Template, F, N),
    '$add_file_info'(Template),
    assert((Head :- TransBody)).

%% For translating a function
%% f(A,B) -> Term
%% we translate first to
%% f(A, B, X) as the head and Term = X as the body
%% and then translate the body.
%% This converts function clauses into relation clauses where
%% the last argument is the result

%% Special case for higher-order functions
'$translate_function'(Head, Test, Body) :-
     '$get_function_type_info'(Head,HeadT, _, ResultType),
    length(HeadT, Len),
    Len \= 1, !,
    '$add_fun_args'(ResultType, Head, FullHead, Body, FullBody),
    freeze_term(FullBody),
    FBody = (FullBody = X),
    FullHead =.. [F|Args],
    '$list2tuple'(Args, ArgsTuple),
    '$translate_body'(((Test, !), FBody), Unfolded),
    '$generate_apply_name'(F, ApplyName),
    A1 = ApplyName(F, ArgsTuple, X),
    '$process_apply_clause'(A1, Unfolded).
'$translate_function'(Head, Test, Body) :-
     FBody = (Body = X),
    freeze_term(X),
    Head =.. P,
    append(P, [X], P1),
    A1 =.. P1,
    P = [F|Args],
    '$generate_apply_name'(F, ApplyName),
    '$add_file_info'(ApplyName(_, _, _)),
    (
      current_predicate(ApplyName/3)
    ->
      true
    ;
      length(Args, ArgN),
      length(TempArgs, ArgN),
      '$list2tuple'(TempArgs, TempArgTuple),
      append(TempArgs, [TempX], AllArgs),
      TempA1 =.. [F|AllArgs],
      assert((ApplyName(F, TempArgTuple, TempX) :- TempA1))
    ),
    '$translate_body'(((Test, !), FBody), Unfolded),
    functor(A1, F1, N1),
    functor(Template, F1, N1),
    '$add_file_info'(Template),
    assert((A1 :- Unfolded)).

%% for translating the body we freeze the vars in goals so
%% they don't get matched against patterns
'$translate_body'(Goals, TransGoals) :-
    freeze_term(Goals),
    '$translate_goal'(Goals, TransGoals).

%% For higher order functions it's possible to write down a rule
%% where the LHS doesn't include all args.
%% EG
%% f : t1 -> (t2, t3) -> t4 -> t5
%% f(X) -> g(h(X))
%% Given this rule is correctly typed then we could rewrite as
%% f(X)(Y,Z)(W) -> g(h(X))(Y,Z)(W)
%% This predicate adds any "missing" arguments
'$add_fun_args'(Type, Term, OutTerm, Body, OutBody) :-
    compound(Type), Type = (T1 -> T2), !,
    '$tuple2list'(T1, T1L),
    length(T1L, N),
    length(Args, N),
    Term1 =.. [Term|Args],
    Body1 =.. [Body|Args],
    '$add_fun_args'(T2, Term1, OutTerm, Body1, OutBody).
'$add_fun_args'(_Type, Term, Term, Body, Body).


/*
'$fix_last_call'(G, FG) :-
    (
      var(G)
    ->
      FG = print__(G)
    ;
      G = (G1, G2)
    ->
      '$fix_last_call'(G2, FG2),
      FG = (G1, FG2)
    ;
     list(G)
    ->
     FG = print__(G)
    ;
      FG = G
    ).

'$wrap_expressions'(T, T) :-
    nonvar(T), T = (_, _), !.
'$wrap_expressions'(T, WT) :-
    var(T), !,
    WT = print__(T).
'$wrap_expressions'(T, WT) :-
    T = match(_, _),
    !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = unify(_, _),
    !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    '$type_info'(_, relation, _, T, _), !,
    WT = T.

'$wrap_expressions'(T, WT) :-
    T = type(_, _), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = type(_, _, _), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = ground(_), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = '$at_forall'(_, _, _), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = '$at_forall_actions'(_, _, _), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = '$forall'(_, _, _), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = '$exists'(_, _), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = '$forall_actions'(_, _, _), !,
    WT = T.
'$wrap_expressions'(T, WT) :-
    T = print__(_),!,
    WT = T.

'$wrap_expressions'(T, WT) :-
    WT = print__(T).
*/

/*
'$unfold_function_calls'(A,B) :-
    '$unfold_function_calls_aux'(A,B1),
    '$fix_last_call'(B1, B).
*/

%% translate qulog goals to prolog goals
%% Note that arguments in goals might contain expressions that
%% need to be evaluated. These evaluations will be unfolded to prolog
%% goals that are called before the goal is called
%'$translate_goal'(X,_) :- errornl('$translate_goal'(X)),fail.

%% For goals whose functors are not atoms we need to go through apply
'$translate_goal'(G, TG) :-
    compound(G), G =.. [F|Args], var(F), !,
    '$list2tuple'(Args, ArgsTuple),
    TG = apply(F, ArgsTuple, _).
'$translate_goal'((B1 & B2), (TB1, TB2)) :-
    !,
    '$translate_goal'(B1, TB1),
    '$translate_goal'(B2, TB2).
'$translate_goal'((B1 ; B2), (TB1, TB2)) :-
    !,
    '$translate_goal'(B1, TB1),
    '$translate_goal'(B2, TB2).
'$translate_goal'((B1 , B2), (TB1, TB2)) :-
    !,
    '$translate_goal'(B1, TB1),
    '$translate_goal'(B2, TB2).

/*
'$translate_goal'(G, TG) :-
    '$is_a_type_test'(G, Term, Type),
    '$expand_funs'(Term, TermT, TermA),
    append(TermA, [ground(TermT), type(TermT, Type)], Goals),
    '$list2tuple'(Goals, TG),
    !.           
*/
'$translate_goal'(G, RG) :-
    list(G),
    !,
    findall(qconsult(F), member(F, G), Query),
    '$list2tuple'(Query, RG).
'$translate_goal'(remember(G), RG) :-
    '$expand_funs'(G, GT, GA),
    append(GA, [remember(GT)], Goals),
    '$list2tuple'(Goals, RG),
    !.
'$translate_goal'(remember_for(G,T), RG) :-
    '$expand_funs'(G, GT, GA),
    '$expand_funs'(T, TT, TG),
    append(GA, TG, GG),
    append(GG, [remember_for(GT, TT)], Goals),
    '$list2tuple'(Goals, RG),
    !.
  '$translate_goal'(rememberA(G), RG) :-
    '$expand_funs'(G, GT, GA),
    append(GA, [rememberA(GT)], Goals),
    '$list2tuple'(Goals, RG),
    !.
'$translate_goal'(rememberA_for(G,T), RG) :-
    '$expand_funs'(G, GT, GA),
    '$expand_funs'(T, TT, TG),
    append(GA, TG, GG),
    append(GG, [rememberA_for(GT, TT)], Goals),
    '$list2tuple'(Goals, RG),
    !.
'$translate_goal'(forget(G), FG) :-
    '$expand_funs'(G, GT, GA),
    append(GA, [forget(GT)], Goals),
    '$list2tuple'(Goals, FG),
    !.
'$translate_goal'(forget_after(G,T), RG) :-
    '$expand_funs'(G, GT, GA),
    '$expand_funs'(T, TT, TG),
    append(GA, TG, GG),
    append(GG, [forget_after(GT, TT)], Goals),
    '$list2tuple'(Goals, RG),
    !.
'$translate_goal'(forgetall(G), FG) :-
    '$expand_funs'(G, GT, GA),
    append(GA, [forgetall(GT)], Goals),
    '$list2tuple'(Goals, FG),
    !.
%% TODO
'$translate_goal'(replace_by(G1, G2), RG) :-
    collect_vars(G1, G1Vars),
    collect_vars(G2, G2Vars),
    intersect_list(G1Vars, G2Vars, []), !,
    '$expand_funs'(G1, G1T, G1A),
    '$expand_funs'(G2, G2T, G2A),
    append(G1A, G2A, GA),
    append(GA, ['$quick_replace'(G1T, G2T)], Goals),
    '$list2tuple'(Goals, RG).
%% TODO
'$translate_goal'(replace_by(G1, G2), G) :-
    !,
    '$translate_goal'(forget(G1), RG1),
    '$translate_goal'(remember(G2), RG2),
    (
      RG1 == forget(G1),
      RG2 == remember(G2)
    ->
      %% there were no expressions to unfold
      G = '$quick_replace'(G1, G2)
    ;
      G = '$replace_by_atomic'((RG1, RG2))
    ).
'$translate_goal'(call(G), FG) :-
    '$expand_funs'(G, GT, GA),
    append(GA, [call_(GT)], Goals),
    '$list2tuple'(Goals, FG),
    !.
'$translate_goal'(do(G), FG) :-
    '$expand_funs'(G, GT, GA),
    append(GA, [do_(GT)], Goals),
    '$list2tuple'(Goals, FG),
    !.
'$translate_goal'(once(B), once(G)) :-
    !,
    '$translate_goal'(B, G).
'$translate_goal'(not(B), not(G)) :-
    !,
    '$translate_goal'(B, G).
'$translate_goal'('$exists'(V, A), Result) :-
    free_vars(A, FreeVars),
    diff_list(FreeVars, V, Unbound),
    (
      Unbound = []
    ->
      Result = once(AT)
    ;
      Result = AT
    ),
    !,
    '$translate_goal'(A, AT).
/* 
'$translate_goal'('$at_forall'(_V, A, B), thread_atomic_goal(forall(AT, once(BT)))) :-
    B = forget(_),
    !,
    '$translate_goal'(A, AT),
    '$translate_goal'(B, BT).
'$translate_goal'('$at_forall'(_V, A, B), forall(AT, once(BT))) :-
    !,
    '$translate_goal'(A, AT),
    '$translate_goal'(B, BT).
'$translate_goal'('$at_forall_actions'(_V, A, B), forall(AT, once(BT))) :-
    !,
    '$translate_goal'(A, AT),
    '$translate_goal'(B, BT).
*/
'$translate_goal'('$forall'(_V, A, B), forall(AT, BT)) :-
    !,
    '$translate_goal'(A, AT),
    '$translate_goal'(B, BT).
'$translate_goal'('$forall_actions'(_V, A, B), forall(AT, BT)) :-
    !,
    '$translate_goal'(A, AT),
    '$translate_goal'(B, BT).
%% Example:
%% r(X, Y, Z) <= X =? Y?t(Y) ++ Z
%% translates to
%% r(X, Y, Z) :- string_concat(Y, Z, X), t(Y)
'$translate_goal'(( SExp '=?' L1 ++ L2), G) :-
    !,
    '$expandS'(SExp,L-Concats1), 
    '$expandSP'(L, (L1 ++ L2), Concats2),
    C = (Concats1,Concats2),
    freeze_term(C),
    '$translate_goal'(C, G).
%% Example:
%% p(X, Y, Z) <= X =? [Y]?q(Y) <> Z
%% translates to
%% p(X, Y, Z) :- append([Y], Z, X), q(Y)
'$translate_goal'(( LExp '=?' L1 <> L2), G) :-
    !,
    '$expandL'(LExp,L-Concats1),
    '$expandLP'(L, (L1 <> L2), Concats2),
    C = (Concats1,Concats2),
    freeze_term(C),
    '$translate_goal'(C, G).
%% The equivalent of =.. for breaking up a compound
'$translate_goal'((B '=?' '@..'(A1, A2)), G) :-
    !,
    '$expand_funs'(A1, AT1, CA1),
    '$expand_funs'(A2, AT2, CA2),
    '$expand_funs'(B, BT, CB),
    append(CA1, ['@..rev'(BT, AT1, AT2)], G1),
    append(CA2, G1, G2),
    append(CB, G2, Goals),
    '$list2tuple'(Goals, G).
%% Note: expression evaluation happens before unification
%% EG
%% 2+X = Y
%% translates to
%% +(2, X, A), unify(A, Y)
'$translate_goal'((A = B), G) :-
    !,
    '$expand_funs'(A, AT, CA),
    '$expand_funs'(B, BT, CB),
    append(CA, [unify(AT, BT)], G1),
    append(CB, G1, G2),
    '$list2tuple'(G2, G).
'$translate_goal'((A \= B), G) :-
    !,
    '$expand_funs'(A, AT, CA),
    '$expand_funs'(B, BT, CB),
    append(CA, [not_unify(AT, BT)], G1),
    append(CB, G1, G2),
    '$list2tuple'(G2, G).
/*
'$translate_goal'((A =:= B), G) :-
    !,
    '$expand_funs'(A, AT, CA),
    '$expand_funs'(B, BT, CB),
    append(CB, [match(AT, BT)|CA], Goals),
    '$list2tuple'(Goals, G).
'$translate_goal'((A =\= B), G) :-
    !,
    '$expand_funs'(A, AT, CA),
    '$expand_funs'(B, BT, CB),
    append(CB, [not_match(AT, BT)|CA], Goals),
    '$list2tuple'(Goals, G).
*/
%% For state variables
'$translate_goal'(':='(A,B), G) :-
    !,
    '$expand_funs'(B, TB, BC),
    (
      '$list2tuple'(BC, BCT)
    ->
      G = thread_atomic_goal((BCT,retractall(A(_)),assert(A(TB))))
    ;
      G = thread_atomic_goal((retractall(A(_)),assert(A(TB))))
    ).
'$translate_goal'('+:='(A,B), G) :-
    !,
    '$expand_funs'(B, TB, BC),
    (
      '$list2tuple'(BC, BCT)
    ->
      G = thread_atomic_goal((BCT, retract(A(N)), N1 is N+TB, assert(A(N1))))
    ;
      G = thread_atomic_goal((retract(A(N)), N1 is N+TB, assert(A(N1))))
    ).
'$translate_goal'('-:='(A,B), G) :-
    !,
    '$expand_funs'(B, TB, BC),
    (
      '$list2tuple'(BC, BCT)
    ->
      G = thread_atomic_goal((BCT, retract(A(N)), N1 is N-TB, assert(A(N1))))
    ;
      G = thread_atomic_goal((retract(A(N)), N1 is N-TB, assert(A(N1))))
    ).
%% Higher order relations - EG
%% curryR(A)(fact(B))(C)
%% translates to
%% fact(B, D), apply(curryR(A), D, E), 'curryR$apply'(E, C, true)
'$translate_goal'(B, G) :-
    '$type_info'(_, function, _, B, _), !,
    B =.. [F|Args],
    '$generate_apply_name'(B, ApplyName),
    '$list2tuple'(Args, Tuple),
    BCall = ApplyName(F, Tuple, true),
    '$expand_funs'(BCall, BT, C),
    (
      '$list2tuple'(C, CG)
    ->
      G = (CG, BT)
    ;
      G = BT
    ).
%% goals with var functors are unfolded in a similar way to above
'$translate_goal'(B, G) :-
    '$var_functor'(B),!,
    B =.. [F|Args],
    '$expand_funs'(Args, TArgs, C),
    BT =.. [F|TArgs],
    (
      '$list2tuple'(C, CG)
    ->
      G = (CG, BT)
    ;
      G = BT
    ).
%% default - call on predicate - args might need evaluating first
'$translate_goal'(B, G) :-
     '$expand_funs'(B, BT, C),
    (
      '$list2tuple'(C, CG)
    ->
      G = (CG, BT)
    ;
      G = BT
    ).

%% T is a term that may contain subterms that need expanding
%% T1 is the resulting term and C is the collection of auxilary
%% goals that evaluate the sub-expressions

'$expand_funs'(T, T1, C) :-
    atomic(T), !, T1 = T, C = [].
'$expand_funs'(T, T1, C) :-
    string(T), !, T1 = T, C = [].
'$expand_funs'(T, T1, C) :-
    var(T), !, T1 = T, C = [].
'$expand_funs'(T, X, C) :-
    T = '$list_constr'(V, G), !,
    '$translate_goal'(G,B),
    '$expand_funs'(V, V1, VG),
    (
      VG = []
    ->
      C = [findall(V1, B, X)]
    ;
      '$list2tuple'(VG, VGT),
      C = [findall(V1, (B,VGT), X)]
    ).
'$expand_funs'(T, X, C) :-
    T = '$set_constr'(V, G), !,
    '$translate_goal'(G,B),
    '$expand_funs'(V, V1, VG),
    (
      VG = []
    ->
      C = [findall(V1, B, X1), sort(X1, Y), X = '$set'(Y)]
    ;
      '$list2tuple'(VG, VGT),
      C = [findall(V1, (B,VGT), X1), sort(X1, Y), X = '$set'(Y)]
    ).
'$expand_funs'(T, X, C) :-
    T = '$set_enum'(B), !,
    '$expand_funs'(B, B1, BC),
    append(BC, [sort(B1, XL), X = '$set'(XL)], C).
'$expand_funs'(T, X, C) :-
    T = '$list2set'(B), !,
    '$expand_funs'(B, B1, BC),
    append(BC, [sort(B1, XL), X = '$set'(XL)], C).
'$expand_funs'(T, X, C) :-
    T = '$set2list'(B), !,
    '$expand_funs'(B, B1, BC),
    append(BC, [B1 = '$set'(X)], C).
'$expand_funs'(T, X, C) :-
    T = (L1 <> L2), !,
    '$expand_funs'(L1, EL1, L1C),
    '$expand_funs'(L2, EL2, L2C),
    append(L1C, L2C, LC),
    append(LC, [append(EL1, EL2, X)], C).
'$expand_funs'(T, X, C) :-
    T = (S1 ++ S2), !,
    '$expand_funs'(S1, ES1, S1C),
    '$expand_funs'(S2, ES2, S2C),
    append(S1C, S2C, SC),
    append(SC, [string_concat(ES1, ES2, X)], C).
'$expand_funs'(T, X, C) :-
    T = '$'(A), !,
    C = [A(X)].
'$expand_funs'(T, X, C) :-
    ('$higher_order_term'(T) ;  '$var_functor'(T) ),
    T =.. [F|P],
    ip_lookup('$tr_prog_vars', PV),
    \+member_eq(F, PV), !,
    '$expand_funs'(F, FP, CF),
    '$expand_funs'(P, TP, CP),
    '$list2tuple'(TP, PTuple),
    C1 = apply(FP, PTuple, X),
    append(CF, CP, CFP),
    append(CFP, [C1], C).
'$expand_funs'(T, X, C) :-
    T =.. [F|P],
    '$type_info'(F, function, (_ -> Types), '$function', _),
    '$higher_order_type'(Types),
    !,
    '$expand_funs'(F, FP, CF),
    '$expand_funs'(P, TP, CP),
    '$list2tuple'(TP, PTuple),
    C1 = apply(FP, PTuple, X),
    append(CF, CP, CFP),
    append(CFP, [C1], C).
'$expand_funs'(T, X, C) :-
    T =.. [F|P],
    '$type_info'(F, function, (_ -> _Types), '$function', _),
    !,
    '$expand_funs'(F, FP, CF),
    '$expand_funs'(P, TP, CP),
    append(TP, [X], P1),
    C1 =.. [FP|P1],
    append(CF, CP, CFP),
    append(CFP, [C1], C).
'$expand_funs'(T, X, C) :-
    list(T), !,
    T = [Head|Tail],
    '$expand_funs'(Head, HeadX, HeadC),
    '$expand_funs'(Tail, TailX, TailC),
    X = [HeadX|TailX],
    append(HeadC, TailC, C).
'$expand_funs'(T, X, C) :-
    compound(T), !,
    T =.. [F|P],
    '$expand_funs'(F, FP, CF),
    '$expand_funs'(P, TP, CP),
    X =.. [FP|TP],
    append(CF, CP, C).
 '$expand_funs'(T, _X, _C) :-
     errornl(failed_expand_funs(T)).

/*
%% Transform individual subterms
%%'$trans_fun'(T, X) :- errornl('$trans_fun'(T, X)),fail.
'$trans_fun'(T, X) :-
    atomic(T), !, X = T.
'$trans_fun'(T, X) :-
    string(T), !, X = T.
'$trans_fun'(T, X) :-
    var(T), !, X = T.
%% for a list constructor the goal is unfolded and then a findall is built
'$trans_fun'(T, X) :-
    T = '$list_constr'(V, G), !,
    ip_lookup(calls, _C),
    errornl(v___(_C)),
    ip_set(calls, []),
    freeze_term(X),
    '$translate_goal'(G,B),
    ip_set(calls, []),
    transform_subterms('$trans_fun', V, V1),
    ip_lookup(calls, C2),
    errornl(o___(G, B, C2)),
    (
      C2 = []
    ->
      ip_set(calls, [findall(V1, B, X)])
    ;
      '$list2tuple'(C,CT),
      ip_set(calls, [findall(V1, (B,C2), X)])
    ).
%% similar to above but the result is sorted and wrapped in a '$set'
'$trans_fun'(T, X) :-
    T = '$set_constr'(V, G), !,
     ip_lookup(calls, C),
    freeze_term(X),
    '$translate_goal'(G,B),
    (
      C = []
    ->
      ip_set(calls, [findall(V, B, Y), sort(Y, X1), X = '$set'(X1)])
    ;
      '$list2tuple'(C,CT),
      ip_set(calls, [findall(V, (B, CT), Y), sort(Y, X1), X = '$set'(X1)])
    ).
'$trans_fun'(T, X) :-
    T = '$set_enum'(B), !,
    ip_lookup(calls, C),
    freeze_term(X),
    append(C, [sort(B, XL), X = '$set'(XL)], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, X) :-
    T = '$list2set'(B), !,
    ip_lookup(calls, C),
    freeze_term(X),
    append(C, [sort(B, XL), X = '$set'(XL)], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, X) :-
    T = '$set2list'(B), !,
    ip_lookup(calls, C),
    freeze_term(X),
    append(C, [B = '$set'(X)], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, X) :-
    T = (L1 <> L2), !,
    ip_lookup(calls, C),
    freeze_term(X),
    append(C, [append(L1, L2, X)], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, X) :-
    T = (L1 ++ L2), !,
    ip_lookup(calls, C),
    freeze_term(X),
    append(C, [string_concat(L1, L2, X)], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, X) :-
    T = '$'(A), !,
    ip_lookup(calls, C),
    append(C, [A(X)], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, X) :-
    ('$higher_order_term'(T) ;  '$var_functor'(T) ),
    T =.. [F|P],
    ip_lookup('$tr_prog_vars', PV),
    \+member_eq(F, PV), !,
    '$list2tuple'(P, PTuple),
    C1 = apply(F, PTuple, X),
    freeze_term(X),
    ip_lookup(calls, C),
    append(C, [C1], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, X) :-
    T =.. [F|P],
    '$type_info'(F, function, (_ -> Types), '$function', _),
    '$higher_order_type'(Types),
    !,
    '$list2tuple'(P, PTuple),
    C1 = apply(F, PTuple, X),
    freeze_term(X),
    %'$expand_out_arith'(C1, C2, IS),
    ip_lookup(calls, C),
    %append(IS, [C2], R),
    append(C, [C1], Calls),
    ip_set(calls, Calls).

'$trans_fun'(T, X) :-
    T =.. [F|P],    
    append(P, [X], P1),
    C1 =.. [F|P1],
    '$type_info'(F, function, (_ -> Types), '$function', _),
    \+'$higher_order_type'(Types),
    !,
    %'$expand_out_arith'(C1, C2, IS),
    ip_lookup(calls, C),
    %append(IS, [C2], R),
    append(C, [C1], Calls),
    ip_set(calls, Calls).
'$trans_fun'(T, T).

'$expand_out_arith'(Call, ACall, AIS) :-
    Call =.. [P|Args],
    '$expand_out_arith_aux'(Args, AArgs, AIS),
    ACall  =.. [P|AArgs].

'$expand_out_arith_aux'([], [], []).
'$expand_out_arith_aux'([A|Args], [AA|AArgs], AIS) :-
    (
     nonvar(A),fail %'$arith'(A)
    ->
     AA = X,
     AIS = [(X is A)|AIS1]
    ;
     AA = A,
     AIS = AIS1
    ),
    '$expand_out_arith_aux'(Args, AArgs, AIS1).
*/

%% T is a compound term whose functor is a variable
'$var_functor'(T) :-
    compound(T),
    functor(T, F, _),
    var(F).

/*
'$builtin_function'('<>'(L1, L2, X), [append(L1, L2, X)]).
'$builtin_function'('++'(S1, S2, X), [string_concat(S1, S2, X)]).
'$builtin_function'('$list2set'(T, X), [sort(T,Y), X = '$set'(Y)]).
'$builtin_function'('$set2list'(T, X), [T = '$set'(X)]).
'$builtin_function'('$list_constr'(T, Cond, X), [findall(T,Cond,X)]).
'$builtin_function'('$set_constr'(T, Cond, X),
                    [findall(T,Cond,X1), sort(X1, X2), X = '$set'(X2)]).
'$builtin_function'('#'(T, X), [length__(T,X)]).



'$get_resource_info'([], [], []).
'$get_resource_info'([A|Args], [T|Types], Res) :-
    ( T = _(T1) ; T = T1 ),!,
    '$get_resource_info'(Args, Types, Res1),
    (
      includes(resources, T1)
    ->
      Res = [A|Res1]
    ;
      Res = Res1
    ).
*/
/*
%% assert the head pattern for a task_atomic program
'$assert_atomic_pattern'(U, Named) :-
    '$decl_kind'(U, M, procedure),
    !,
    functor(Term, U, M),
    assert('$task_atomic'(Term, Named)).
'$assert_atomic_pattern'(U, Named) :-
    '$get_map'(task_atomic(U, Named), R),
      writeL(stderr,
                      ["Error: The task_atomic declaration ", R,
                       " has no corresponding type declaration ", nl_]),
      thread_exit.
*/


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%  Run checks on user programs
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/*
'$run_checks' :-
    errornl(check_valid_terms),
    global_state_set(num_errors, 0),
    ( '$check_valid_terms',fail ; true ),
    global_state_lookup(num_errors, N),
    (
      N > 0
    ->
      writeL(stderr, [nl_, nl_, N,
                               ' term validation errors detected - exiting',
                               nl_,nl_]),
      thread_exit
    ;
      fail
    ).    
 
'$run_checks' :-
    '$check_valid_types',
    '$check_definitions',
    fail.
'$run_checks'.


'$run_tr_checks' :-
    %% handle_percepts must be defined and be of the right type
    (
      '$user_type_info'(handle_percepts_, Kind, _Type, _Term, _Vars)
    ->
      (
        Kind \= action
      ->
        '$display_error'(handle_percepts_wrong_kind)
      ;
        true
     )
    ;
      true
    ),
    (
      '$should_suppress_show_type'(mod_controls),
      '$user_type_info'(mod_controls, MCKind, MCType, _MCTerm, MCVars)
    ->
      true
    ;
      (
        MCKind \= relation
      ->
        '$display_error'(controls_wrong_kind)
      ;
        MCType = ['!'('$list_type'(control_action)), '$list_type'(control_action)]
      ->
        true
      ;
        MCType = ['!'('$list_type'(control_action)),
                '?'('$list_type'(control_action))]
      ->
        true
      ;
       '$display_error'(controls_wrong_type(MCType, MCVars))
      )
    ),
    (
      '$user_type_info'(handle_message, MKind, MType, _MTerm, MVars)
    ->
      (
        MKind \= action
      ->
        '$display_error'(handle_message_wrong_kind)
      ;
        MType = ['+?'(term), '!'(handle)]
      ->
        true
      ;
        MType = ['+?'(term), handle]
      ->
        true
      ;
       '$display_error'(handle_message_wrong_type(MType, MVars))
      )
    ;
      true
    ).

*/

%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Check that the types used in declarations and type definitions are OK
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
'$check_valid_types' :-
    once('$current_consulted_file'(File)),
    '$defined_type'(TName, T),
    \+'$resource_info'(T, _, _), % TODO - is this necessary?
    '$saved_input_with_vars'((TName ::=  T), VarNames, File),
    ip_set(valid_type_context, valid_type_check((TName ::=  T), VarNames)),
    '$is_a_valid_type'(T),
    fail.
'$check_valid_types' :-
    '$relation_type_info'(F, Types, _),
    '$saved_input_with_vars'((F :  Decl), VarNames, _),
    member(T, Types),
    ip_set(valid_type_context, valid_type_check((F :  Decl), VarNames)),
    '$is_a_valid_type'(T),
    fail.
'$check_valid_types' :-
    '$user_type_info'(F, RelAct, Types, _, _),
    '$saved_input_with_vars'((F :  Decl), VarNames, _),
    ( RelAct = action ; RelAct = function),
    member(T, Types),
    ip_set(valid_type_context, valid_type_check((F :  Decl), VarNames)),
    '$is_a_valid_type'(T),
    fail.
'$check_valid_types' :-
    '$belief_type_info'(_, Types, BelClause),
    '$saved_input_with_vars'( belief(BelClause), VarNames, _),
    member(T, Types),
    ip_set(valid_type_context, valid_type_check(belief(BelClause), VarNames)),
    '$is_a_valid_type'(T),
    fail.
'$check_valid_types' :-
    '$percept_type_info'(_, Types, BelClause),
    '$saved_input_with_vars'( percept(BelClause), VarNames, _),
    member(T, Types),
    ip_set(valid_type_context, valid_type_check(percept(BelClause), VarNames)),
    '$is_a_valid_type'(T),
    fail.
'$check_valid_types' :-
    '$user_type_info'(F, _RelAct, Types, _, _),
    '$saved_input_with_vars'((F :  Decl), VarNames, _),
    member(T, Types),
    ip_set(valid_type_context, valid_type_check((F :  Decl), VarNames)),
    '$is_a_valid_type'(T),
    fail.
'$check_valid_types' :-
    '$user_type_info'(_, defined_type, TName, '$enum_type'(T), _),
    sort(T, TS),
    member(X, TS),
    once((delete(X, T, TX), member(X, TX))),
    '$display_error'(repeated_enum_type(TName, X)).
'$check_valid_types' :-
    '$user_type_info'(_, defined_type, TName, '$constr_enum_type'(T), _),
    findall(F, (member(E, T), functor(E, F, _)), Functors),
    sort(Functors, FunctorsS),
    member(X, FunctorsS),
    once((delete(X, Functors, FunctorsX), member(X, FunctorsX))),
    '$display_error'(repeated_constr_enum_type(TName, X)).
    
%% check for overlaps in defined enum types
'$check_valid_types' :-
    '$user_type_info'(_, defined_type, TName, '$enum_type'(T), _),
    '$check_for_type_enum_overlaps'(TName, T),
    fail.
'$check_valid_types' :-
    '$user_type_info'(_, defined_type, TName, '$constr_enum_type'(T), _),
    '$check_for_type_constr_enum_overlaps'(TName, T),
    fail.
%% See if atom in enum type is used as a type or as a constructor in enum type
'$check_valid_types' :-
    '$user_type_info'(_, defined_type, TName, '$enum_type'(Enum), _),
    member(A, Enum),
    (
     '$user_type_info'(A, _, _, _, _)
    ->
     '$display_error'(enum_is_type(A, TName))
    ;
     true
    ),
    '$defined_type'(Constr, '$constr_enum_type'(ConstrList)),
    (
     member(C, ConstrList), functor(C, A, _)
    ->
     '$display_error'(enum_is_constr_enum(A, TName, Constr))
    ;
     true
    ),
    fail.
'$check_valid_types' :-
    '$user_type_info'(_, defined_type, TName, '$constr_enum_type'(Enum), _),
    member(A, Enum),
    '$real_functor'(A, F, _),
    (
     '$user_type_info'(F, _, _, _, _)
    ->
     '$display_error'(constr_enum_is_type(A, TName))
    ;
     true
    ),
    '$defined_type'(Constr, '$constr_enum_type'(ConstrList)),
    TName \= Constr,
    member(C, ConstrList),
    '$real_functor'(C, F, _),
    (
     '$not_identical_mod_vars'(A, C),
     \+type_error_remember(overlapping_constructor__(F))
    ->
     assert(type_error_remember(overlapping_constructor__(F))),
     '$display_error'(constr_enum_is_constr_enum(A, TName, Constr))
    ;
     true
    ),
    fail.
'$check_valid_types' :-
    '$defined_type'(TName, (A..B)),
    '$check_for_type_range_overlaps'(TName, A, B),
    fail.
/*
'$check_valid_types' :-
    '$user_type_info'(_, function, (T1 -> T2), _),
    '$saved_input_with_vars'((F :  (T1 -> T2)), VarNames),
    ip_set(valid_type_context, valid_type_check((F :  (T1 -> T2)), VarNames)),
    '$is_a_valid_type'((T1 -> T2)),
    fail.
*/
'$check_valid_types'.

'$check_for_type_enum_overlaps'(TName, Enum) :-
    Enum = [N|_],
    integer(N),
    '$defined_type'(RName, (A..B)),
    once((
	  member(M, Enum),
	  between(A, B, M)
	 )),
    once((
	  member(M1, Enum), \+ between(A, B, M1),
	  between(A, B, M2), \+ member(M2, Enum)
	 )),
    (
     type_error_remember(overlapping_types(RName, TName))
    ->
     fail
    ;
     assert(type_error_remember(overlapping_types(TName, RName))),
     '$display_error'(overlapping_types(TName, RName))
    ).
'$check_for_type_enum_overlaps'(TName, Enum) :-
    '$defined_type'(RName, '$enum_type'(OtherEnum)),
    TName \= RName,
    once(( member(M, Enum), member(M, OtherEnum) )),
    once((
	  member(M1, Enum), \+ member(M1, OtherEnum),
	  member(M2, OtherEnum), \+ member(M2, Enum)
	 )),
    (
     type_error_remember(overlapping_types(RName, TName))
    ->
     fail
    ;
     assert(type_error_remember(overlapping_types(TName, RName))),
     '$display_error'(overlapping_types(TName, RName))
    ).

'$check_for_type_constr_enum_overlaps'(TName, Enum) :-
    '$defined_type'(RName, '$constr_enum_type'(OtherEnum)),
    TName \= RName,
    once(( member(M, Enum), member(M, OtherEnum) )),
    once((
	  member(M1, Enum), \+ member(M1, OtherEnum),
	  member(M2, OtherEnum), \+ member(M2, Enum)
	 )),
    (
     type_error_remember(overlapping_types(RName, TName))
    ->
     fail
    ;
     assert(type_error_remember(overlapping_types(TName, RName))),
     '$display_error'(overlapping_types(TName, RName))
    ).


'$check_for_type_range_overlaps'(TName, A, B) :-
    '$defined_type'(RName, (A1..B1)),
    A1 > A, A1 =< B,
    B1 > B,
    (
     type_error_remember(overlapping_types(RName, TName))
    ->
     fail
    ;
     assert(type_error_remember(overlapping_types(TName, RName))),
     '$display_error'(overlapping_types(TName, RName))
    ).

/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Check that the dynamicB decls are OK
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%
'$check_dynamicP' :-
    '$dynamicP'(P),
    (
     '$is_dynamicP_decl'(P)
    ->
     fail
    ;
     writeL(stderr, ["Error: Invalid dynamicP declaration ",
			      dynamicP(P), nl_]),
     thread_exit
    ).
*/
'$update_macros' :-
    '$user_type_info'(A, defined_type, B, '$type_macro'(T), VarNames),
    \+'$is_a_type'(T),
    retract('$user_type_info'(A, defined_type, B, '$type_macro'(T), VarNames)),
    (
      compound(T)
    ->
      assert('$user_type_info'(A, defined_type, B, '$constr_enum_type'([T]),
                               VarNames))
    ;
      assert('$user_type_info'(A, defined_type, B, '$enum_type'([T]), VarNames))
    ),
    fail.
'$update_macros'.


'$update_unknown_kinds' :-
    once('$current_consulted_file'(File)),
    '$saved_input_with_vars'(Clause, Vars, File),
    '$update_unknown_kinds_aux'(Clause, Vars),
    fail.
'$update_unknown_kinds' :-
    retract('$user_type_info'(A, unknown, Types, Head, V)),
    assert('$user_type_info'(A, relation, Types, Head, V)),fail.
'$update_unknown_kinds'.


'$update_unknown_kinds_aux'(Clause, Vars) :-
     Clause = (Head :: _Test <= _Body), !,
    check_against_declaration(Head, Clause, Vars, relation).
'$update_unknown_kinds_aux'(Clause, Vars) :-
     Clause = (Head <= _Body), !,
    check_against_declaration(Head, Clause, Vars, relation).
'$update_unknown_kinds_aux'(Clause, Vars) :-
     Clause = (Head :: _Test -> _Body), !,
    check_against_declaration(Head, Clause, Vars, function).
'$update_unknown_kinds_aux'(Clause, Vars) :-
     Clause = (Head -> _Body), !,
    check_against_declaration(Head, Clause, Vars, function).
'$update_unknown_kinds_aux'(Clause, Vars) :-
     Clause = (Head :: _Test ~>> _Body), !,
    check_against_declaration(Head, Clause, Vars, action).
'$update_unknown_kinds_aux'(Clause, Vars) :-
     Clause = (Head ~>> _Body), !,
    check_against_declaration(Head, Clause, Vars, action).
'$update_unknown_kinds_aux'(Clause, Vars) :-
     Clause = '$tr_prog'(Head, _), !,
    check_against_declaration(Head, Clause, Vars, tr_prog).

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Check definitions
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% This tests the validity/types/modes of definitions of
%% relations, functions and TR programs.

'$check_definitions' :-
    once('$current_consulted_file'(File)),
    '$saved_input_with_vars'(Clause, Vars, File),
    '$check_definitions_aux'(Clause, Vars),
    fail.
'$check_definitions'.

'$check_definitions_aux'((_A ::= _AT), _Vars) :- !.
'$check_definitions_aux'(belief(_), _) :- !.
'$check_definitions_aux'(percept(_), _) :- !.
'$check_definitions_aux'(discrete(_), _) :- !.
'$check_definitions_aux'(durative(_), _) :- !.
'$check_definitions_aux'(resources(_), _) :- !.
'$check_definitions_aux'(task_start(_), _) :- !.
'$check_definitions_aux'(task_atomic(_), _) :- !.
'$check_definitions_aux'(global_int_decl(_,_), _) :- !.
'$check_definitions_aux'(global_num_decl(_,_), _) :- !.
'$check_definitions_aux'((_:_), _Vars) :- !.
'$check_definitions_aux'((?- _), _Vars) :- !.
'$check_definitions_aux'(Clause, Vars) :-
    Clause = '$tr_prog'(_A, _B), !,
    once('$check_tr_prog'(Clause, Vars)).
'$check_definitions_aux'(Clause, Vars) :-
    Clause = (_Head :: _Test -> _Body), !,
    %%errornl(checking_function(Clause, Vars)),
    ip_set(template_constraints, []),
    once('$check_function'(Clause, Vars)).
'$check_definitions_aux'(Clause, Vars) :-
    Clause = (_Head -> _Body), !,
    %%errornl(checking_function(Clause, Vars)),
    ip_set(template_constraints, []),
    once('$check_function'(Clause, Vars)).
'$check_definitions_aux'(Clause, Vars) :-
    Clause = (_ :: _ ~>> _), !,
    ip_set(template_constraints, []),
    once('$check_action'(Clause, Vars)).
'$check_definitions_aux'(Clause, Vars) :-
    Clause = ( _ ~>> _), !,
    ip_set(template_constraints, []),
    once('$check_action'(Clause, Vars)).
'$check_definitions_aux'(Clause, Vars) :-
    %errornl(checking_relation(Clause)),
    ip_set(template_constraints, []),
    once('$check_relation'(Clause, Vars)).

'$check_relation'(Clause, VarNames) :-
    Clause = (Head :: Test <= Body),
    !,
    '$check_relation_aux'(Head, Test, Body, Clause, VarNames).
'$check_relation'(Clause, VarNames) :-
    Clause = (Head <= Body),
    !,
    '$check_relation_aux'(Head, true, Body, Clause, VarNames).
'$check_relation'(Clause, VarNames) :-
    Clause = (Head :: Test),
    !,
    '$check_relation_aux'(Head, Test, true, Clause, VarNames).
'$check_relation'(Clause, VarNames) :-
    Clause = Head,
    !,
    '$check_relation_aux'(Head, true, true, Clause, VarNames).

'$check_ultimate_functor'(Head, Clause, VarNames) :-
    '$get_ultimate_functor'(Head, F),
    (
      atom(F)
    ->
      true
    ;
      '$display_error'(non_atom_functor(F, Head, Clause, VarNames))
    ).

'$check_relation_aux'(Head, Test, Body, Clause, VarNames) :-
    %global_state_set('$error_level', 1),
    freeze_term(Clause),
    '$check_ultimate_functor'(Head, Clause, VarNames),
    '$check_head_for_evaluation'(Head, Clause, VarNames),
    '$check_goal_expressions'(Test, Clause, VarNames, rel),
    '$check_goal_expressions'(Body, Clause, VarNames, rel),
    %thaw_term(XFVars),
    FullBody = (Test & Body),
    '$initialize_copy_and_save_context'(Clause, VarNames,
                                        FullBody, FullBodyCopy, relation),
    ip_set(type_checker_context, relation(Clause, VarNames)),
    once(
         (
           manage_errors1,
           check_clause(Head, FullBody, FullBodyCopy),
           manage_errors2
         ;
           true
         )
        ).

'$check_action'(Clause, VarNames) :-
    Clause = (Head :: Test ~>> Body),
    !,
    '$check_action_aux'(Head, Test, Body, Clause, VarNames).
'$check_action'(Clause, VarNames) :-
    Clause = (Head ~>> Body),
    !,
    '$check_action_aux'(Head, true, Body, Clause, VarNames).
'$check_action'(Clause, VarNames) :-
    Clause = (Head :: Test),
    !,
    '$check_action_aux'(Head, Test, true, Clause, VarNames).
'$check_action'(Clause, VarNames) :-
    Clause = Head,
    !,
    '$check_action_aux'(Head, true, true, Clause, VarNames).

'$check_action_aux'(Head, Test, Body, Clause, VarNames) :-
    freeze_term(Clause),
    '$check_head_for_evaluation'(Head, Clause, VarNames),
    '$check_goal_expressions'(Test, Clause, VarNames, any),
    '$check_goal_expressions'(Body, Clause, VarNames, any),
    %thaw_term(XFVars),
    FullBody = (Test & Body),
    '$initialize_copy_and_save_context'(Clause, VarNames,
                                        FullBody, FullBodyCopy, any),
    ip_set(type_checker_context, action(Clause, VarNames)),
    once(
         (
           manage_errors1,
           check_clause(Head, FullBody, FullBodyCopy),
           manage_errors2
         ;
           true
         )
        ).


check_against_declaration(Head, _Clause, _VarNames, function) :-
    '$get_ultimate_functor'(Head, A),
    '$type_info'(A, HeadKind1, Types1, Head1, _),
    Head1 \= Head,
    Types1 = [Def], '$type_info'(_, defined_type, Def, DefType, _),
    DefType = '$type_macro'(Types), Types = (_ -> _), !,
    retract('$user_type_info'(A, HeadKind1, Types1, Head1, V)),
    assert('$user_type_info'(A, function, Types, '$function', V)).
check_against_declaration(Head, Clause, VarNames, Kind) :-
    '$get_ultimate_functor'(Head, A),
    Kind \== function,
    '$type_info'(_, HeadKind, Types, Head, _), !,
    (
      HeadKind = unknown
    ->
     functor(Head, P, N),
     functor(CHead, P, N),
      retract('$user_type_info'(A, HeadKind, Types, _, V)),
      assert('$user_type_info'(A, Kind, Types, CHead, V))
    ;
      HeadKind = Kind
    ->
      true
    ;
     HeadKind = function, Kind = relation,
     functor(Head, F, _), compound(F)
    ->
     true
    ;
     '$display_error'(kind_mismatch(Clause, VarNames)),fail
    ).
check_against_declaration(_Head, _Clause, _VarNames, _Kind).
    

%%'$check_function'(Clause, VarNames) :- errornl('$check_function'(Clause, VarNames)),fail.
'$check_function'(Clause, VarNames) :-
    Clause = (Head :: Test -> Body), !,
    %global_state_set('$error_level', 1),
    freeze_term(Clause),
    '$check_head_for_evaluation'(Head, Clause, VarNames),
    '$check_goal_expressions'(Test, Clause, VarNames, rel),
    '$check_expressions'(Body, Clause, VarNames),
    %thaw_term(XFVars),
    (
      '$get_function_type_info'(Head, _Args, _Types, _ResultType)
    ->
      true
    ;
      '$display_error'(undeclared_function(Clause, VarNames))
    ),
    FullBody = function(Test, Body),
    '$initialize_copy_and_save_context'(Clause, VarNames,
                                        FullBody, FullBodyCopy, function),
    once(
         (
           manage_errors1,
           check_function_clause(Head, FullBody, FullBodyCopy),
           manage_errors2
         ;
           true
	 )
        ).

'$check_function'(Clause, VarNames) :-
    Clause = (Head  -> Body), !,
    %global_state_set('$error_level', 1),
    freeze_term(Clause),
    '$check_head_for_evaluation'(Head, Clause, VarNames),
    '$check_expressions'(Body, Clause, VarNames),
    %thaw_term(XFVars),
    (
      '$get_function_type_info'(Head, _Args, _Types, _ResultType)
    ->
      true
    ;
      '$display_error'(mismatched_function_types(Clause, VarNames))
    ),
    FullBody = function(true, Body),
    '$initialize_copy_and_save_context'(Clause, VarNames,
                                        FullBody, FullBodyCopy, function),
    once(
         (
           manage_errors1,
           check_function_clause(Head, FullBody, FullBodyCopy),
           manage_errors2
         ;
           true
	 )
        ).

    
'$check_head_for_evaluation'(Head, Clause, VarNames) :-
    Head =.. [_P|Args],
    '$check_head_for_evaluation_aux'(Args, Clause, VarNames).

'$check_head_for_evaluation_aux'([], _, _).
'$check_head_for_evaluation_aux'([A|Args], Clause, VarNames) :-
    '$get_ultimate_functor'(A, F),
    (
      atom(F), '$type_info'(F, function, _, _, _)
    ->
      '$display_error'(eval_in_head(A, Clause, VarNames))
    ;
      true
    ),
    '$check_head_for_evaluation_aux'(Args, Clause, VarNames).

%%'$check_goal_expressions'(G, Clause, VarNames, RA) :-  errornl('$check_goal_expressions'(G, Clause, VarNames, RA)),fail.
'$check_goal_expressions'(true, _, _, _) :-
    %% special case
    !.

'$check_goal_expressions'(show, Clause, VarNames, _) :-
    !,
    '$display_error'(bad_show(show, Clause, VarNames)).
'$check_goal_expressions'(show(A), Clause, VarNames, _) :-
    !,
    '$display_error'(bad_show(show(A), Clause, VarNames)).

'$check_goal_expressions'(G, _Clause, _VarNames, _) :-
    compound(G), functor(G, F, _), var(F), !.
'$check_goal_expressions'(G, Clause, VarNames, _) :-
    '$type_info'(_, discrete, _, G, _), !,
    '$display_error'(bad_discrete_goal(G, Clause, VarNames)).
'$check_goal_expressions'(G, Clause, VarNames, _) :-
    '$type_info'(_, durative, _, G, _), !,
    '$display_error'(bad_durative_goal(G, Clause, VarNames)).
'$check_goal_expressions'(G, Clause, VarNames, _) :-
    '$type_info'(_, tr_prog, _, G, _), !,
    '$display_error'(bad_tr_prog_goal(G, Clause, VarNames)).
'$check_goal_expressions'((G1 & G2), Clause, VarNames, RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, RA),
    '$check_goal_expressions'(G2, Clause, VarNames, RA).
'$check_goal_expressions'((G1 ; G2), Clause, VarNames, RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, RA),
    '$check_goal_expressions'(G2, Clause, VarNames, RA).
'$check_goal_expressions'(not(G1), Clause, VarNames, _) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, rel).
'$check_goal_expressions'(once(G1), Clause, VarNames, RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, RA).
'$check_goal_expressions'(call(G1), Clause, VarNames, RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, RA).
'$check_goal_expressions'(do(G1), Clause, VarNames, RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, RA).
'$check_goal_expressions'('$exists'(_, G1), Clause, VarNames, _RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, rel).
'$check_goal_expressions'('$at_forall'(_, G1, G2), Clause, VarNames, _RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, rel),
    '$check_goal_expressions'(G2, Clause, VarNames, rel).
'$check_goal_expressions'('$at_forall_actions'(_, G1, G2),
                          Clause, VarNames, RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, RA),
    '$check_goal_expressions'(G2, Clause, VarNames, RA).
'$check_goal_expressions'('$forall'(_, G1, G2), Clause, VarNames, _RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, rel),
    '$check_goal_expressions'(G2, Clause, VarNames, rel).
'$check_goal_expressions'('$forall_actions'(_, G1, G2), Clause, VarNames, RA) :-
    !,
    '$check_goal_expressions'(G1, Clause, VarNames, RA),
    '$check_goal_expressions'(G2, Clause, VarNames, RA).
'$check_goal_expressions'((G2 '=?' G1), Clause, VarNames, _) :-
    !,
    '$check_expressions'(G1, Clause, VarNames),
    '$check_expressions'(G2, Clause, VarNames).
'$check_goal_expressions'((G1 = G2), Clause, VarNames, _) :-
    !,
    '$check_expressions'(G1, Clause, VarNames),
    '$check_expressions'(G2, Clause, VarNames).
'$check_goal_expressions'(remember(G), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        ( var(G) ; functor(G, F, _), var(F) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        G = A(_), '$state_belief'(A)
      ->
        '$display_error'(bad_belief_global_goal(A, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel(remember(G), Clause, VarNames))
    ).
'$check_goal_expressions'(remember_for(G,T), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        ( var(G) ; functor(G, F, _), var(F) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        G = A(_), '$state_belief'(A)
      ->
        '$display_error'(bad_belief_global_goal(A, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel(remember_for(G,T), Clause, VarNames))
    ).
'$check_goal_expressions'(rememberA(G), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        ( var(G) ; functor(G, F, _), var(F) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel(rememberA(G), Clause, VarNames))
    ).
'$check_goal_expressions'(rememberA_for(G,T), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        ( var(G) ; functor(G, F, _), var(F) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel(rememberA_for(G,T), Clause, VarNames))
    ).
'$check_goal_expressions'(forget(G), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        ( var(G) ; functor(G, F, _), var(F) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        G = A(_), '$state_belief'(A)
      ->
        '$display_error'(bad_belief_global_goal(A, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel(forget(G), Clause, VarNames))
    ).
'$check_goal_expressions'(forget_after(G,T), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        ( var(G) ; functor(G, F, _), var(F) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        G = A(_), '$state_belief'(A)
      ->
        '$display_error'(bad_belief_global_goal(A, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel(forget_after(G,T), Clause, VarNames))
    ).
'$check_goal_expressions'(replace_by(G1, G2), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        ( var(G1) ; functor(G1, F1, _), var(F1) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        G1 = A(_), '$state_belief'(A)
      ->
        '$display_error'(bad_belief_global_goal(A, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G1)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G1, Clause, VarNames))
      ),
      (
        ( var(G2) ; functor(G2, F2, _), var(F2) )
      ->
        '$display_error'(bad_belief_goal(G, Clause, VarNames))
      ;
        G2 = B(_), '$state_belief'(B)
      ->
        '$display_error'(bad_belief_global_goal(B, Clause, VarNames))
      ;
        '$dynamic_type_info'(_, _, G2)
      ->
        true
      ;
        '$display_error'(bad_belief_goal(G2, Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel(replace_by(G1, G2), Clause, VarNames))
    ).
'$check_goal_expressions'(':='(A, B), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        '$state_belief'(A)
      ->
        true
      ;
        '$display_error'(not_a_global(A, ':='(A, B), Clause, VarNames))
      )
     ;
      '$display_error'(action_in_rel(':='(A, B), Clause, VarNames))
    ).
'$check_goal_expressions'('+:='(A, B), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        '$state_belief'(A)
      ->
        true
      ;
        '$display_error'(not_a_global(A, '+:='(A,B), Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel('+:='(A,B), Clause, VarNames))
    ).
'$check_goal_expressions'('-:='(A, B), Clause, VarNames, RA) :-
    !,
    (
      RA = any
    ->
      (
        '$state_belief'(A)
      ->
        true
      ;
        '$display_error'(not_a_global(A, '-:='(A,B), Clause, VarNames))
      )
    ;
      '$display_error'(action_in_rel('-:='(A,B), Clause, VarNames))
    ).
    
'$check_goal_expressions'(ground(G), Clause, VarNames, _) :-
    !,
    '$check_expressions'(G, Clause, VarNames).
'$check_goal_expressions'(template(G), Clause, VarNames, _) :-
    !,
    '$check_expressions'(G, Clause, VarNames).
'$check_goal_expressions'(type(_G1, G2), Clause, VarNames, _) :-
    !,
    ip_set(valid_type_context, valid_type_check(Clause, VarNames)),
    '$is_a_valid_type'(G2).
'$check_goal_expressions'(typeG(G1, G2), Clause, VarNames, _) :-
    !,
    (
     var(G1)
    ->
     true
    ;
     '$display_error'(must_be_variable(G1, typeG(G1, G2), Clause, VarNames))
    ),
    ip_set(valid_type_context, valid_type_check(Clause, VarNames)),
    '$is_a_valid_type'(G2).
'$check_goal_expressions'(isa(G1, G2), Clause, VarNames, _) :-
    !,
    '$check_expressions'(G1, Clause, VarNames),
    ip_set(valid_type_context, valid_type_check(Clause, VarNames)),
    '$is_a_valid_type'(G2).

/*
   removed all references to is_a_type_test
  '$check_goal_expressions'(G, Clause, VarNames, _) :-
    '$is_a_type_test'(G, Term, Type),
    errornl('$is_a_type_test'(G, Term, Type)),
    ground(Type),
    !,
    (
     var(Term)
    ->
     true
    ;
     '$display_error'(must_be_variable(Term, G, Clause, VarNames))
    ),
    ip_set(valid_type_context, valid_type_check(Clause, VarNames)),
    '$is_a_valid_type'(Type).
*/

'$check_goal_expressions'(type(G1, G2, _), Clause, VarNames, _) :-
    !,
    '$check_expressions'(G1, Clause, VarNames),
    ip_set(valid_type_context, valid_type_check(Clause, VarNames)),
    '$is_a_valid_type'(G2).
'$check_goal_expressions'(G, Clause, VarNames, _) :-
    '$type_info'(_, relation, _, G, _), !,
    '@..rev'(G, _F, Args),
    %G =.. [_F|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_goal_expressions'(G, Clause, VarNames, RA) :-
    '$type_info'(_, action, Types, G, _), !,
    (
      RA = any
    ->
    '@..rev'(G, _F, Args),
      %G =.. [_F|Args],
      '$check_action_expressions'(Args, Types, Clause, VarNames)
    ;
      '$display_error'(action_in_rel(G, Clause, VarNames))
    ).
'$check_goal_expressions'(G, _Clause, _VarNames, _) :-
    '$type_info'(_, function, _, G, _), !.
'$check_goal_expressions'(G, Clause, VarNames, _) :-
    compound(G), functor(G, F, _),
    '$type_info'(F, function, _, _, _), !,
    '$display_error'(expression_in_relact(G, Clause, VarNames)).
'$check_goal_expressions'(G, Clause, VarNames, _) :-
    atomic(G), !,
    '$display_error'(expression_in_relact(G, Clause, VarNames)).
'$check_goal_expressions'(G, _Clause, _VarNames, _) :-
    var(G), !.
'$check_goal_expressions'(G, _Clause, _VarNames, _) :-
    '$get_ultimate_functor'(G, F),
    var(F), !.
'$check_goal_expressions'([_|_], _Clause, _VarNames, _) :- !.
    
'$check_goal_expressions'(Term, Clause, VarNames, _) :-
    '$display_error'(bad_goal(Term, Clause, VarNames)).

%%'$check_action_expressions'(A,B,C,D) :- errornl('$check_action_expressions'(A,B,C,D)),fail.
'$check_action_expressions'([], [], _Clause, _VarNames).
'$check_action_expressions'([A|Args], [T|Types], Clause, VarNames) :-
    T = actcall,  !,
    '$check_goal_expressions'(A, Clause, VarNames, any),
    '$check_action_expressions'(Args, Types, Clause, VarNames).
'$check_action_expressions'([A|Args], [T|Types], Clause, VarNames) :-
    ( T = relcall ; T = qcall ), !,
    '$check_goal_expressions'(A, Clause, VarNames, rel),
    '$check_action_expressions'(Args, Types, Clause, VarNames).
'$check_action_expressions'([A|Args], [_T|Types], Clause, VarNames) :-
    '$check_expressions'(A, Clause, VarNames),
    '$check_action_expressions'(Args, Types, Clause, VarNames).

%%'$check_expressions'(G, C, V) :- errornl('$check_expressions'(G, C, V)), fail.
/*
  '$check_expressions'(Type, Clause, VarNames) :-
    '$defined_type'(Type, _), !,
    thaw_term((Clause, VarNames)),
    '$vars2names'(Clause, VarNames),
    '$display_error'(type_as_term_error(Type, Clause)),
    fail. %process_type_as_term_error(Type, Clause, VarNames).
*/
'$check_expressions'(G, _, _) :-
    \+compound(G), !.
'$check_expressions'('`'(_A), _Clause, _VarNames) :-
    !.
'$check_expressions'('$'(A), Clause, VarNames) :-
    !,
    (
      '$state_belief'(A)
    ->
      true
    ;
      '$display_error'(not_a_global(A, '$'(A), Clause, VarNames))
    ).

/*
  '$check_expressions'('::'(V, G), Clause, VarNames) :-
    !,
    '$check_expressions'(V, Clause, VarNames),
    '$check_goal_expressions'(G, Clause, VarNames, rel).
*/
'$check_expressions'(G, Clause, VarNames) :-
    '$type_info'(_, durative, _, G, _), !,
    G =.. [_F|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_expressions'(G, Clause, VarNames) :-
    '$type_info'(_, discrete, _, G, _), !,
    G =.. [_F|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_expressions'(G, Clause, VarNames) :-
    '$type_info'(_, relation, _, G, _), !,
    G =.. [_F|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_expressions'(G, Clause, VarNames) :-
    '$type_info'(_, action, _, G, _), !,
    G =.. [_F|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_expressions'(G, Clause, VarNames) :-
    '$type_info'(_, tr_prog, _, G, _), !,
    G =.. [_F|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_expressions'(Expr, Clause, VarNames) :-
    Expr = F(_,_), F == '$list_constr',
    Expr = '$list_constr'(_,  G),
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel).
'$check_expressions'(Expr, Clause, VarNames) :-
    Expr = F(_,_), F == '$set_constr',
    Expr = '$set_constr'(_,  G),
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel).
'$check_expressions'(Expr, Clause, VarNames) :-
    Expr = F(_,_), F == '?',
    Expr = (_ ? G),
    !,
    '$check_goal_expressions'(G, Clause, VarNames, rel).
'$check_expressions'(Expr, Clause, VarNames) :-
    Expr = '$set'([H|T]),
    !,
    '$check_expressions'(H, Clause, VarNames),
    '$check_expressions'(T, Clause, VarNames).
'$check_expressions'(Expr, _Clause, _VarNames) :-
    Expr = '$set_enum'([]), !.
'$check_expressions'(Expr, Clause, VarNames) :-
    Expr = '$set_enum'([H|T]),
    !,
    collect_vars([H|T], Vars),
    (
      Vars = []
    ->
      true
    ;
      '$display_error'(expression_query_mode(Expr, VarNames, Vars, _)),
      fail
    ),
    '$check_expressions'(H, Clause, VarNames),
    '$check_expressions'(T, Clause, VarNames).
'$check_expressions'(Expr, Clause, VarNames) :-
    list(Expr), Expr = [H|T],
    !,
    '$check_expressions'(H, Clause, VarNames),
    '$check_expressions'(T, Clause, VarNames).
'$check_expressions'(Expr, Clause, VarNames) :-
    Expr = '$tuple'(Lst),
    !,
    '$check_expressions'(Lst, Clause, VarNames).

'$check_expressions'(Term, Clause, VarNames) :-
    '$type_info'(_, defined_type, _, Type, _),
    Type = '$type_macro'(TypeM),
    compound(TypeM),
    functor(Term, F, N),
    functor(TypeM, F, N),
    !,
    Term =.. [_|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_expressions'(Term, Clause, VarNames) :-
    compound(Term),
    '$real_functor'(Term, F, N),
    '$constructor_type_arity'(F, _), !,
    (
      '$constructor_type_arity'(F, N)
    ->
      true
    ;
      '$display_error'(wrong_num_args(Term, Clause, VarNames))
    ),
    Term =.. [_|Args],
    '$check_expressions'(Args, Clause, VarNames).
'$check_expressions'(Term, Clause, VarNames) :-
    (
      '$get_ultimate_functor'(Term, F), var(F)
    ->
      !
    ;
      (
        '$get_function_type_info'(Term, _Args, _Types, _)
      ->
       
        (
          '$get_function_type_info'(Term, Args, Types, _),
          map('$tuple_len', [Args, L]),
          map('$tuple_type_len', [Types, L]),
          map('$tuple_to_list', [Args, ArgsL])
        ->
          true
        ;
          !, '$display_error'(wrong_num_args(Term, Clause, VarNames))
        )
      ;
        fail
      ),
      !,
      '$check_expressions'(ArgsL, Clause, VarNames)
    ).
'$check_expressions'(Term, Clause, VarNames) :-
    compound(Term),
    functor(Term, P, _N),
    '$type_info'(P, function, (_T1 -> _), '$function', _), !,
    %'$tuple_to_list'(T1, T1L),
    %length(T1L, TLen),
    '$display_error'(already_declared(Term, Clause, VarNames)).
'$check_expressions'(Term, Clause, VarNames) :-
    compound(Term),
    functor(Term, P, _N),
    '$type_info'(P, _, _Types, _, _), !,
    %length(Types, TLen),
    '$display_error'(already_declared(Term, Clause, VarNames)).
    
'$check_expressions'(Term, Clause, VarNames) :-
    compound(Term), !,
    (
      thaw_term(Clause),
      '$vars2names'(Clause, VarNames),
      functor(Term, P, N),
      (
        Clause = '$tr'(Head, Rule)
      ->
        writeL(stderr, ["Warning: ", P/N,
                        " is undeclared in the rule", nl_]),
        write_a_tr_rule(stderr, Rule),
        writeL(stderr, [nl_, "of the TR procedure ", Head, nl_])
      ;
        writeL(stderr, ["Warning: ", P/N,
                        " is undeclared in", nl_,
                        show_(Clause),nl_])
      ),
      (
        '$get_ultimate_functor'(Term, R),
        '$type_info'(R, Kind, _, _, _)
      ->
        writeL(stderr, ["However, there is a ", Kind,
                        " declaration for ", R, nl_])
      ;
        true
      ),
      fail
    ;
      true
    ).
'$check_expressions'(Term, Clause, VarNames) :-
    '$display_error'(bad_expression(Term, Clause, VarNames)).
%'$check_expressions'(_Term, _Clause, _VarNames).



'$freeze_underscores'(T, Vars) :-
    collect_vars(T, TVars),
    freeze_term(Vars, FVars),
    freeze_term(TVars),
    thaw_vars(FVars).

'$get_underscores'(T, Vars, Unders) :-
    collect_vars(T, TVars),
    filter('$not_in_vars'(Vars), TVars, Unders).
    
'$not_in_vars'(Vars, X) :-
    member((A = _B), Vars),
    A == X,
    !,
    fail.
'$not_in_vars'(_Vars, _X).
   
'$lookForDefinedType'(UTypes,[atomic|UTypes2]) :-
    list(UTypes),
    delete(atom, UTypes, UTypes1),
    delete(num, UTypes1, UTypes2), !.

'$lookForDefinedType'(UTypes,[Type|RUTypes]) :-
    list(UTypes),
    '$defined_type'(Type, '$union_type'(OTypes)),
    forall( member(T,OTypes), member(T,UTypes)),
    !,
    findall(X, (member(X, UTypes), \+member(X, OTypes)), RUTypes).
'$lookForDefinedType'(UTypes,UTypes). 


'$initialize_copy_and_save_context'(Clause, VarNames, Body, BodyCopy, Kind) :-
    '$get_underscores'(Clause, VarNames, Unders),
    ip_set('$underscore_vars', Unders),
    copy_term((Clause, VarNames, Body), (ClauseCopy, VarNamesCopy, BodyCopy)),
    ip_set(type_checker_context, Kind(Clause, VarNames)),
    ip_set(type_checker_copy_context, Kind(ClauseCopy, VarNamesCopy)),
    ip_set(check_index_, 1),
    global_state_set(gs_check_index_, 1).
    





%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Support for term transformation
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'$expandL'(L,AL-(AL=L)):-
   (var(L) ; list(L)),
   !.
'$expandL'('$set_constr'(E, Conds),L-(findall(E,Conds,L1), sort(L1, L))):-
   !.
'$expandL'('$list_constr'(E, Conds),L-(findall(E,Conds,L))):-
   !.
'$expandL'(L1<>L2, AL-(Expands1,Expands2,append(AL1,AL2,AL))):-
    !,
  '$expandL'(L1, AL1-Expands1),
  '$expandL'(L2,AL2-Expands2).
'$expandL'(L,AL-(AL=L)).


'$expandS'(S,AS-(AS=S)):-
   (var(S) ; string(S)),
   !.
'$expandS'(S1 ++ S2, AS-(Expands1,Expands2,string_concat(AS1,AS2,AS))):-
  '$expandS'(S1, AS1-Expands1),
  '$expandS'(S2,AS2-Expands2).


 '$expandLP'(L, L1 <> L2 <> L3, Trans) :-
       !,
       (
	L1 = L11+[]?Cond
       ->
	Trans=( append(L11,RL,L),Cond,Apps)
       ;
        L1= L11?Cond
       ->
	Trans = (append(L11,RL,L),Cond,Apps)
       ;
	Trans= ( append(L1,RL,L),Apps)
       ),
       '$expandLP'(RL, L2 <> L3 , Apps).  

 '$expandLP'(L, (L1 <> L2), Trans) :-
       (
	(L1=L11?Cond1, L2=L22?Cond2)
       -> 
	%elist_test(L11,E1),
	%elist_test(L22,E2),
	Trans=(append(L11,L22,L),Cond1,Cond2)
       ;
	L1=L11?Cond1
       ->
	%elist_test(L11,E1),
	%elist_test(L2,E2),
	Trans=(append(L11,L2,L),Cond1)
       ;
        L2=L22:Cond2
       ->
	%elist_test(L1,E1),
	%elist_test(L22,E2),
	Trans=(append(L1,L22,L),Cond2)
       ;
	%elist_test(L1,E1),
	%elist_test(L2,E2),
        Trans=(append(L1,L2,L))
       ).
           


elist_test((_L+[]), true) :- !.
elist_test(L, true) :- list(L), L \= [], !.
elist_test(L,(L\=[])).

 '$expandSP'(L, L1 ++ L2 ++ L3, Trans) :-
     !,
     (
       L1 = S1+""?Cond
     ->
       Trans=( string_concat(S1,RL,L),Cond,Concats)
     ;
       L1= S1?Cond
     ->
       Trans = (string_concat(S1,RL,L),Cond,Concats)
     ;
       Trans= ( string_concat(L1,RL,L),Concats)
     ),
     '$expandSP'(RL, L2 ++ L3 , Concats).  

 '$expandSP'(LR, (L1++ L2), Trans) :-
     (
       (L1=S1?Cond1, L2=S2?Cond2)
     -> 
       %estring_test(S1,E1),
       %estring_test(S2,E2),
       Trans=(string_concat(S1,S2,LR),Cond1,Cond2)
     ;
       L1=S1?Cond1
     ->
       %estring_test(S1,E1),
       %estring_test(L2,E2),
       Trans=(string_concat(S1,L2,LR),Cond1)
     ;
        L2=S2?Cond2
     ->
       %estring_test(L1,E1),
       %estring_test(S2,E2),
       Trans=(string_concat(L1,S2,LR),Cond2)
     ;
       %estring_test(L1,E1),
       %estring_test(L2,E2),
       Trans=(string_concat(L1,L2,LR))
     ).

estring_test((_L+""),true) :- !.
estring_test(L,true) :- string(L), L \= "".
estring_test(L,(L\="")).



%% above are only for top-level
/*
trans_term((A where B), (A :: B)) :- !.
trans_term((A then B), (A -> B)) :- !.
trans_term((A & B), (A, B)) :- !.

trans_term((A or_while B), (A ++ B)) :- !.
trans_term((A until B), (A ^^ B)) :- !.
trans_term((A or_while_until B), (A +^ B)) :- !.
trans_term((not A), (\+ A)) :- !.
trans_term((A -- WR), Trans) :-
    !,
    (
      WR =   wait(T)/repeat(N)
    ->
      Trans = wait_repeat(A, T, N)
    ;
      WR = wait(T)
    ->
      Trans = wait_repeat(A, T, 1)
    ;
      fail
    ).
trans_term( (@ Qs ::  C1 => C2), forall(Qs,C1,C2)) :- !.
trans_term( (E in SorL), in__(E,SorL) ) :- !.

trans_term( ( L1 ++ L2 ?= SExp), (Concats1,Concats2)) :- !, 
    '$expandS'(SExp,L-Concats1), 
    '$expandSP'(L, (L1 ++ L2), Concats2).
trans_term( ( L1 <> L2 ?= LExp), (Concats1,Concats2)) :- !, 
    '$expandL'(LExp,L-Concats1), 
    '$expandLP'(L, (L1 <> L2), Concats2).

trans_term( (E subs S), ( S = S1, sub_string(S1,_,LE,_,E), E \= "")) :- !. 


trans_term(X, X).

*/
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Auxliary predicates
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%% determine if a given rule mentions times
'$times_of'(_, _, has_times, has_times) :- !.
'$times_of'((_ ++ time(_)), _RHS, _Times, has_times) :- !.
'$times_of'((_ ++ time(_,_)), _RHS, _Times, has_times) :- !.
'$times_of'((_ ++ time(_) ^^ _), _RHS, _Times, has_times) :- !.
'$times_of'((_ ++ time(_,_) ^^ _), _RHS, _Times, has_times) :- !.
'$times_of'((_ ^^ time(_)), _RHS, _Times, has_times) :- !.
'$times_of'((_ ^^ time(_,_)), _RHS, _Times, has_times) :- !.
'$times_of'(_, _, T, T).

'$flip_sign'(-, +).
'$flip_sign'(+, -).


'$list2tuple'([X], X) :- !.
'$list2tuple'([X|Y], (X, Ys)) :-
    '$list2tuple'(Y, Ys).

'$index_member'(X, I, L) :-
    L = [_|_],!,
    append(Y, [X|_], L),
    length(Y, N),
    I is N+1.
'$index_member'(X, I, L) :-
    '$tuple2list'(L, L1),
    append(Y, [X|_], L1),
    length(Y, N),
    I is N+1.

'$tuple_to_list'(TA, A) :-
    compound(TA), TA = '$tuple'(A), !.
'$tuple_to_list'(A, [A]).

'$list_to_tuple'([A], A) :- !.
'$list_to_tuple'(As, '$tuple'(As)).

'$tuple_type_to_list'(TA, A) :-
    compound(TA), TA = '$tuple_type'(A), !.
'$tuple_type_to_list'(A, [A]).

'$list_to_tuple_type'([A], A) :- !.
'$list_to_tuple_type'(As, '$tuple_type'(As)).


'$tuple2list'(A, B) :-
    var(A), !, B = [A].
'$tuple2list'((A,B), Result) :-
     !,
     '$tuple2list'(A, AL),
     '$tuple2list'(B, BL),
     append(AL,BL,Result).
'$tuple2list'(A, [A]).

'$tuple_len'(TA, B) :-
    compound(TA), TA = '$tuple'(A), !,
    length(A, B).
'$tuple_len'(_, 1).

'$tuple_type_len'(TA, B) :-
    compound(TA), TA = '$tuple_type'(A), !,
    length(A, B).
'$tuple_type_len'(_, 1).

'$disj2list'(A, B) :-
    var(A), !, B = [A].
'$disj2list'((A|B), Result) :-
     !,
     '$disj2list'(A, AL),
     '$disj2list'(B, BL),
     append(AL,BL,Result).
'$disj2list'(A, [A]):-
      '$is_term'(A).

/*
'$get_map'(TR, TR1) :-
    thaw_term(TR),
    '$expand_all_map'(TR, TR1, Vars),
    '$do_unifs'(Vars),
    collect_vars(TR1, UVars),
    '$unify_underscore'(UVars).
*/

'$set_var_names'(Term, VarNames, Thawed) :-
    thaw_term(Term, Thawed),
    '$do_unifs'(VarNames),
    collect_vars(Term, UVars),
    '$unify_underscore'(UVars).




/*
  '$add_plus_modes'(T, MT) :-
    T =.. [P|Args],
    '$add_plus_modes_aux'(Args, MArgs),
    MT =..[P|MArgs].

'$add_plus_modes_aux'([], []).
'$add_plus_modes_aux'([A|Args], [(+A)|MArgs]) :-
    '$add_plus_modes_aux'(Args, MArgs).

'$add_to_type_info'(Term, InArgs, OutArgs, InTypes, OutTypes) :-
    '$type_info'(_, relation, Types, Term), !,
    Term =.. [_|Args],
    '$split_on_mode'(Args, Types, InArgs, OutArgs, InTypes, OutTypes).
'$add_to_type_info'(Term, InArgs, OutArgs, InTypes, OutTypes) :-
    '$type_info'(_, action, Types, Term), !,
    Term =.. [_|Args],
    '$split_on_mode'(Args, Types, InArgs, OutArgs, InTypes, OutTypes).
'$add_to_type_info'(Term, InArgs, OutArgs, InTypes, OutTypes) :-
    '$get_ultimate_functor'(Term, R),
    '$type_info'(R, Kind, _, G), G \= '$function', !,
    (
      '$get_function_type_info'(Term,  Args, Types, _ResultType)
    ->
      Args = [A1|RArgs],
      Types = [T1|RTypes],
      '$tuple_to_list'(A1, A1L),
      '$tuple_to_list'(T1, T1L),
      '$split_on_mode'(A1L, T1L, InArgs1, OutArgs, InTypes1, OutTypes),
      append(RArgs, InArgs1, InArgs),
      append(RTypes, InTypes1, InTypes)
    ;      
      writeL(stderr, [nl_, 'Error: there is already a ', Kind,
                      ' declaration for ', R, nl_, nl_]),
      fail
    ).
    
'$add_to_type_info'(Term, InArgs, OutArgs, InTypes, OutTypes) :-
     Term =.. [R|Args],
    findall(term, member(_, Args), Types),
    length(Args, N),
    length(VarArgs, N),
    Pattern =.. [R|VarArgs],
    assert('$user_type_info'(R, relation, Types, Pattern)),
    assert('$saved_input_with_vars'((R : '<='(Types)), [])),
    writeL(stderr, [nl_, 'Warning: undeclared relation ',  R/N,
		    nl_, 'Declaring with type ', '<='(Types), nl_, nl_]),
    '$add_to_type_info'(Term, InArgs, OutArgs, InTypes, OutTypes).
    

'$split_on_mode'([], [], [], [], [], []).
'$split_on_mode'([A|Args], [T|Types], InArgs, OutArgs, InTypes, OutTypes) :-
    '$split_on_mode'(Args, Types, InArgs1, OutArgs1, InTypes1, OutTypes1),
    (
     nonvar(T), ( T = '!'(T1) )
    ->
     InArgs = [A|InArgs1],
     OutArgs = OutArgs1,
     InTypes = [T1|InTypes1],
     OutTypes = OutTypes1
    ;
     nonvar(T), ( T = ?(T1) ; T = '^'(T1) )
    ->
     InArgs = InArgs1,
     OutArgs = [A|OutArgs1],
     InTypes = InTypes1,
     OutTypes = [T1|OutTypes1]
    ;
     InArgs = InArgs1,
     OutArgs = [A|OutArgs1],
     InTypes = InTypes1,
     OutTypes = [T|OutTypes1]
    ).

'$check_type_mode_head'(Term, OutArgs, Vars, Clause) :-
    '$type_info'(_, relation, Types, Term, _), !,
    Term =.. [_|Args],
    '$check_type_mode_head_aux'(Args, Types, OutArgs, Vars, Clause, relation).
'$check_type_mode_head'(Term, OutArgs, Vars, Clause) :-
    '$type_info'(_, action, Types, Term, _), !,
    Term =.. [_|Args],
    '$check_type_mode_head_aux'(Args, Types, OutArgs, Vars, Clause, action).
'$check_type_mode_head'(Term, OutArgs, Vars, Clause) :-
    '$get_ultimate_functor'(Term, R),
    '$type_info'(R, Kind, TType, G, _), !,
    (
      G \= '$function', \+'$belief'(G), \+'$percept'(G),
      '$get_function_type_info'(Term,  Args, Types, _ResultType)
    ->
      Args = [A1|RArgs],
      Types = [T1|RTypes],
      map('$add_input', RTypes, IRTypes),      
      '$tuple_to_list'(A1, A1L),
      '$tuple_to_list'(T1, T1L),
      append(RArgs, A1L, AllArgs),
      append(IRTypes, T1L, AllTypes),
      '$check_type_mode_head_aux'(AllArgs, AllTypes, OutArgs, Vars, Clause, relation)
    ;
     '$belief'(G),
     Term =.. [_|Args]
     ->
      '$check_type_mode_head_aux'(Args, TType, OutArgs, Vars, Clause, belief)
    ;
    '$percept'(G),
     Term =.. [_|Args]
     ->
      '$check_type_mode_head_aux'(Args, TType, OutArgs, Vars, Clause, percept)
    ;
      functor(Term, R, N),
      functor(G, R, NG),
      writeL(stderr, [nl_, 'Error: the definition of ', R/N, 
                      ' does not match the declaration of the ',
                      Kind, ' ', R/NG, nl_, nl_]),
      fail
    ).
    
'$check_type_mode_head'(Term, OutArgs, Vars, Clause) :-
    Term =.. [R|Args],
    findall(term, member(_, Args), Types),
    length(Args, N),
    length(VarArgs, N),
    Pattern =.. [R|VarArgs],
    assert('$user_type_info'(R, relation, Types, Pattern, Vars)),
    assert('$saved_input_with_vars'((R : '<='('$tuple'(Types))), Vars)),
    writeL(stderr, [nl_, 'Warning: undeclared relation ',  R/N,
		    nl_, 'Declaring with type ', '<='('$tuple'(Types)), nl_, nl_]),
    '$check_type_mode_head'(Term, OutArgs, Vars, Clause).

'$check_type_mode_head_aux'([], [], [], _Vars, _Clause, _).
'$check_type_mode_head_aux'([A|Args], [T|Types], OutArgs, Vars, Clause, RelAct) :-
     (
      nonvar(T), ( T = '!'(T1) )
    ->
      ip_set(type_order, eq),
      '$hasTypes'([A], [T1], relhead_type(RelAct, Vars, Clause)),
      freeze_term(A),
      OutArgs = OutArgs1
    ;
     nonvar(T), ( T = '^'(T1) )
    ->
      ip_set(type_order, le),
      '$hasTypes'([A], [T1], relhead_type(RelAct, Vars, Clause)),
      OutArgs = [A|OutArgs1]
    ;
     nonvar(T), ( T = ?(T1))
    ->
      ip_set(type_order, eq),
      '$hasTypes'([A], [T1], relhead_type(RelAct, Vars, Clause)),
      OutArgs = [A|OutArgs1]
    ;
     nonvar(T), ( T = '+?'(T1))
    ->
      ip_set(type_order, eq),
      '$hasTypes'([A], [T1], relhead_type(RelAct, Vars, Clause)),
      OutArgs = OutArgs1
    ;
      RelAct = action
     ->
      ip_set(type_order, ge),
      '$hasTypes'([A], [T], relhead_type(RelAct, Vars, Clause)),
      freeze_term(A),
      OutArgs = OutArgs1
     ;       
      ip_set(type_order, eq),
      '$hasTypes'([A], [T], relhead_type(RelAct, Vars, Clause)),
      OutArgs = [A|OutArgs1]
    ),
   '$check_type_mode_head_aux'(Args, Types, OutArgs1, Vars, Clause, RelAct).
*/



'$frozen_term'(T) :-
    freeze_term(T, Vars), Vars \= [], !, fail.
'$frozen_term'(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Checking against grammar but more exact checking
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'$is_term'(T) :-
    atom(T).
'$is_term'(T) :-
    \+list(T),
    compound(T).


'$is_valid_resource_decl'(Rs) :-
    forall(member(R, Rs), '$type_info'(_, defined_type, _, R, _)).

'$is_defined_type_decl'(TName, T, VarNames) :-
    '$is_term'(TName), !,
    TName =.. [TN|TNArgs],
    atom(TN),
    forall(member(A, TNArgs), var(A)),
    '$is_a_type'(T),
    %% above checks it is a valid type description 
    %%        - range, enum, cons or union or an atom or var
    collect_vars(TName, HeadVars),
    collect_vars(T, BodyVars),
    (
      diff_list(HeadVars, BodyVars, [])
    ->
      true
    ;
      '$display_error'(invalid_type_unmatched_vars((TName ::= T), VarNames))
    ).
       
%%%%keith term type  defs-- lots of changes following
/*
  '$just_new_constructors'((C1|C2)):-
    !,
    '$just_new_constructors'(C1),
   '$just_new_constructors'(C2).
'$just_new_constructors'(C) :-
    errornl('$just_new_constructors'(C)),
     '$is_term'(C),
    C =.. [P|Args],
    atom(P),
    \+ is_declared_type_name(P), 
    forall(member(A, Args), '$is_a_stype'(A)).
*/

'$no_types'(C) :-
    \+is_declared_type_name(C). 

'$just_new_constructor'(C) :-
     '$is_term'(C),
    C =.. [P|Args],
    atom(P),
    \+is_declared_type_name(P), !,
    forall(member(A, Args), (var(A) ; '$is_a_type'(A))).
'$just_new_constructor'(C) :-
    ip_lookup(defined_type_check, V), !, nonvar(V),
    V = dtc(A, AT, VarNames),
    '$display_error'(invalid_constructor_type((A ::= AT), VarNames, C)).

%%%%keith above needs tightening so that a constructor functor (or atom) only
% appears in one constructor type def. This allows union types, enum and range types
% to be types of the constructor arguments. 

'$is_struct_type'(T) :-
    '$is_term'(T),
    T =.. [P|Args],
    atom(P),
    forall(member(A, Args), '$is_a_type'(A)).

/*
is_declared_type(A,AT) :-
    '$tmp_decl'(defined_type(A, AT)).
is_declared_type(A, AT) :-
   '$defined_type'(A,AT).
*/

is_declared_type_name(TN) :-
    '$defined_type'(A,_),
    functor(A,TN,_),
    !. 
   
'$decl_kind'({}(_), relation).
'$decl_kind'((_ -> _), function).

'$is_moded_struct_type'(T) :-
    '$is_term'(T),
    T =.. [P|Args],
    atom(P),
    forall(member(A, Args), '$is_a_moded_type'(A)).

'$is_term_stype'(T) :-
    var(T),
    !.
'$is_term_stype'(T) :-
    is_primitive_or_defined_type(T).



 is_primitive_or_defined_type(T) :-
     member(T, [atom,nat,int,num,string,atomic,term, top, code, '$none_', belief,percept,actcall,relcall,trcall]),
     !. 
 is_primitive_or_defined_type(T):-
    once('$defined_type'(T,_)). 

%%'$check_enum_atom_type'(T) :-
%%    '$no_types'(T), !.
'$check_enum_atom_type'(T) :-
    is_declared_type_name(T),
    ip_lookup(defined_type_check, V), !, nonvar(V),
    V = dtc(A, AT, VarNames),
    '$display_error'(invalid_atom_enum_type((A ::= AT), VarNames, T)).
'$check_enum_atom_type'(_).

%%'$is_a_type'(T) :-    errornl(is_a_type______(T)),fail.
'$is_a_type'(X) :- var(X), !.
'$is_a_type'('$enum_type'(TList)) :-
    !,
    (
     forall(member(T, TList), (atom(T), '$check_enum_atom_type'(T)))
    ;
     forall(member(T, TList), string(T))
    ;
     forall(member(T, TList), number(T))
    ).
'$is_a_type'('$constr_enum_type'(TList)) :-
    !,
    forall(member(T, TList), (compound(T), '$just_new_constructor'(T))).
'$is_a_type'( '$union_type'(Types)) :-
    !,
    %'$tuple2list'(Types, TList),
    forall(member(T, Types), '$is_a_type'(T)).
'$is_a_type'( '$list_type'(T)) :-
    !,
    '$is_a_type'(T).
'$is_a_type'( '$set_type'(T)) :-
    !,
    '$is_a_type'(T).
'$is_a_type'( '$type_macro'(T)) :-
    !,
    '$is_a_type'(T).
'$is_a_type'('$tuple_type'(T)) :-
    !,
    forall(member(T1, T), '$is_a_type'(T1)).
'$is_a_type'((T1 -> T2)) :-
    !,
    '$is_a_type'(T1),
    '$is_a_type'(T2).
'$is_a_type'('<='(Types)) :-
    !,
    forall(member(T1, Types), '$is_a_type'(T1)).
'$is_a_type'('~>'(Types)) :-
    !,
    forall(member(T1, Types), '$is_a_type'(T1)).
'$is_a_type'('~>>'(Types)) :-
    !,
    forall(member(T1, Types), '$is_a_type'(T1)).

'$is_a_type'(typeE(T)) :-
    !,
    '$is_a_type'(T). 
'$is_a_type'(T) :-
    '$is_a_stype'(T),
    !. 
'$is_a_type'(C1) :-
    atom(C1), !,
    (
     is_declared_type_name(C1)
    ->
     true
    ;
     ip_lookup(defined_type_check, V), nonvar(V)
    ->
     V = dtc(A, AT, VarNames),
     '$display_error'(invalid_type((A ::= AT), VarNames, C1))
    ;
     fail
    ). 
'$is_a_type'(C1) :-
    !,
    (
      compound(C1),
      '$defined_type'(C1,_)
    ->
      true
    ;
      ip_lookup(defined_type_check, V), nonvar(V),
       V = dtc(A, AT, VarNames),
      '$display_error'(invalid_constructor_type((A ::= AT), VarNames, C1))
    ).
'$is_a_type'(T) :-
    errornl(failed_is_a_type______(T)),fail.

tuple_member(T, (T,_Ts)).
tuple_member(T, (_,Ts)) :- 
    !,
    tuple_member(T,Ts).
tuple_member(T, T). 

/*
  alt_member(T, (T|_Ts)).
alt_member(T, (_|Ts)) :- 
    !,
    alt_member(T,Ts).
alt_member(T, T) :- T \= '$nothing'. 
*/

'$is_a_stype'(T) :-
    var(T), !.
'$is_a_stype'([]) :- !, fail.
'$is_a_stype'('$none_') :-
    !.
'$is_a_stype'(T) :-
    is_primitive_or_defined_type(T),
    !,
    (
     compound(T)
    ->
     '$is_struct_type'(T) 
    ;
     true
    ).

% so a specialisation of a type does not need to be prefixed
'$is_a_stype'((N..M)) :-
    !,
    integer(N), integer(M),
    N<M.
'$is_a_stype'( '$enum_type'(Types)) :-
    closed_list(Types),
    !,
    forall(tuple_member(T, Types), '$is_a_type'(T)).
'$is_a_stype'( '$constr_enum_type'(Types)) :-
    closed_list(Types),
    !,
    forall(tuple_member(T, Types), '$is_a_type'(T)).
/*
'$is_a_stype'(Types) :-
    closed_list(Types),
    !,
    forall(member(T, Types), atom(T)).
*/
%%%keith end of changes to type checking/inf

'$is_a_moded_type'(T) :-
    nonvar(T),
    T = +T1, !,
    '$is_a_type'(T1).
'$is_a_moded_type'(T) :-
    '$is_a_type'(T), 
     !.

'$is_resource_decl'(Rs) :-
    closed_list(Rs),
    forall(member(R, Rs),
	   (atom(R) ; compound(R), R = (R1:T1), atom(R1), atom(T1))).

%%%%%keith  these following have been changd a lot 
/*
'$is_dynamicP_decl'(D) :-
    atom(D),
    !.
'$is_dynamicP_decl'(D) :-
    \+list(D),
     !,
    '$is_struct_type'(D).
'$is_dynamicB_decl'(D) :-
    atom(D),
    !.
'$is_dynamicB_decl'(D) :-
    \+list(D),
     !,
    '$is_struct_type'(D).
*/
'$is_relation_decl'(R) :-
   '$is_moded_struct_type'(R).

'$is_function_decl'((F -> T)) :-
    '$is_struct_type'(F),
    '$is_a_type'(T).

/*
'$is_procedure_decl'(P) :-
    \+ list(P),
    !,
   '$is_moded_struct_type'(P).
*/
  '$is_durative_decl'(D) :-
    \+list(D),
     !,
    '$is_struct_type'(D).
'$is_discrete_decl'(D) :-
    \+list(D),
     !,
    '$is_struct_type'(D).

'$is_task_start_decl'(D) :-
    atom(D).
'$is_task_atomic_decl'(D) :-
    atom(D), !.

/*
'$is_task_atomic_decl'(D) :-
    compound(D),
    D = A+Xs,
    atom(A),
    closed_list(Xs),
    forall(member(X, Xs), hasTypes(X, resources, task_atomic_resources(D))).
% above now checks that only declared resources can be given in task atomic decl
%%%%keith end of these changes

'$is_tr_procedure'(TR) :-
    compound(TR),
    TR = (Head .. PBody),
    compound(PBody),
    PBody = {}(Body),
    '$tuple2list'(Body, Rules),
    forall(member(Rule, Rules), '$is_tr_rule'(Rule)).
'$is_tr_rule'(R) :-
    compound(R),
    R = (Guard -> Action),
    '$is_tr_rule_guard'(Guard),
    '$is_extended_tr_rule_action'(Action).
'$is_tr_rule_guard'(G) :-
    compound(G),
    (
     G = ( Guard +^ OrWhile )
    ;
     G = ( Guard ++ OrWhile )
    ;
     G = ( Guard ^^ OrWhile )
    ),!,
    '$is_guard_term_seq'(Guard),
    '$is_or_while'(OrWhile).
'$is_tr_rule_guard'(Guard) :-
    '$is_guard_term_seq'(Guard).
'$is_guard_term_seq'(Guard) :-
    nonvar(Guard),
    '$tuple2list'(Guard, Goals),
    forall(member(G, Goals), ('$is_goal_term'(G), G \= '!')).
'$is_or_while'(OW) :-
    nonvar(OW),
    OW = time(T),
    !,
    number(T).
'$is_or_while'(OW) :-
    nonvar(OW),
    OW = time(G, T),
    !,
    number(T),
    '$is_guard_term_seq'(G).
'$is_or_while'(OW) :-
    '$is_guard_term_seq'(OW).
'$is_extended_tr_rule_action'(Action) :-
    compound(Action),
    Action = (A ~~ G), !,
    '$is_tr_rule_action'(A),
    '$is_guard_term_seq'(G).
'$is_extended_tr_rule_action'(Action) :-
    '$is_tr_rule_action'(A).
'$is_tr_rule_action'(A) :-
    compound(A),
    A = {}(As),
    '$tuple2list'(As, AList),
    forall(member(Act, AList), '$is_primitive_action'(Act)).
'$is_tr_rule_action'(A) :-
    nonvar(A),
    '$is_tr_procedure'(A).
*/

'$is_goal_term'(G) :-
    var(G), !, fail.
/*
'$is_goal_term'(G) :-
    functor(G, P, N),
    '$decl_kind'(P, N, T),
    ( T = dynamicB;  T = dynamicP ).
*/'$is_goal_term'(G) :-
    '$type_info'(_, action, _, G, _), !.
'$is_goal_term'(G) :-
    '$type_info'(_, relation, _, G, _), !.
'$is_goal_term'(G) :-
    '$type_info'(_, function, _, G, _), !,
    G \= '$function'.
'$is_goal_term'(type(_, _)).
'$is_goal_term'(type(_, _, _)).
'$is_goal_term'(typeG(_, _)).
'$is_goal_term'(typeG(_, _, _)).
'$is_goal_term'(ground(_)).
'$is_goal_term'(remember(_)).
'$is_goal_term'(rememberA(_)).
'$is_goal_term'(remember_for(_,_)).
'$is_goal_term'(rememberA_for(_,_)).
'$is_goal_term'(forget(_)).
'$is_goal_term'(forget_after(_,_)).
'$is_goal_term'(replace_by(_)).
'$is_goal_term'(call(_)).
'$is_goal_term'(do(_)).
'$is_goal_term'('$at_forall'(_, _, _)).
'$is_goal_term'('$forall'(_, _, _)).
'$is_goal_term'('$exists'(_, _)).
'$is_goal_term'('$at_forall_actions'(_, _, _)).
'$is_goal_term'('$forall_actions'(_, _, _)).
'$is_goal_term'(term).

%%%%keith we must allow sending of terms

% '$is_goal_term'(true).
% '$is_goal_term'(fail).
%%%%%keith  should not allow ! in goals, or true or fail
% '$is_goal_term'('!').

%%'$is_a_valid_type'(T) :- errornl('$is_a_valid_type'(T)),fail.
'$is_a_valid_type'(T) :-
    var(T), !.
'$is_a_valid_type'(T) :-
    T = typeE(T1), !,
    '$is_a_valid_type'(T1).
'$is_a_valid_type'(T) :-
     is_primitive_or_defined_type(T), !.
'$is_a_valid_type'(T) :-
    T = '$enum_type'(Enum),!,
    forall(member(X, Enum), atom(X)).
'$is_a_valid_type'(T) :-
    T = '$constr_enum_type'(Enum),!,
    forall(member(X, Enum), '$is_a_constructor_type'(X)).
'$is_a_valid_type'(T) :-
    T = '$union_type'(Types), !,
    forall(member(ET, Types), '$is_a_valid_type'(ET)).
'$is_a_valid_type'(T) :-
    T = '$tuple_type'(Types), !,
    forall(member(Ty, Types), '$is_a_valid_type'(Ty)).
'$is_a_valid_type'(T) :-
    T = '$set_type'(Type), \+ (compound(Type) , Type = (_,_)), !,
    '$is_a_valid_type'(Type).
'$is_a_valid_type'(T) :-
    T = '$list_type'(Type), !,
    '$is_a_valid_type'(Type).
'$is_a_valid_type'(T) :-
    T = (T1 -> T2), !,
    '$is_a_valid_type'(T1),
    '$is_a_valid_type'(T2).
'$is_a_valid_type'(T) :-
    T = '<='(T1), !,
    '$is_a_valid_type'(T1).
'$is_a_valid_type'(T) :-
    T = '~>'(T1), !,
    '$is_a_valid_type'(T1).
'$is_a_valid_type'(T) :-
    moded(T), 
    T = M(T1),
    (
      is_a_constructor_type(T1, _, IT1)
    ->
      check_constructor_mode(M, IT1)
    ;
      '$is_a_valid_type'(T1)
    ), !.  
'$is_a_valid_type'(T) :-
    \+ moded(T),
    compound(T),
    functor(T, P, _),
    '$defined_type'(P, _), !.
'$is_a_valid_type'((N .. M)) :-
    integer(N), integer(M), N =< M, !.
'$is_a_valid_type'(T) :-
    ip_lookup(valid_type_context, valid_type_check(Clause, Vars)),
    '$display_error'(invalid_type(Clause, Vars, T)).

%%check_constructor_mode(M, T) :- errornl(check_constructor_mode(M, T)),fail.
check_constructor_mode(M, T) :-
    moded(T),
    T = M1(T1), !,
    (
      M = '?'
    ->
      ( M1 = '?' ; M1 = '+?' )
    ;
      M = '+?'
    ->
      M1 = '+?'
    ;
      true
    ),
    (
      is_a_constructor_type(T1, _, T2)
    ->
      check_constructor_mode(M1, T2)
    ;
      true
    ).
check_constructor_mode(_M, _T).

'$is_a_constructor_type'(T) :-
    compound(T),
    functor(T, F, N), N > 0,
    T =.. [F|Args],
    atom(F), 
    forall(member(Ty, Args), '$is_a_valid_type'(Ty)).

'$a_supplied_atom_type'(T) :-
    '$resource_type'(T).
'$a_supplied_atom_type'(T) :-
    '$defined_type'(T, _).
     
'$a_valid_constr_type'(T) :-
    atom(T), !.
'$a_valid_constr_type'(T) :-
    compound(T), !,
    '$is_a_valid_type'(T).
%%
%% if we have a type definition like
%% foo ::= bar
%% then the parser converts the RHS to a union type with a single entry
%% - this is treated as a type macro
%% However if the RHS is not a type then it should be converted to
%% an enumerated type with one element
%% This does the checking and replacement as necessary
%%
/*
'$check_single_enum' :-
    '$user_type_info'(A,B,C,D,_),
    D = '$union_type'([Single]),
    errornl('$user_type_info'(A,B,C,D)),
    functor(Single, P, _),
    \+'$user_type_info'(P, defined_type, _, _, _),
    retract('$user_type_info'(A,B,C,D,V)),
    assert('$user_type_info'(A,B,C,'$enum_type'([Single]),V)),
    fail.
'$check_single_enum'.
*/   
/*
  '$check_if_relation_type' :-
    forall('$user_type_info'(A,B,C,D,E), errornl('$user_type_info'(A,B,C,D,E))),
    '$user_type_info'(F, relation, Types, Head, V),
    once(
	 (
	  '$tmp_clause'(action_clause((Head ~>> _)))
	 ;
	  '$tmp_clause'(action_clause((Head ::  _ ~>>  _)))
	 )
	),
    retract('$user_type_info'(F, relation, Types, Head, V)),
    assert('$user_type_info'(F, action, Types, Head, V)),
    fail.
'$check_if_relation_type'.
*/