


%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.

%?- [qconsult].
?-op(300,fx,'watch__').
?- dynamic(watch/0).

%KKunwat?-dynamic('$watched'/1).
?-dynamic('$watch_clauses'/1).
?-dynamic('$watch_saved'/3).
?-dynamic('$query_var_types'/1).
?- assert('$num_solutions'(5)).
?-dynamic('$top_level_consulted_file'/1).

version('0.4').

main(_Args) :-
    version(V),
    catch('$qulog_process_initial_goal', Ball, '$qulog_catch_init_goal'(Ball)),
    write_term_list([wa('QuLog Version '), w(V), nl]),
    global_state_set('$indent_level', 0),
    global_state_set('$break_level', 0),
    global_state_set('$prompt', '| ?? '),
    '$qulog_interpreter'.

'$qulog_process_initial_goal' :-
    '$initial_goal'(GS),
    open_string(read(GS), S),
    read(S, Goal),
    call(Goal),    
    !.
'$qulog_process_initial_goal'.

'$qulog_catch_init_goal'(Ball) :-
    Ball = exception(Data),
    !,
    '$valid_exception_term'(Data, Type, Sev, Goal, Msg0),
    '$get_exception_message'(Data, Type, Sev, Goal, Msg0, Msg),
    write_term_list(stderr, Msg),
    thread_exit.
'$qulog_catch_init_goal'(Ball) :-
    errornl(Ball),
    thread_exit.


'$qulog_interpreter' :-
    ip_lookup('$locally_handled_exceptions', Handlers),
    (   var(Handlers)
        ->
            Handlers = []
        ;
            true
    ),
    ip_set('$locally_handled_exceptions', [abort(_, _, _)|Handlers]),
    catch('$qulog_fetch_execute', Ball, '$qulog_interpreter_catch'(Ball)),
    ip_set('$locally_handled_exceptions', Handlers),
    !.
'$qulog_interpreter' :-
    '$qulog_interpreter'.
 
'$qulog_fetch_execute' :-
    repeat,
    global_state_lookup('$prompt',Prompt),
    global_state_set(watch__,0),
    write_term_list([nl, wa(Prompt)]),
    flush_output,
    ip_set('$in_interpreter_query', true),
    interpreter_read_query(FullQuery, Vars),
    '$get_underscores'(FullQuery, Vars, Unders),
    ip_set('$global_vars', []),
    ip_set('$local_vars', []),
    ip_set('$underscore_vars', Unders),
    freeze_term(FullQuery),
    '$num_solutions'(Sols),
    (
     FullQuery = '$num_query'(N, Query)
    ->
     global_state_set('$total_solns2print', N),
      FVars = Vars
    ;
     FullQuery = '$vars_query'(QVars, Query)
    ->
     filter('$keep'(QVars), Vars, FVars),
     global_state_set('$total_solns2print', Sols)
    ; 
     FullQuery = '$num_vars_query'(N, QVars, Query)
    ->
     filter('$keep'(QVars), Vars, FVars),
     global_state_set('$total_solns2print', N)
    ; 
     Query = FullQuery,
     FVars = Vars,
     global_state_set('$total_solns2print', Sols)
    ),
    ip_set('$in_interpreter_query', fail),
    (
     '$qulog_exit_interpreter'(Query)
    ;
     '$qulog_solve_query'(Query, FVars)
    ),
    !.

'$keep'(FreeVars, (V = _)) :- member_eq(V, FreeVars).
'$ignore'(FreeVars, U) :- \+'$keep'(FreeVars, U).

'$qulog_interpreter_catch'(Ball) :-
    '$qulog_interpreter_catch_msg'(Ball),
    fail.
    
'$qulog_interpreter_catch_msg'(Ball) :-
    nonvar(Ball),
    Ball = '$exception'(abort(_, _, [Msg])),
    !,
    write_atom(stderr, Msg),
    write_atom(stderr, '\n').
'$qulog_interpreter_catch_msg'(Ball) :-
    nonvar(Ball),
    Ball = exception(undefined_predicate(_, Goal, _)),
    functor(Goal, Pred, Arity),
    !,
    write_atom(stderr, 'no definition for '),
    write(stderr, Pred),
    write_atom(stderr, '/'),
    '$write_integer'(stderr, Arity),
    write_atom(stderr, '\n').
'$qulog_interpreter_catch_msg'(Ball) :-
    nonvar(Ball),
    Ball = exception(Data),
    !,
    '$valid_exception_term'(Data, Type, Sev, Goal, Msg0),
    '$get_exception_message'(Data, Type, Sev, Goal, Msg0, Msg),
    write_term_list(stderr, Msg),
    '$exception_severity'(Sev, DefaultAction, _, _),
    call_predicate(DefaultAction).
'$qulog_interpreter_catch_msg'(Ball) :-
    nonvar(Ball),
    Ball = ctrlC_reset,
    !.
'$qulog_interpreter_catch_msg'(Ball) :-
    write_atom(stderr, caught),
    write_atom(stderr, ':'),
    errornl(Ball).

'$qulog_exit_interpreter'(G) :- G == end_of_file.

/*
 * Solve the query and print the result.
 *
 * mode '$solve_query'(@term, @list(compound)).
 */
'$qulog_solve_query'(Query, Vars) :-
    nl,
    global_state_set('$has_solution', no),
    global_state_lookup('$total_solns2print', M),
    global_state_set('$num_solns2print', M),
    %freeze_term(Query),
    once('$get_call_kind'(Query, RealQuery, Vars, Kind)),
    ip_lookup('$local_vars', Locals),
    %iff_list(Vars, Locals, QVars),
    filter('$ignore'(Locals), Vars, QVars),
    '$qulog_solve_query_aux'(Kind, RealQuery, QVars).

'$qulog_solve_query_aux'(relation(QVT), Query, Vars) :-
    '$qulog_interpreter_call'(Query),
    global_state_set('$has_solution', yes),
    '$qulog_print_result'(Vars, QVT),
    !,
    fail.
'$qulog_solve_query_aux'(relation(_QVT), _Query, _Vars) :-
    global_state_lookup('$has_solution', no),
    !,
    write('no\n'),
    fail.
'$qulog_solve_query_aux'(relation(_QVT), _Query, _Vars) :-
    !,
    global_state_lookup('$total_solns2print', M),
    global_state_lookup('$num_solns2print', N),
    (
     N = M
    ->
     write('no more solutions\n')
    ;
     true
    ),
    fail.
'$qulog_solve_query_aux'(expression(_T), Query, _Vars) :-
    %nl,
    ip_lookup(colons_used, false),
    '$qulog_interpreter_call'(Query),
    !,
    fail.
'$qulog_solve_query_aux'(expression(_T), Query, _Vars) :-
    global_state_lookup('$total_solns2print', M),
    (
      '$qulog_interpreter_call'(Query),
      fail
    ;
      global_state_lookup('$num_solns2print', N),
      (
        N = M
      ->
        write('no more values\n')
      ;
        true
      ),
      global_state_set('$num_solns2print', M),
      fail
    ).
    
/*
'$qulog_solve_query_aux'(expression(_T), _Query, _Vars) :-
    write('\nUndefined - no applicable rule\n'),
    !,
    fail.
*/
'$qulog_solve_query_aux'(action(QVT), Query, Vars) :-
    '$qulog_interpreter_call'(Query),
    (
      Vars = []
    ->
      true
    ;
    '$qulog_print_bindings_section'(Vars, QVT)
    ),    
    write('success\n'),
    !,
    fail.
'$qulog_solve_query_aux'(action(_QVT), _Query, _Vars) :-
    write('fail\n'),
    !,
    fail.

'$print_alt_expression_soln'(M, Term) :-
    global_state_decrement('$num_solns2print', N),
    (
      M is N+1
    ->
      true
    ;
      write_term_list(['...',nl])
    ),
    flush_output,
    (
      N = 0
    ->
      true
    ;
      print__aux(Term),
      fail 
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%% Put Query interface code here %%%%%%%%%%%%%%
%K moved to here from compiler
% Debgging code

watch__ (function(_NF), Call) :-
    watch,
    !,
    global_state_increment(watch__,_WL),
    Call.
watch__ (function(_NF), Call) :-
    Call. 

  
watch__ (NR,Call) :- 
    watch,
   !,
    global_state_increment(watch__,WL),
    open_string(write, Stream),
    %% PJR - remove comments below to unset var names
    %(
      name_vars(Call),
      writeTerm(Stream, Call),
    %  fail
    %;
    %  true
    %),
    stream_to_string(Stream, String),
    string_to_list(String, Chars),
    (
     write(WL), write(':'),
      forall(member(C, Chars), put_code(C)),
      nl
     ;
      write(WL), write(':'),
      forall(member(C, Chars), put_code(C)),
      write(' no (more) proofs\n'), 
      fail
    ),
    Call=..[_|Args],
    collect_vars(Call,CallVars),
    construct(CallVars,NamePairs),
% above constructs a list of Name=Var pairs of all vars in Call
% that will have been given names by above name_vars(Call)
% This list is passed into the watched call so that output bindings can be displayed
    NewCall=..[NR,NamePairs|Args],
    NewCall,
    (
      writeL([WL,":",Call, " succeeded\n" ])
    ;
      display_call_num(WL), 
      forall(member(C, Chars), put_code(C)), 
      writeL([" seeking another proof\n"]), 
      fail
    ).

watch__ (NR, Call) :-
    Call=..[_|Args],
    NewCall =.. [NR,_|Args],
    NewCall. 

display_call_num(1) :- !,  nl, writeL([1, ":"]).
display_call_num(WL) :- writeL([WL, ":"]).




?-op(1150,fx,prolog).
%?-op(1150,xfx,<~).


%%'$qulog_interpreter_call'(X) :- errornl('$qulog_interpreter_call'(X)), fail.
/*
  '$qulog_interpreter_call'(consult(File)) :-
    '$top_level_consulted_file'(File1),
    File \= File1,
    !,
    writeL(stderr, ["Error: Only one file can be consulted at the top level of the interpreter. ", nl_, File, " has already been consulted.", nl_]),
    fail.
'$qulog_interpreter_call'(consult(File)) :-
    !,
    assert('$top_level_consulted_file'(File)),
    top_level_qconsult(File).
*/
'$qulog_interpreter_call'(pconsult(File)) :-
    !,
    consult(File).
'$qulog_interpreter_call'(watch) :-
    !,
    assert(watch).

'$qulog_interpreter_call'(set_num_answers(N)) :-
    !,
    N > 0,
    retract('$num_solutions'(_)),
    assert('$num_solutions'(N)).

'$qulog_interpreter_call'(unwatch) :-
    !,
    retractall(watch).

'$qulog_interpreter_call'(watch(Rel)) :-
    '$type_info'(Rel, relation, _, _, _),
      '$watched'(Rel),
      !,
      write_term_list([Rel,' already being watched\n']).
     
%% Watching functions
'$qulog_interpreter_call'(watch(Fun)) :-
      '$type_info'(Fun, function, _, Type, _),
      !,
      (
        Type = '$tuple'(TL)
      ->
        length(TL, N)
      ;
        N = 1
      ),
      N1 is N+1,
      %% N is the arity of the function, N1 is the arity of the relation
      %% (includes result arg)
      atom_concat2(Fun,'_watched',NFun),
      functor(FunCall, Fun, N1),
      functor(HeadPattern, Fun, N1),
      FunCall =.. [_|Args],
      WFunCall =.. [NFun|Args],
      watch_trans_function_clauses(Fun, N),
      forall(watch_retract(HeadPattern, Body),
             assert('$watch_saved'(Fun, HeadPattern, Body))),
      assert((FunCall :- watch__ (function(NFun),WFunCall))),
      remember(watch),
      assert('$watched'(Fun)). 
%% Watching relations
'$qulog_interpreter_call'(watch(Rel)) :-
      !,
      once((
       '$type_info'(Rel, relation, _, RelCall, _)
	;
	'$type_info'(Rel, action, _, RelCall, _)
      )),
      functor( RelCall, Rel, N),
      atom_concat2(Rel,'_watched',NRel),
      RelCall=..[Rel | _Args],
      %% remember stuff XXX
      functor(HeadPattern, Rel, N),
      watch_trans_clauses(Rel, N),
      forall(watch_retract(HeadPattern, Body),
             assert('$watch_saved'(Rel, HeadPattern, Body))),
      assert((RelCall :- watch__ (NRel,RelCall))),
      remember(watch),
      assert('$watched'(Rel)). 

'$qulog_interpreter_call'(unwatch(Rel)) :-
      retract('$watched'(Rel)),
      !,
      retractall('$watch_clauses'(Rel)),
      '$type_info'(Rel, relation, _, RelCall, _),
      functor( RelCall, Rel, N),
      atom_concat2(Rel,'_watched',NRel),
      functor( NRelCall, NRel, N),
      retract( (RelCall:- watch__ _)),
      retractall(NRelCall),
      forall(retract('$watch_saved'(Rel, HeadPattern, Body)),
               assert((HeadPattern :- Body))).

'$qulog_interpreter_call'(unwatch(Rel)) :-
      !,
      write_term_list([Rel,' was not being watched\n']).

'$qulog_interpreter_call'(watchC(Rel)) :-
      '$watch_clauses'(Rel),
      !,
      write_term_list([Rel,' clauses already being watched\n']).
 
'$qulog_interpreter_call'(watchC(Rel)) :-
       '$watched'(Rel),
       !,
      assert('$watch_clauses'(Rel)).

'$qulog_interpreter_call'(watchC(Rel)) :-
    (
     '$type_info'(Rel, relation, _, _, _)
    ;
     '$type_info'(Rel, action, _, _, _)
    ),
    !,
    assert('$watch_clauses'(Rel)),
    '$qulog_interpreter_call'(watch(Rel)).
'$qulog_interpreter_call'(watchC(Fun)) :-
      '$type_info'(Fun, function, _, _, _),
      !,
      assert('$watch_clauses'(Fun)),
      '$qulog_interpreter_call'(watch(Fun)).

'$qulog_interpreter_call'(unwatchC(Rel)) :-
    retract(  '$watch_clauses'(Rel) ),
    !.

'$qulog_interpreter_call'(unwatchC(Rel)) :-
      !,
      write_term_list([Rel,' clauses were not being watched\n']).


'$qulog_interpreter_call'(prolog) :-
    !,
    global_state_lookup('$prompt', Prompt),
    global_state_set('$prompt', '| ?- '),
    interpreter,
    global_state_set('$prompt', Prompt).

'$qulog_interpreter_call'(prolog(Query)) :-
    !,
    thaw_term(Query),
    call_prolog(Query).

    
'$qulog_interpreter_call'(types) :-
    !,
    types.

'$qulog_interpreter_call'(types(Rel)) :-
    !,
   types(Rel).
'$qulog_interpreter_call'(stypes(Rel)) :-
    !,
   stypes(Rel).
'$qulog_interpreter_call'(stypes) :-
    !,
    stypes.
'$qulog_interpreter_call'(show) :-
    !,
    show.
'$qulog_interpreter_call'(show(Rel)) :-
    !,
   show(Rel).

'$qulog_interpreter_call'(isa(Term,T)) :-
    !,
    thaw_term((Term, T)),
    isa(Term,T). 
'$qulog_interpreter_call'(type(Term,T,A)) :-
    !,
    thaw_term((Term, T, A)),
    type(Term,T,A). 
  '$qulog_interpreter_call'(type(Term,T)) :-
    !,
    thaw_term((Term, T)),
    type(Term,T,_).
'$qulog_interpreter_call'(typeC(Term,T,A)) :-
    !,
    thaw_term((Term, T, A)),
    type(Term,T,A). 

'$qulog_interpreter_call'(log(Addr)) :-
   logger(Addr, all).
'$qulog_interpreter_call'(teleor) :-
   teleor.
'$qulog_interpreter_call'(default_agent(MyName, Addr, Interface)) :-
   default_agent(MyName, Addr, Interface).
'$qulog_interpreter_call'(bs) :-
   '$show_bs'.
   
 

'$qulog_interpreter_call'(Query) :-
    thaw_term(Query),
     call(Query).

%%'$get_call_kind'(X, _, _, _) :- errornl('$get_call_kind'(X)),fail.

'$get_call_kind'([File], interpreter_consult(File), _, action([])) :-
    !.
'$get_call_kind'(consult(File), interpreter_consult(File), _, action([])) :-
    !.
'$get_call_kind'(pconsult(File),pconsult(File), _, action([])) :-
    !.
'$get_call_kind'(watch,watch, _, action([])) :-
    !.
'$get_call_kind'(unwatch, unwatch,_, action([])) :-
    !.
'$get_call_kind'(set_num_answers(N), set_num_answers(N),_, action([])) :-
    !.

'$get_call_kind'(watch(Rel), watch(Rel), _,  action([])) :-
    !.
'$get_call_kind'(unwatch(Rel),unwatch(Rel), _, action([])) :-
      !.
'$get_call_kind'(watchC(Rel), watchC(Rel), _, action([])) :-
      !.
'$get_call_kind'(unwatchC(Rel),unwatchC(Rel), _, action([])) :-
    !.
'$get_call_kind'(prolog, prolog,_, action([])) :-
    !.
'$get_call_kind'(prolog(Query),prolog(Query), _, relation([])) :-
    !.
'$get_call_kind'(types,types,_, action([])) :-
    !.
'$get_call_kind'(types(Rel),types(Rel),_, action([])) :-
    !.
'$get_call_kind'(stypes,stypes,_, action([])) :-
    !.
'$get_call_kind'(stypes(Rel),stypes(Rel),_, action([])) :-
    !.
'$get_call_kind'(show, show,_, action([])) :-
    !.
'$get_call_kind'(show(Rel), show(Rel), _, action([])) :-
    !.
'$get_call_kind'(log(Addr), log(Addr), _, action([])) :-
    !.
'$get_call_kind'(unlog, unlog, _, action([])) :-
    !.
'$get_call_kind'(bs, bs, _, action([])) :-
    !.
'$get_call_kind'(teleor, teleor, _, action([])) :-
    !.
'$get_call_kind'(default_agent(N,A,B), default_agent(N, A, B), _, action([])) :-
    !.
'$get_call_kind'(type(Term,T,A),type(Term,T,A), _Vars, relation([])) :-
    !. 
'$get_call_kind'(type(Term,T),type(Term,T),_Vars, relation([Term:T])) :-
    !.
'$get_call_kind'(typeC(Term,T,A),typeC(Term,T,A),_Vars, relation([])) :-
    !. 
'$get_call_kind'(isa(Term,T),isa(Term,T),_Vars, relation([])) :-
    !.
%%'$get_call_kind'(TypeCheck, typeG(Term, Type), _, relation([])) :-
%%    '$is_a_type_test'(TypeCheck, Term, Type), !.
'$get_call_kind'(Query, TQuery, Vars, GKind) :-
     '$check_query'(Query, Vars), !,
    ip_set(colons_used, false),
    %%ip_lookup('$underscore_vars', Unders),
    (
      check_query_mode_type(Query, Vars, Kind)
    ->
      ip_set(query_type___, Kind),
      '$translate_query'(Query, TQuery),
      thaw_term(TQuery),
      (
        Query = (_ :: _)
      ->
        ip_set(colons_used, true)
      ;
        true
      ),
      thaw_term(Kind, KVars),
      findall(Kind, '$vars2names'(Kind, Vars), [GKind]),
      freeze_term(KVars)
    ;
      writeL(stderr, ["Error: Illegal query ", Query, nl_]),
      fail
    ).
'$get_call_kind'(Query, _TQuery, _Vars, _GKind) :-
    writeL(stderr, ["Error: Illegal query ", Query, nl_]),
    fail.

in__(E,SorL) :- string(SorL), !, sub_string(SorL,_,1,_,E).
in__(E,SorL) :- closed_list(SorL), member(E,SorL).

length__(SorL,N) :- string(SorL), !, string_length(SorL,N).
length__(SorL,N) :- closed_list(SorL), length(SorL,N).

call_prolog(Query) :- var(Query), !, Query = 42.
call_prolog('&'(Q1, Q2)) :-
    !,
    call_prolog(Q1),
    call_prolog(Q2).
call_prolog(Q) :-
    call(Q).
    
%%'$qulog_print_result'(Vars, GorE) :-errornl('$qulog_print_result'(Vars, GorE)),fail.
    
'$qulog_print_result'([], _) :-
    !,
    write('yes\n'),
    flush_output.
'$qulog_print_result'(Vars, GorE) :-
    GorE == expr,
    !,
    (
      \+ '$has_bindings'(Vars)
    ->
      write_term_list([nl]),
      flush_output
    ;
      '$qulog_interpreter_cont_section'(Vars)
    ).
'$qulog_print_result'(Vars, QVT) :-
    global_state_decrement('$num_solns2print', N),
    global_state_lookup('$total_solns2print', M),
    %%errornl(p__(M, N, QVT)),
    (
     M is N+1
    ->
     true
    ;
     write_term_list(['...',nl])
    ),
    '$qulog_print_bindings_section'(Vars, QVT),
    flush_output,
    (
     N = 0
    ->
     true
    ;
     fail	
    ),
    global_state_set('$num_solns2print', M),
    '$qulog_interpreter_cont_section'(Vars).

'$has_bindings'(Vars) :-
    '$vars2names'(Vars, Vars),
    member((A=B), Vars), A \= B.
/*
'$qulog_print_bindings_section'(Instantiations, QVT) :-
    '$get_values'(Instantiations, ValuesList),
    get_all_types('$tuple'(ValuesList), Types),
    (
     Types = alt_types(TTypes)
    ->
     map(strip_tuple__, TTypes, STypes),
     zip__(STypes, ZTypes),
     map(merge_types__, ZTypes, MZTypes),
     '$qulog_print_bindings_section_aux'(Instantiations, MZTypes, QVT)
    ;
     Types = '$tuple'(TupTypes),
    '$qulog_print_bindings_section_aux'(Instantiations, TupTypes, QVT)
    ), !.
*/
strip_tuple__('$tuple'(L), L) :- !.
strip_tuple__(X, [X]).

'$qulog_print_bindings_section_aux'([], [], _QVT).
'$qulog_print_bindings_section_aux'([=(Value, Var)|Tail], [Type|Types], QVT) :-
    member(('$$var$$'(Var) : _(TypeB)), QVT),!,
    (
     identical_patterns(Type, TypeB)
     ->
     writeL([uq_(Var), " = ", wr_(Value : Type),nl_])
    ;
     writeL([uq_(Var), " = ", wr_(Value : TypeB), " (", Type, ")", nl_])
    ),
    '$qulog_print_bindings_section_aux'(Tail, Types, QVT).
'$qulog_print_bindings_section_aux'([=(Value, Var)|Tail], [Type|Types], QVT) :-
    ip_lookup('$interpreter_vmt', VMT),
    errornl(o___(Var, Value, VMT)),
    writeL([uq_(Var), " = ", wr_(Value : Type),nl_]),
    '$qulog_print_bindings_section_aux'(Tail, Types, QVT).
/*
'$qulog_print_bindings_section'([], _).
'$qulog_print_bindings_section'([=(Value, Var)|Tail], QVT) :-
    member(('$$var$$'(Var) : _(Type)), QVT), !,
    '$remove_repeats_from_union'(Type, SType),
    writeL([Var, ' = ', wr_(Value : SType),nl_]),
    '$qulog_print_bindings_section'(Tail, QVT).
'$qulog_print_bindings_section'([=(Value, Var)|Tail], QVT) :-
    writeL([Var, ' = ', wr_(Value),nl_]),
    '$qulog_print_bindings_section'(Tail, QVT).
*/

'$qulog_print_bindings_section'([], _).
'$qulog_print_bindings_section'(ValVarList, _QVT) :-
    map('$strip_out_vals', ValVarList, Values),
    collect_vars(Values, ValuesVars),
    map('$add_mode_type', ValuesVars, VMT),
    map('$bindings_get_type'(VMT), Values, ValTypes),
    !,
    collect_vars(ValTypes, TypeVars),
    reverse(TypeVars, RevTypeVars),
    '$name_type_vars'([_|RevTypeVars]),
    '$qulog_print_bindings_section_entries'(ValVarList, ValTypes).

'$qulog_print_bindings_section_entries'([], []).
'$qulog_print_bindings_section_entries'([=(Value,Var)|Vs], [T1|Ts]) :-
    ip_lookup('$interpreter_vmt', VMT),
    ( var(VMT) -> VMT = [] ; true ),
    (
     type_member_of(Value:_M(T), VMT)
    ->
     writeL([uq_(Var), " = ", wr_(Value:T), nl_])
    ;
    writeL([uq_(Var), " = ", wr_(Value:T1), nl_])
    ),
    '$qulog_print_bindings_section_entries'(Vs, Ts).

'$strip_out_vals'(=(Value, _), Value).

'$split_on_type'((A:B), A, B).

'$strip_out_types'((_:B), B).

'$add_mode_type'(V, (V : '!'(_))).

'$bindings_get_type'(VMT, Term, Type) :-
    has_type2(Term, Type, VMT, _).
    
'$name_type_vars'([]).
'$name_type_vars'([V|Vs]) :-
    (
      set_var_name(V, 'Ty')
    ;
      true
    ), !,
    '$name_type_vars'(Vs).


    

'$get_values'([], []).
'$get_values'([=(Value, _)|Tail], [Value|VTail]) :-
    '$get_values'(Tail, VTail).

'$qulog_interpreter_cont_section'(_Vars) :-
    get_line(Line),
    string2tokens(Line, Tokens, _),
    (
     Tokens = []
    ->
     true
    ;
     Tokens = [op('..')]
    ->
     global_state_lookup('$total_solns2print', M),
     global_state_set('$num_solns2print', M),
     fail
    ;
     Tokens = [op('..'), int(N)]
    ->
      global_state_set('$num_solns2print', N),
      global_state_set('$total_solns2print', N),
      fail
     ;
      true
    ).


interpreter_consult(File) :-
    '$top_level_consulted_file'(File1),
    File \= File1,
    !,
    writeL(stderr, ["Error: Only one file can be consulted at the top level of the interpreter. ", nl_, File, " has already been consulted.", nl_]),
    fail.

interpreter_consult(File) :-
    '$consulted_timestamp'(G, _, OldTime),
    File \= G,
    concat_atom([G, '.qlg'], Gqlg),
    stat(Gqlg, stat(Time, _)),
    Time > OldTime, !,
    %% at least one file that File consults directly or indirectly
    %% has changed so remove add info from all files.
    forall(retract('$consulted_timestamp'(F, _, _)), '$delete_all_code'(F)),
    top_level_qconsult(File).
interpreter_consult(File) :-
    concat_atom([File, '.qlg'], Fileqlg),
    stat(Fileqlg, stat(Time, _)),
    (
      '$consulted_timestamp'(File, File, Time)
    ->
      writeL(["Neither ", File, " nor any files consulted from this file",
              nl_, "has changed since the last consult - file not (re)consulted", nl_])
    ;
      forall(retract('$consulted'(File1)),
             assert('$saved_consulted'(File1))),
      retractall('$consulted_timestamp'(File, File, _)),
      assert('$consulted_timestamp'(File, File, Time)),
      top_level_qconsult(File)
    ).
      

