%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.

?- compile_time_only(consult(op_decls)).
	  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Writer
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

writeTerm(T) :-
    writeTerm(stdout, T).

writeTerm(Stream, T) :-
    var(T), !, write(Stream, T).
writeTerm(Stream, T) :-
    atom(T),
    atom_concat(T1, '_watched', T), !,
    write(Stream, T1).
writeTerm(Stream, T) :-
    atom(T), T = '$nil', !,
    write(Stream, '()').
writeTerm(Stream, T) :-
    atom(T), !,
    writeq(Stream, T).
writeTerm(Stream, T) :-
    integer(T), !, write(Stream, T).
writeTerm(Stream, T) :-
    number(T), !, write(Stream, T).
writeTerm(Stream, T) :-
    string(T), !, write(Stream, T).
/*
writeTerm(Stream, T) :-
    list(T), ground(T), sort(T, T), !,
    write(Stream, '{'),
    writeT_list_elems(Stream, T),
    write(Stream, '}').
*/
writeTerm(Stream, '$eval'(Exp)) :-
    !,
    write(Stream, '!'),
    writeTerm(Stream, Exp).
writeTerm(Stream, '$set'([])) :-
    !,
    write(Stream, '{}').
writeTerm(Stream, '$set'(T)) :-
    !,
    write(Stream, '{'),
    writeT_list_elems(Stream, T),
    write(Stream, '}').
writeTerm(Stream, '$set_enum'(T)) :-
    !,
    write(Stream, '{'),
    writeT_list_elems(Stream, T),
    write(Stream, '}').

writeTerm(Stream, T) :-
    list(T), !,
    writeT_list(Stream, T).
writeTerm(Stream, T) :-
    compound(T),
    freeze_term(T),
    writeT_compound(Stream, T),
    fail.
writeTerm(_Stream, _).

    
writeT_list(Stream, T) :-
    write(Stream, '['),
    freeze_term(T, V),
    writeT_list_elems(Stream, T),
    thaw_term(V),
    write(Stream, ']').
writeT_list_elems(Stream, [X]) :-
    !,
    writeTerm(Stream, X).
writeT_list_elems(Stream, [X1, X2|T]) :-
    !,
    writeTerm(Stream, X1),
    write(Stream, ', '),
    writeT_list_elems(Stream, [X2|T]).
writeT_list_elems(Stream, [X|T]) :-
    writeTerm(Stream, X),
    write(Stream, ',..'),
    writeTerm(Stream, T).

%writeT_compound(Stream, Term) :- errornl(write___(Term)),fail.
writeT_compound(Stream, Term) :-
    '$is_clause_defn'(Term), !,
    show_clause(Term, Stream).
writeT_compound(Stream, belief(F:T)) :-
    !,
    write(Stream, 'belief '),
    writeTerm(Stream, F:T).
/*
  writeT_compound(Stream, Term) :-
    '$is_a_type_test'(Term, V1, T1), !,
    writeT_compound(Stream, type(V1, '!'(T1))).
*/
writeT_compound(Stream, Term) :-
    functor(Term, F, _),
    var(F),
    !,
    Term =.. [_|Args],
    writeTerm(Stream, F),
    write(Stream, '('),
    writeT_args(Stream, Args, ', '),
    write(Stream, ')').
writeT_compound(Stream, F(X)) :-
    X == '$none_',
    !,
    writeTerm(Stream, F),
    write(Stream, '()').

writeT_compound(Stream, '$$var$$'(T)) :-
    !,
    write(Stream, T).
writeT_compound(Stream, remember(T)) :-
    !,
    write(Stream, 'remember '),
    writeTerm(Stream, T).
writeT_compound(Stream, remember_for(G, T)) :-
    !,
    write(Stream, 'remember '),
    writeTerm(Stream, G),
    write(Stream, ' for '),
    writeTerm(Stream, T).
writeT_compound(Stream, rememberA(T)) :-
    !,
    write(Stream, 'rememberR '),
    writeTerm(Stream, T).
writeT_compound(Stream, rememberA_for(G, T)) :-
    !,
    write(Stream, 'rememberR '),
    writeTerm(Stream, G),
    write(Stream, ' for '),
    writeTerm(Stream, T).
writeT_compound(Stream, forget(T)) :-
    !,
    write(Stream, 'forget '),
    writeTerm(Stream, T).
writeT_compound(Stream, forget_after(G, T)) :-
    !,
    write(Stream, 'forget '),
    writeTerm(Stream, G),
    write(Stream, ' after '),
    writeTerm(Stream, T).
writeT_compound(Stream, replace_by(T1, T2)) :-
    !,
    write(Stream, 'replace '),
    writeTerm(Stream, T1),
    write(Stream, ' by '),
    writeTerm(Stream, T2).

writeT_compound(Stream, call(T)) :-
    !,
    write(Stream, 'call '),
    writeTerm(Stream, T).
writeT_compound(Stream, do(T)) :-
    !,
    write(Stream, 'do '),
    writeTerm(Stream, T).
writeT_compound(Stream, '@..'(T1, T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' @.. '),
    writeTerm(Stream, T2).
writeT_compound(Stream, to(T1, T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' to '),
    writeTerm(Stream, T2).
writeT_compound(Stream, from(T1, T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' from '),
    writeTerm(Stream, T2).
writeT_compound(Stream, fork_as(T1, T2)) :-
    !,
    write(Stream, 'fork '),
    writeTerm(Stream, T1),
    write(Stream, ' as '),
    writeTerm(Stream, T2).
writeT_compound(Stream, (type T)) :-
    !,
    write(Stream, 'type '),
    writeTerm(Stream, T).
writeT_compound(Stream, (T1 & T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' & '),
    writeTerm(Stream, T2).
writeT_compound(Stream, (T1 ; T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' ; '),
    writeTerm(Stream, T2).
writeT_compound(Stream, (T1 , T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ', '),
    writeTerm(Stream, T2).
writeT_compound(Stream, (T1 .. T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' .. '),
    writeTerm(Stream, T2).
writeT_compound(Stream, ((T0 -> T1) -> T2)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (T0 -> T1)),
    write(Stream, ')'),
    write(Stream, ' -> '),
    writeTerm(Stream, T2).
writeT_compound(Stream, (T1 -> T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' -> '),
    writeTerm(Stream, T2).
writeT_compound(Stream, (T1 | T2)) :-
    !,
    writeTerm(Stream, T1),
    write(Stream, ' | '),
    writeTerm(Stream, T2).
writeT_compound(Stream, (A ::= B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' ::= '),
    writeTerm(Stream, B).
writeT_compound(Stream, global_int_decl(A,B)) :-
    !,
    write(Stream, 'int '),
    writeTerm(Stream, A),
    write(Stream, ' := '),
    writeTerm(Stream, B).
writeT_compound(Stream, global_num_decl(A,B)) :-
    !,
    write(Stream, 'num '),
    writeTerm(Stream, A),
    write(Stream, ' := '),
    writeTerm(Stream, B).    
writeT_compound(Stream, A > B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' > '),
    writeTerm(Stream, B).
writeT_compound(Stream, A >= B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' >= '),
    writeTerm(Stream, B).
writeT_compound(Stream, A < B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' < '),
    writeTerm(Stream, B).
writeT_compound(Stream, A =< B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' =< '),
    writeTerm(Stream, B).
writeT_compound(Stream, A = B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' = '),
    writeTerm(Stream, B).
writeT_compound(Stream, A \= B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' \\= '),
    writeTerm(Stream, B).
writeT_compound(Stream, A @> B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' @> '),
    writeTerm(Stream, B).
writeT_compound(Stream, A @>= B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, '@ >= '),
    writeTerm(Stream, B).
writeT_compound(Stream, A @< B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' @< '),
    writeTerm(Stream, B).
writeT_compound(Stream, A @=< B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' @=< '),
    writeTerm(Stream, B).
writeT_compound(Stream, A @= B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' @= '),
    writeTerm(Stream, B).
writeT_compound(Stream, A '=?' B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' =? '),
    writeTerm(Stream, B).
writeT_compound(Stream, A $= B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' $= '),
    writeTerm(Stream, B).
writeT_compound(Stream, A in B) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' in '),
    writeTerm(Stream, B).
writeT_compound(Stream, '$list_constr'(A, B)) :-
    !,
    write(Stream, '['),
    writeTerm(Stream, A),
    write(Stream, ' :: '),
    writeTerm(Stream, B),
    write(Stream, ']').
writeT_compound(Stream, '$set_constr'(A, B)) :-
    !,
    write(Stream, '{'),
    writeTerm(Stream, A),
    write(Stream, ' :: '),
    writeTerm(Stream, B),
    write(Stream, '}').
writeT_compound(Stream, start_agent(A,B,C)) :-
    !,
    write(Stream, start_agent),
    write(Stream, ' '),
    writeTerm(Stream, A),
    write(Stream, ' '),
    writeTerm(Stream, B),
    write(Stream, ' '),
    writeTerm(Stream, C).
writeT_compound(Stream, start_drive(A,B)) :-
    !,
    write(Stream, start_drive),
    write(Stream, ' '),
    writeTerm(Stream, A),
    write(Stream, ' '),
    writeTerm(Stream, B).
writeT_compound(Stream, start_task(A,B)) :-
    !,
    write(Stream, start_task),
    write(Stream, ' '),
    writeTerm(Stream, A),
    write(Stream, ' '),
    writeTerm(Stream, B).
writeT_compound(Stream, kill_task(A)) :-
    !,
    write(Stream, kill_task),
    write(Stream, ' '),
    writeTerm(Stream, A).

/*
writeT_compound(Stream, '$at_forall'(V, G1, G2)) :-
    !,
    write(Stream, '@ ('),
    writeT_varlist(Stream, V),
    write(Stream, ') ('),
    writeTerm(Stream, G1),
    write(Stream, ' => '),
    writeTerm(Stream, G2),
    write(Stream, ')').
writeT_compound(Stream, '$at_forall_actions'(V, G1, G2)) :-
    !,
    write(Stream, '@ '),
    writeT_varlist(Stream, V),
    write(Stream, ' ('),
    writeTerm(Stream, G1),
    write(Stream, ' ~>> '),
    writeTerm(Stream, G2),
    write(Stream, ')').
*/
writeT_compound(Stream, '$exists'(V, G)) :-
    !,
    write(Stream, 'exists '),
    writeT_varlist(Stream, V),
    write(Stream, ' '),
    writeTerm(Stream, G).
writeT_compound(Stream, '$forall'(V, G1, G2)) :-
    !,
    write(Stream, 'forall '),
    writeT_varlist(Stream, V),
    write(Stream, ' ('),
    writeTerm(Stream, G1),
    write(Stream, ' => '),
    writeTerm(Stream, G2),
    write(Stream, ')').
writeT_compound(Stream, '$forall_actions'(V, G1, G2)) :-
    !,
    write(Stream, 'forall '),
    writeT_varlist(Stream, V),
    write(Stream, ' ('),
    writeTerm(Stream, G1),
    write(Stream, ' ~>> '),
    writeTerm(Stream, G2),
    write(Stream, ')').

writeT_compound(Stream, '$type_macro'(Type)) :-
    !,
    writeTerm(Stream, Type).
writeT_compound(Stream, '$union_type'(Args)) :-
    !,
    writeT_args(Stream, Args, ' || ').
writeT_compound(Stream, '$enum_type'([A])) :-
    !,
    writeTerm(Stream, A), write(Stream, ' | ').
writeT_compound(Stream, '$enum_type'(Args)) :-
    !,
    writeT_args(Stream, Args, ' | ').
writeT_compound(Stream, '$constr_enum_type'(Args)) :-
    !,
    writeT_args(Stream, Args, ' | ').
writeT_compound(Stream, '@'(A, B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' @ '),
    writeTerm(Stream, B).
writeT_compound(Stream, ':='(A, B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' := '),
    writeTerm(Stream, B).
writeT_compound(Stream, '+:='(A, B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' +:= '),
    writeTerm(Stream, B).
writeT_compound(Stream, '-:='(A, B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' -:= '),
    writeTerm(Stream, B).
writeT_compound(Stream, '$'(A)) :-
    !,
    write(Stream, '$'),
    writeTerm(Stream, A).
writeT_compound(Stream, (A <= B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' <= '),
    writeTerm(Stream, B).
writeT_compound(Stream, '<='(A)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' <= ').
writeT_compound(Stream, (A ~>> B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' ~>> '),
    writeTerm(Stream, B).
writeT_compound(Stream, '~>>'(A)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' ~>> ').
writeT_compound(Stream, '~>'(A)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' ~> ').
writeT_compound(Stream, (A :: B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' :: '),
    writeTerm(Stream, B).   
writeT_compound(Stream, (::(A))) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' ::').
writeT_compound(Stream, '$set'(A, B)) :-
    !,
    write(Stream, '{'),
    writeTerm(Stream, A),
    write(Stream, ' :: '),
    writeTerm(Stream, B),
    write(Stream, '}').
writeT_compound(Stream, (A .. B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, A),
    write(Stream, ' .. '),
    writeTerm(Stream, B),
    write(Stream, ')').
writeT_compound(Stream, '$set_type'(A)) :-
    !,
    write(Stream, '{'),
    writeTerm(Stream, A),
    write(Stream, '}').
writeT_compound(Stream, '$list_type'(A)) :-
    !,
    write(Stream, '['),
    writeTerm(Stream, A),
    write(Stream, ']').
writeT_compound(Stream, '$set2list'(A)) :-
    !,
    write(Stream, '[]('),
    writeTerm(Stream, A),
    write(Stream, ')').
writeT_compound(Stream, '$list2set'(A)) :-
    !,
    write(Stream, '{}('),
    writeTerm(Stream, A),
    write(Stream, ')').
writeT_compound(Stream, (# A)) :-
    !,
    write(Stream, '#'),
    writeTerm(Stream, A).
writeT_compound(Stream, (A <> B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' <> '),
    writeTerm(Stream, B).
writeT_compound(Stream, (A ++ B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' ++ '),
    writeTerm(Stream, B).
writeT_compound(Stream, -A) :-
    !,
    write(Stream, '-'),
    writeT_wrap_arith(Stream, A).
writeT_compound(Stream, ?A) :-
    !,
    write(Stream, '?'),
    writeAnnotatedTerm(Stream, A).
writeT_compound(Stream, '!'(A)) :-
    !,
    write(Stream, '!'),
    writeAnnotatedTerm(Stream, A).
writeT_compound(Stream, '??'(A)) :-
    !,
    write(Stream, '??'),
    writeTerm(Stream, A).
writeT_compound(Stream, '+?'(A)) :-
    !,
    writeAnnotatedTerm(Stream, A),
    write(Stream, '?').
writeT_compound(Stream, +A) :-
    !,
    write(Stream, '+'),
    writeT_wrap_arith(Stream, A).
writeT_compound(Stream, (A <=)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' <=').

writeT_compound(Stream, inter(A, B)) :-
    !,
    writeT_wrap_sets(Stream, A),
    write(Stream, ' inter '),
    writeT_wrap_sets(Stream, B).
writeT_compound(Stream, union(A, B)) :-
    !,
    writeT_union_diff1(Stream, A),
    write(Stream, ' union '),
    writeT_union_diff2(Stream, B).
writeT_compound(Stream, diff(A, B)) :-
    !,
    writeT_union_diff1(Stream, A),
    write(Stream, ' diff '),
    writeT_union_diff2(Stream, B).
    
writeT_compound(Stream, (A ** B)) :-
    !,
    writeT_wrap_arith(Stream, A),
    write(Stream, ' ** '),
    writeT_wrap_arith(Stream, B).
writeT_compound(Stream, (A ? B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' ? '),
    (
      '$has_operator'(B)
    ->
      write(Stream, '('),
      writeTerm(Stream, B),
      write(Stream, ')')
    ;
      writeTerm(Stream, B)
    ).
writeT_compound(Stream, (A : B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' : '),
    writeTerm(Stream, B).
writeT_compound(Stream, (A * B)) :-
    !,
    writeT_mult_div1(Stream, A),
    write(Stream, ' * '),
    writeT_mult_div2(Stream, B).
writeT_compound(Stream, (A / B)) :-
    !,
    writeT_mult_div1(Stream, A),
    write(Stream, ' / '),
    writeT_mult_div2(Stream, B).
writeT_compound(Stream, (A + B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' + '),
    writeTerm(Stream, B).
writeT_compound(Stream, (A - B)) :-
    !,
    writeTerm(Stream, A),
    write(Stream, ' - '),
    writeT_minus2(Stream, B).
writeT_compound(Stream, A) :-
    A = '$tuple'([]),
    !,
    write(Stream, '()').
writeT_compound(Stream, A) :-
    A = '$tuple'(Args),
    !,
    write(Stream, '('),
    writeT_args(Stream, Args, ', '),
    write(Stream, ')').
writeT_compound(Stream, A) :-
    A = '$tuple_type'([]),
    !,
    write(Stream, '()').
writeT_compound(Stream, A) :-
    A = '$tuple_type'(Args),
    !,
    write(Stream, '('),
    writeT_args(Stream, Args, ', '),
    write(Stream, ')').
writeT_compound(Stream, A) :-
    A = alt_types(Args),
    !,
    writeT_args(Stream, Args, ' | ').
writeT_compound(Stream, '$tr_prog'(Head, Rules)) :-
    !,
    writeTerm(Stream, Head),
    write(Stream, ' {'), nl(Stream),
    write_tr_rules(Stream, Rules),
    write(Stream, '    }'), nl(Stream).
writeT_compound(Stream, Rule) :-
    Rule = '~>'(_, _), !,
    write_a_tr_rule(Stream, Rule).
writeT_compound(Stream, durative(D)) :-
    !,
    write(Stream, 'durative '),
    (
     list(D)
    ->
     writeT_args(Stream, D, ', ')
    ;
     writeTerm(Stream, D)
    ).
writeT_compound(Stream, discrete(D)) :-
    !,
    write(Stream, 'discrete '),
    (
     list(D)
    ->
     writeT_args(Stream, D, ', ')
    ;
     writeTerm(Stream, D)
    ).
writeT_compound(Stream, resources(D)) :-
    !,
    write(Stream, 'resources '),
    (
     list(D)
    ->
     writeT_args(Stream, D, ', ')
    ;
     writeTerm(Stream, D)
    ).
writeT_compound(Stream, task_start(D)) :-
    !,
    write(Stream, 'task_start '),
    (
     list(D)
    ->
     writeT_args(Stream, D, ', ')
    ;
     writeTerm(Stream, D)
    ).
writeT_compound(Stream, task_atomic(D)) :-
    !,
     writeT_args(Stream, D, ', ').
writeT_compound(Stream, '`'(A)) :-
    !,
    write(Stream, '`'),
    writeTerm(Stream, A).

writeT_compound(Stream, A) :-
    A =.. [F|Args],
    writeTerm(Stream, F),
    write(Stream, '('),
    writeT_args(Stream, Args, ', '),
    write(Stream, ')').

write_tr_rules(_Stream, []).
write_tr_rules(Stream, [R|Rs]) :-
    tab(Stream, 4),
    write_a_tr_rule(Stream, R),
    nl(Stream),
    write_tr_rules(Stream, Rs).

write_a_tr_rule(Stream, '~>'(LHS, RHS)) :-
    write_a_tr_LHS(Stream, LHS),
    write(Stream, ' ~> '),
    nl(Stream),
    tab(Stream, 11),
    write_a_tr_RHS(Stream, RHS).

write_a_tr_LHS(Stream, '~^'(G, G1)) :-
    !,
    writeTerm(Stream, G),
    write(Stream, ' ~^ '),
    writeTerm(Stream, G1).
write_a_tr_LHS(Stream, '$while_until'(OP, G, '$time'(OPT, G1,T))) :-
    !,
    writeTerm(Stream, G),
    write_term_list(Stream, [' ', OP, ' ']),
    writeTerm(Stream, G1),
    write_term_list(Stream, [' ', OPT, ' ']),
    writeTerm(Stream, T).
write_a_tr_LHS(Stream, '$while_until'(OP, G, '$time'(OPT, T))) :-
    !,
    writeTerm(Stream, G),
    write_term_list(Stream, [' ', OP, ' ', OPT, ' ']),
    writeTerm(Stream, T).
write_a_tr_LHS(Stream, '$while_until'(OP, G, G1)) :-
    !,
    writeTerm(Stream, G),
    write_term_list(Stream, [' ', OP, ' ']),
    writeTerm(Stream, G1).
write_a_tr_LHS(Stream, '$time'(OP, G, T)) :-
    !,
    writeTerm(Stream, G),
    write_term_list(Stream, [' ', OP, ' ']),
    writeTerm(Stream, T).
write_a_tr_LHS(Stream, '$while'(OP, G, '$time'(OPT, W, T))) :-
    !,
    writeTerm(Stream, G),
    write_term_list(Stream, [' ', OP, ' ']),
    writeTerm(Stream, W),
    write_term_list(Stream, [' ', OPT, ' ']),
    writeTerm(Stream, T).
write_a_tr_LHS(Stream, '$while'(OP, G, '$time'(OPT, T))) :-
    !,
    writeTerm(Stream, G),
    write_term_list(Stream, [' ', OP, ' ', OPT, ' ']),
    writeTerm(Stream, T).
write_a_tr_LHS(Stream, '$while'(OP, G, A)) :-
    !,
    writeTerm(Stream, G),
    write_term_list(Stream, [' ', OP, ' ']),
    writeTerm(Stream, A).
write_a_tr_LHS(Stream, '$until'(OP, G, A)) :-
    !,
    write_a_tr_LHS(Stream, G),
    write_term_list(Stream, [' ', OP, ' ']),
    write_a_tr_until(Stream, A).
write_a_tr_LHS(Stream, G) :-
    !,
    writeTerm(Stream, G).


write_a_tr_until(Stream, '$time'(OP, U, T)) :-
    !,
    writeTerm(Stream, U), 
    write_term_list(Stream, [' ', OP, ' ']),
    writeTerm(Stream, T).
write_a_tr_until(Stream, U) :-
    writeTerm(Stream, U).

write_a_tr_RHS(Stream, '++'(A,B)) :-
    !,
    write_a_tr_action(Stream, A),
     write(Stream, ' ++ '),
    writeTerm(Stream, B).
write_a_tr_RHS(Stream, '$wait_repeat'(A,B)) :-
    !,
    write_a_tr_action(Stream, A),
     write(Stream, ' wait '),
    write_wait_repeat(Stream, B).
write_a_tr_RHS(Stream, A) :-
    write_a_tr_action(Stream, A).

write_wait_repeat(Stream, '$repeat'(OP, W, R)) :-
    !,
    writeTerm(Stream, W),
    write_term_list(Stream, [' ', OP, ' ']),
    writeTerm(Stream, R).
write_wait_repeat(Stream, W) :-
        !,
    writeTerm(Stream, W).

    
write_a_tr_action(Stream, timed_seq(TS)) :-
    !,
    write_a_tr_timed_seq(Stream, TS).
write_a_tr_action(_Stream, []).
write_a_tr_action(Stream, [A|As]) :-
    writeTerm(Stream, A),
    (
      As = []
    ->
      true
    ;
      write(Stream, ', ')
    ),
    write_a_tr_action(Stream, As).

write_a_tr_timed_seq(_Stream, []).
write_a_tr_timed_seq(Stream, [A|As]) :-
    (
      A = '$time'(OP, A1, T)
    ->
      write_a_tr_action(Stream, A1),
      write_term_list(Stream, [' ', OP, ' ']),
      writeTerm(Stream, T)
    ;
      write_a_tr_action(Stream, A)
    ),
    (
      As = []
    ->
      true
    ;
      write(Stream, ' ; ')
    ),
    write_a_tr_timed_seq(Stream, As).
    
writeT_args(Stream, [A], _Op) :-
    !,
    writeTerm(Stream, A).
writeT_args(Stream, [A|T], Op) :-
    !,
    writeTerm(Stream, A),
    write(Stream, Op),
    writeT_args(Stream, T, Op).
    
writeAnnotatedTerm(Stream, A) :-
    var(A), !,
    writeTerm(Stream, A).
writeAnnotatedTerm(Stream, A) :-
    compound(A),
    ( A = '<='(_) ; A = (_ -> _)), !,
    write(Stream, '('),
    writeTerm(Stream, A),
    write(Stream, ')').
writeAnnotatedTerm(Stream, A) :-
    writeTerm(Stream, A).

writeT_mult_div1(Stream, (A+B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A+B)),
    write(Stream, ')').
writeT_mult_div1(Stream, (A-B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A-B)),
    write(Stream, ')').
writeT_mult_div1(Stream, A) :-
    writeTerm(Stream, A).

writeT_mult_div2(Stream, (A+B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A+B)),
    write(Stream, ')').
writeT_mult_div2(Stream, (A-B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A-B)),
    write(Stream, ')').
writeT_mult_div2(Stream, (A*B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A*B)),
    write(Stream, ')').
writeT_mult_div2(Stream, (A/B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A/B)),
    write(Stream, ')').
writeT_mult_div2(Stream, A) :-
    writeTerm(Stream, A).
    
writeT_minus2(Stream, A+B) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A+B)),
    write(Stream, ')').
writeT_minus2(Stream, A-B) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, (A-B)),
    write(Stream, ')').
writeT_minus2(Stream, A) :-
    writeTerm(Stream, A).

writeT_wrap_arith(Stream, A) :-
    arith(A),
    !,
    write(Stream, '('),
    writeTerm(Stream, A),
    write(Stream, ')').
writeT_wrap_arith(Stream, A) :-
    writeTerm(Stream, A).

arith((_ + _)).
arith((_ - _)).
arith((_ * _)).
arith((_ / _)).
arith((_ ** _)).

writeT_wrap_sets(Stream, A) :-
    setop(A), !,
    write(Stream, '('),
    writeTerm(Stream, A),
    write(Stream, ')').
writeT_wrap_sets(Stream, A) :-
    writeTerm(Stream, A).
    
%setop(inter(_, _)).
setop(union(_, _)).
setop(diff(_, _)).

/*writeT_union_diff1(Stream, union(A, B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, A),
    write(Stream, ' union '),
    writeTerm(Stream, B),
    write(Stream, ')').
writeT_union_diff1(Stream, diff(A,B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, A),
    write(Stream, ' diff '),
    writeTerm(Stream, B),
    write(Stream, ')').
*/
writeT_union_diff1(Stream, A) :-
    writeTerm(Stream, A).
writeT_union_diff2(Stream, diff(A,B)) :-
    !,
    write(Stream, '('),
    writeTerm(Stream, A),
    write(Stream, ' diff '),
    writeTerm(Stream, B),
    write(Stream, ')').
writeT_union_diff2(Stream, A) :-
    writeTerm(Stream, A).



writeT_varlist(Stream, [H1, H2|T]) :-
    !,
    writeTerm(Stream, H1),
    write(Stream, ', '),
    writeT_varlist(Stream, [H2|T]).
writeT_varlist(Stream, [H1]) :-
    writeTerm(Stream, H1).

writeL(List) :-
    writeL(stdout, List).
writeL(_Stream, []).
writeL(Stream, [T|Rest]) :-
    var(T), !,
    write(Stream, T),
    writeL(Stream, Rest).
writeL(Stream, [nl_|Rest]) :-
    !,
    nl(Stream),
    writeL(Stream, Rest).
writeL(Stream, [sp_(N)|Rest]) :-
    !,
    tab(Stream, N),
    writeL(Stream, Rest).
writeL(Stream, [T|Rest]) :-
    atom(T), !,
    writeTerm(Stream, T),
    writeL(Stream, Rest).
writeL(Stream, [T|Rest]) :-
    string(T), !,
    write_string(Stream, T),
    writeL(Stream, Rest).
writeL(Stream, [q_(T)|Rest]) :-
    !,
    write(Stream, T),
    writeL(Stream, Rest).
writeL(Stream, [uq_(T)|Rest]) :-
    !,
    write(Stream, T),
    writeL(Stream, Rest).
/*
writeL(Stream, [q(T)|Rest]) :-
    !,
    write(Stream, T),
    writeL(Stream, Rest).
*/
writeL(Stream, [wr_(T)|Rest]) :-
    !,
    name_vars(T),
    writeTerm(Stream, T),
    writeL(Stream, Rest).
writeL(Stream, [show_(T)|Rest]) :-
    !,
    show_clause(T, Stream),
    writeL(Stream, Rest).
writeL(Stream, [X|Rest]) :-
    writeTerm(Stream, X),
    writeL(Stream, Rest).

