%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.

?- dynamic('$running_teleo'/0).

?- compile_time_only(consult(op_decls)).
?- dynamic('$current_actions'/1).
%?- op(1023, xfx,  '~~' ).
%?- op(600, fx,  'display' ).
%?- op(500, fy, '++').
%?- op(500, fy, '--').

/*
display([]) :- nl, !.
display([nl|Items]):- nl, !, display(Items).
display([Item|Items]) :- !, write(Item),tab(1),display(Items).
display(Item) :- write(Item),nl.
*/


%% atomic(TR, Resource)
?- dynamic(task_atomic/2).
%% has_priority(TR1, TR2) TR1 has priority over TR2 
%%?- dynamic(has_priority/2).

%% discreet actions
?- dynamic('$discrete_action'/1).

?- dynamic('__resources__'/3).
?- dynamic('__running__'/2).
?- dynamic('__waiting__'/2).
?- dynamic('__new_percepts__'/0).
?- dynamic('__time__'/2).
%?- dynamic(resources_hook/0).
%?- dynamic(task_atomic/1).
?- dynamic('$delayed_tr_call'/1).
?- dynamic('$agent_info'/3).
?- dynamic(overlapping_resources_/2).

dump_tr(TR, AT, StartIndex, Now, InTerms, OutTerms, InDepend, Stamps, InTime, OutDelay, OldAction, Actions, InResources) :-
    clause(tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms, InDepend, Stamps,
                   InTime, OutDelay, OldAction, Actions,
		   InResources),
           Body),
    writeR((tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms, InDepend, Stamps,
		    InTime, OutDelay, OldAction, Actions, 
		    InResources) :-
           Body)),write('.'),
    nl,
    fail.

dump_overlap(TR, Changed, CurrentIndex, I) :-
    clause(overlap_at(TR, Changed, CurrentIndex, I), Body),
    writeR((overlap_at(TR, Changed, CurrentIndex, I) :- Body)),write('.'),
    nl,
    fail.
dump_retry(TR_I, TR, RID, Now, Changed, I) :-
    clause(retry_from(TR_I, TR, RID, Now, Changed, I), Body),
    writeR((retry_from(TR_I, TR, RID, Now, Changed, I) :- Body)),write('.'),
    nl,fail.

all_resources_used(R) :-
    '$task_atomic'(_,_), !,
    R = no.
all_resources_used(yes).

teleor :-
    '$running_teleo', !.
teleor :-
    assert('$running_teleo'),
    '$run_tr_checks',
    global_state_set('$prompt', '| ?~> ').

'__can_run__'(Task, Resources, Id) :-
    (
      '__resources__'(dummy, Task, _)
    ->
      '$replace'('__resources__'(dummy, Task, _),
              '__resources__'(dummy, Task, Resources))
    ;
      remember('__resources__'(dummy, Task, Resources))
    ),
    '__running__'(dummy, Task),   % currently running
    Id == none,  % a new call
    !,
    (
      '__waiting__'(dummy, _)   % is there a waiting task
    ->                         
      retract('__running__'(dummy, Task)), % if so give up Resources
      assert('__waiting__'(dummy, Task)),  % put at back of wait queue
      fail                      % now fail this can_run call
    ;
      all_free_excluding(Resources, Task)
    ->
      true        % else let it continue running
    ;
      retract('__running__'(dummy, Task)), % if so give up Resources
      assert('__waiting__'(dummy, Task)),  % put at back of wait queue
      fail                      % now fail this can_run call
    ).                                

'__can_run__'(Task, _Resources, _Id) :-
    '__running__'(dummy, Task),   % Task already running, continuing atomic call
    !.

'__can_run__'(Task, Resources, _) :-
    remember('__waiting__'(dummy, Task)),
    head_of_wait_queue(Task),  % Task head of wait queue
    all_free(Resources),      % all resource needs are free
    !,   	               % let it become running and succeed call
    retract('__waiting__'(dummy, Task)), 
    assert('__running__'(dummy, Task)).

'__can_run__'(Task, Resources, _) :-
    '__new_percepts__',  % not head of the wait queue, but can do switch test
    %write_term_list(['Checking if ',Task,' can jump the wait queue\n']),
    can_use_resources_now(Task,Resources),
    !,           % let it become running task and succeed call
    retract('__waiting__'(dummy, Task)), 
    assert('__running__'(dummy, Task)).


can_use_resources_now(Task,Resources) :-
    all_free(Resources),       % all resource needs are free
    will_conform_to_queue_etiquette(Task,Resources),
    %write_term_list(['No task in front of ',Task, ' needs a  resource it needs\n']),
    !. 
    % make sure not usurping a resource needed 
    % by a waiting task in front of Task on wait queue

can_use_resources_now(Task,AtPRes) :-
    current_predicate(app_conform_to_queue_etiquette/3),
    all_free(AtPRes),
    findall( T-Res, (before(T,Task),'__resources__'(dummy,T,Res),
                             overlap(Res,AtPRes)), InFront),
    app_conform_to_queue_etiquette(Task,AtPRes,InFront).
/*
    write_term_list(['\nLetting ',Task,
                              ' run even though a task in front\n', 
			      'is waiting for overlapping resources\n']).
*/
before(T,Task) :-
   '__waiting__'(dummy,T),
   (T=Task -> !,  % ! will stop backtracking for more sols to waiting(T)
     fail
     ;
     true).     

will_conform_to_queue_etiquette(Task, Resources) :-
    in_front_of(Task,T),
    '__resources__'(dummy,T,TRes),
    overlap(TRes,Resources),
    !,
    fail.
will_conform_to_queue_etiquette(_Task, _Resources).


head_of_wait_queue(Task) :-
   '__waiting__'(dummy, T), !,
   T=Task.     % Task first waiting task
 
in_front_of(Task,T) :-
   '__waiting__'(dummy, T),
   (
     T=Task
   ->
     !,        % will stop backtracking to get more sols to waiting(T)
     fail
   ;
     true
   ).

overlap(all, _) :- !.
overlap(_, all) :- !.
overlap(Res1,Res2) :-
   member(R1,Res1),
   member(R2,Res2),
   '$are_overlapping_resources'(R1, R2).

'$are_overlapping_resources'(R, R).
'$are_overlapping_resources'(R1, R2) :-
    overlapping_resources_(R1, R2).
'$are_overlapping_resources'(R1, R2) :-
    overlapping_resources_(R2, R1).

/*
  no_overlap(L1, L2) :-
    member(X, L1),
    member(X, L2),
    !,
    fail.
no_overlap(_, _).
*/


all_free(all) :-
    !,
    \+'__running__'(dummy, _).  
all_free(Res) :-
    member(R,Res),
    '__running__'(dummy,T),
    '__resources__'(dummy,T,OtherRes),
    overlap([R], OtherRes),
    !, fail.
all_free(_).


no_overlap([], _) :- !.
no_overlap(_, []) :- !.
no_overlap([X|_], [X|_]) :- !, fail.
no_overlap([X|Xs], Ys) :-
    Ys = [Y|_],
    X @< Y, !,
    no_overlap(Xs, Ys).
no_overlap(Xs, [_Y|Ys]) :-
    no_overlap(Xs, Ys).

%resource_member(_,all) :- !.
%resource_member(R,RList) :- member(R,RList), !.

all_free_excluding(Res, Task) :-
    member(R,Res),
    '__running__'(dummy,T),
    T \= Task,
    '__resources__'(dummy,T,OtherRes),
    overlap([R], OtherRes),
    !, fail.
all_free_excluding(_, _). 

list_back([X], Y) :- !, Y = X. 
list_back([_|Xs], Y) :- list_back(Xs, Y).

ready_to_run :-
    '__running__'(dummy, _), !, fail.
ready_to_run :-
    '__waiting__'(dummy, Task), !,
    thread_symbol(Task).
    
%% merge sorted lists removing repeats

pred_order_merge(Sign1(P1)/N1, Sign2(P2)/N2) :-
    t(P1, N1, Sign1) @< t(P2, N2, Sign2).


merge([], X, X) :- !.
merge(X, [], X) :- !.
merge([X1|R1], [X1|R2], [X1|RM]) :-
    !,
    merge(R1, R2, RM).
merge([X1|R1], [X2|R2], [X1|RM]) :-
    X1 @< X2,
    merge(R1, [X2|R2], RM).
merge([X1|R1], [X2|R2], [X2|RM]) :-
    merge([X1|R1], R2, RM).


% try an until goal: try(Goal) succeeds iff Goal succeeds - no vars are
% instantiated
try(Goal) :-
    try_not(Goal),
    !,
    fail.
try(_).

% try an until goal: try_not(Goal) succeeds iff Goal fails  - no vars are
% instantiated
try_not(Goal) :-
    call(Goal),
    !,fail.
try_not(_).



%% get_delay(Now, Times, Delay) is true
%% if Delay is the smallest element of the list Time bigger than now.
%% -1 if no such time.
get_delay(Now, Times, Delay) :-
    get_min_time(Times, Now, 1e20, Min),
    get_delay_aux(Now, Min, Delay).

get_min_time([], _, M, M).
get_min_time([T|Rest], Now, InMin, Min) :-
    T >= Now, T < InMin, !,
    get_min_time(Rest, Now, T, Min).
get_min_time([_|Rest], Now, InMin, Min) :-
    get_min_time(Rest, Now, InMin, Min).

get_delay_aux(_Now, 1e20, -1) :- !.
get_delay_aux(Now, Timeout, Delay) :- Delay is Timeout - Now.




check_if_should_clean_up(no) :-
    !, clean_up.
check_if_should_clean_up(_).

clean_up :-
    thread_symbol(Task),
    retractall('__running__'(dummy, Task)),
    retractall('__waiting__'(dummy, Task)),
    retractall('__resources__'(dummy, Task,_)).

'$delay_tr_call'(G) :-
    assert('$delayed_tr_call'(G)).


eval(_OldActions, _Stamps, _Delay, TermSeq, _TimeStamp) :-
    log_resources,
    resources_callback,
    TermSeq = [wait_choice(_, _)|_],
    all_resources_used(yes),
    !,
    %% so wait until nothing is running and at head of wait queue
    thread_wait_on_goal(ready_to_run, [wait_preds(['__running__'/2])]),
    list_back(TermSeq, rule_choice(TR, _, _, _, _, _)),
    start_task(TR).
%% No delays
eval(OldActions, Stamps, Delay, TermSeq, TimeStamp) :-
    Delay < 0,
    !,
    thread_wait_on_goal(update_call_chain(TermSeq, Delay1, NewTermSeq,
                                          OldActions, NewActions,
                                          Stamps, NewStamps, [], NewTimeStamp),
                        [wait_preds(TimeStamp)]),
    !, % eliminate choice points
    eval(NewActions,  NewStamps, Delay1, NewTermSeq, NewTimeStamp).
%% A delay
eval(OldActions, Stamps, Delay, TermSeq, TimeStamp) :-
    !,
    eval_delays(OldActions, Stamps, Delay, TermSeq, TimeStamp).

eval_delays(OldActions, Stamps, Delay, TermSeq, TimeStamp) :-
    thread_wait_on_goal(update_call_chain(TermSeq, Delay1, NewTermSeq,
                                          OldActions, NewActions,
                                          Stamps, NewStamps, [], NewTimeStamp),
                        [wait_preds(TimeStamp), wait_for(Delay)]),
    !, % eliminate choice points
    eval(NewActions,  NewStamps, Delay1, NewTermSeq, NewTimeStamp).
eval_delays(OldActions, Stamps, _Delay, TermSeq, _TimeStamp) :-
    %% timeout occurred
    %%errornl('timeout occurred'),
    thread_atomic_goal(update_eval_info(TermSeq, Delay, NewTermSeq,
                                        OldActions, NewActions,
                                        Stamps, NewStamps, NewTimeStamp)),
    !, % eliminate choice points
    eval(NewActions,  NewStamps, Delay, NewTermSeq, NewTimeStamp).


update_eval_info(TermSeq, Delay, NewTermSeq,
                 OldActions, NewActions, Stamps, NewStamps, NewTimeStamp) :-
    update_call_chain(TermSeq, Delay, NewTermSeq,
                      OldActions, NewActions, Stamps, NewStamps, _,
                      NewTimeStamp),
    !.
update_eval_info(TermSeq, Delay, NewTermSeq,
                 OldActions, NewActions, Stamps, NewStamps, NewTimeStamp) :-
    get_delay_from_stack(TermSeq, Delay),
    update_predicate_timestamps(Stamps, NewStamps),
    %%errornl(update_predicate_timestamps(Stamps, NewStamps)),
    get_predicate_timestamps([+'__time__'/2], NewTimeStamp),
    NewTermSeq = TermSeq,
    NewActions = OldActions.

    
%% TermSeq is the previous call stack.
%% Changed is the list of annotated preds that have changed.
%% If a new call stack is constructed then this succeeds
%% instantiating the other args

'$update_stamps_get_changed'(Stamps, Changed) :-
    update_predicate_timestamps(Stamps, NewStamps),
    diff_list(NewStamps, Stamps, Diff),
    map('$strip_star', Diff, Changed).

'$strip_star'((A * _), A).

'$display_changed_preds'(Changed) :-
    write('Changed Predicates: '),
    '$display_changed_preds_aux'(Changed),
    nl.
'$display_changed_preds_aux'([]).
'$display_changed_preds_aux'([(+P/_N)|Rest]) :-
    !,
    write_term_list(['++',P, '  ']),
    '$display_changed_preds_aux'(Rest).
'$display_changed_preds_aux'([(-P/_N)|Rest]) :-
    !,
    write_term_list(['--', P, '  ']),
    '$display_changed_preds_aux'(Rest).

update_call_chain(TermSeq, Delay, NewTermSeq,
		  OldActions, NewActions, Stamps, NewStamps, CheckChanged,
                  NewTimeStamp) :-
    '$update_stamps_get_changed'(Stamps, Changed),
    %%'$display_changed_preds'(Changed),
    %%errornl(check_changed(CheckChanged, Changed, Delay)),
    Changed \== CheckChanged,   %%%% no changes affect 'watched' preds
    gettimeofday(Now), %% fix the current time for the computation
    thread_atomic_goal(
		       update_call_chain_aux(TermSeq, NewTermSeq,
					     OldActions, NewActions,
					     Depends, Now, Delay, Changed)),
    %%errornl(got_here_update_call_chain(Delay)),
    get_predicate_timestamps(Depends, NewStamps),
    get_predicate_timestamps([+'__time__'/2], NewTimeStamp),
    '$update_chain_call_goals'.

 '$update_chain_call_goals':-
    '$delayed_tr_call'(_), !,
    forall(retract('$delayed_tr_call'(G)), '$call_tr_call'(G)),
    '$update_time_fact'.
'$update_chain_call_goals'.


'$call_tr_call'(G) :- call(G), !.
'$call_tr_call'(G) :- writeL(stderr, ["The call ", G, " failed", nl_]).
    
%% special case - only new_percepts changed which implies this is a
%% waiting task and we only need to test the wait entry as nothing
%% else can change below this point in the stack
%% Keith: as discussed this is a minor optimization that we will keep
/*
  update_call_chain_aux([Term|TermSeq], NewTermSeq,
		      OldActions, NewActions, NewStamps, Now, Delay, Changed) :-
    Term = wait_choice(TR, Depend),
    Changed = [+'__new_percepts__'/0],
    !,
    tr_rule(TR, wait, 1, Now, TermSeq, NewTermSeq, Depend, NewStamps, [], Delay,
            OldActions, NewActions, no).
*/
%% normal case.
%% NewRule will be instantiated to true iff a new stack is built.
update_call_chain_aux(TermSeq, NewTermSeq,
		      OldActions, NewActions, Depends, Now, Delay, Changed) :-
    update_call_chain_aux1(TermSeq, NewTermSeq, OldActions, NewActions,
                           Depends, Now, Delay, Changed, NewRule),
    %errornl(update_call_chain_aux1(TermSeq, NewTermSeq, OldActions, NewActions,
    %                       NewStamps, Now, Delay, Changed, NewRule)),
    nonvar(NewRule),
    !.


%% Step through the stack processing on the way out of recursion
update_call_chain_aux1([], _NewTermSeq, _OldActions, _NewActions,
                           _Depends, _Now, _Delay, _Changed, _NewRule).

update_call_chain_aux1([Term|TermSeq], NewTermSeq, OldActions, NewActions,
                           Depends, Now, Delay, Changed, NewRule) :-
    update_call_chain_aux1(TermSeq, NewTermSeq, OldActions, NewActions,
                           Depends, Now, Delay, Changed, NewRule),
    update_call_chain_aux_term(Term, TermSeq, NewTermSeq,
                               OldActions, NewActions,
                               Depends, Now, Delay, Changed, NewRule).

%% A change has been found further down the stack and so the stack
%% has been recalculated already so don't test any more
%% on the way out of recursion
update_call_chain_aux_term(_, _, _, _, _, _, _, _, _, NewRule) :-
    nonvar(NewRule), !.

%% No change found so far so see if a change happens at this stack entry

%% The entry is a timed sequence - has the timeout expired?
%% As before
update_call_chain_aux_term(timed_seq(Timeout, S, All, Depend,
                                     Time, InResources),
                           TermSeq, NewTermSeq, Action, NewActions,
                           OutDepends, Now, Delay, _Changed, NewRule) :-
    Timeout < Now,
    !,
    %% timeout expired - 
    %% get the next element in the timed sequence
    update_timed_seq(S, All, S1), 
    S1 = [TR:T|_],
    update_call_chain_TS(T, Now, S1, All, Depend, Time, OutTSTimes,
                         InResources, TermSeq, TermsTS),
    update_call_chain_aux2(TR, TermsTS, NewTermSeq,
                           Depend, OutDepends, OutTSTimes, Now, Delay,
                           Action, NewActions, InResources),
    NewRule = true.
%% The entry is at a wait_choice - this is only fired
%% if __new_percepts__ has just been asserted and there are other changes
%% to percepts - this could mean a different rule of TR should be chosen
%% As before
update_call_chain_aux_term(wait_choice(TR, Depend), TermSeq, NewTermSeq,
                           Action, NewActions,
                           OutDepends, Now, Delay, _Changed, NewRule) :-
    %% The update is at a wait_choice
    %% overlap(Changed, [+'__new_percepts__'/0]),
    tr_rule(TR, wait, 1, Now, TermSeq, NewTermSeq, Depend, OutDepends,
            [], Delay, Action, NewActions, no),
    !,
    NewRule = true.

update_call_chain_aux_term(rule_choice(TR, TR_I, RID, Depend,
                                       Time, ResourcesUsed),
                           Rest, NewTermSeq, Action, NewActions,
                           OutDepends, Now, Delay, Changed, NewRule) :-
    retry_from(TR_I, TR, RID, Now, Changed, RetryIndex),
    tr_rule(TR, RID, RetryIndex, Now, Rest, NewTermSeq, Depend,
            OutDepends, Time, Delay,
            Action, NewActions, ResourcesUsed),
    !,
    NewRule = true.

%% This catchall deals with no change for calls on tr_rule as well as
%% no changes for timed sequences and for wait_choice
update_call_chain_aux_term(_, _, _, _, _, _, _, _, _, _) :-
    !.



update_call_chain_TS(-1, _Now, _S1, _All, _Depend, Time, OutTSTimes,
                     _InResources, Rest, TermsTS) :-
    !,
    Time = OutTSTimes,
    TermsTS = Rest.
update_call_chain_TS(T, Now, S1, All, Depend, Time, OutTSTimes,
                     InResources, Rest, TermsTS) :-
    TSNow is T + Now,
    OutTSTimes = [TSNow|Time],
    NewTerm = timed_seq(TSNow, S1, All, Depend, Time, InResources),
    TermsTS = [NewTerm|Rest].


update_call_chain_aux2(TR, InTermSeq, _NewTermSeq,
                       _Depend, _Stamps, _InTime, _Now, _OutDelay,
                       OldAction, _Actions, _InResources) :-
      TR = pa([error]), !,
      throw(error(error_timeout, TR, InTermSeq, OldAction)).
update_call_chain_aux2(TR, InTermSeq, NewTermSeq,
                       Depend, OutDepends, InTime, Now,  OutDelay,
                       OldAction, Actions, _InResources) :-
      TR = pa(Actions), !,
      %% the timed sequence element is a primitive action
      NewTermSeq = InTermSeq,
      OutDepends = Depend,
      do_action(OldAction, Actions, InTermSeq),
      %gettimeofday(Now),
      get_delay(Now, InTime, OutDelay).

update_call_chain_aux2(TR, InTermSeq, NewTermSeq,
                       Depend, OutDepends, InTime, Now, OutDelay,
                       OldAction, Actions, _InResources) :-
      TR = rpa(Actions), !,
      %% the timed sequence element is a primitive action
      NewTermSeq = InTermSeq,
      OutDepends = Depend,
      do_action(OldAction, rpa(Actions), InTermSeq), 
      get_delay(Now, InTime, OutDelay).

      
update_call_chain_aux2(TR, InTermSeq, NewTermSeq,
                       Depend, OutDepends, InTime, Now, Delay,
                       OldAction, Actions, InResources) :-
    %% otherwise a TR so update the call stack
    tr_rule(TR, none, 1, Now, InTermSeq, NewTermSeq, Depend, OutDepends,
            InTime, Delay, OldAction, Actions, 
            InResources).

get_delay_from_stack([rule_choice(_, _, _, _, InTime, _)|_], Delay) :-
    !,
    gettimeofday(Now),
    get_delay(Now, InTime, Delay).
      
get_delay_from_stack([timed_seq(T, _, _, _, _, _),
		      rule_choice(_, _, _, _, InTime, _)|_], Delay) :-
    gettimeofday(Now),
    get_delay(Now, [T|InTime], Delay).
   

'$update_time_fact' :-
    gettimeofday(T),
    (
      '__time__'(dummy,_)
    ->
      retract('__time__'(dummy,_)),
      assert('__time__'(dummy, T))
    ;
      %assert('$start_time'(T)),
      assert('__time__'(dummy, T))
    ),
     '$start_time'(T1),
     Delta is T - T1,
    display_beliefs(Delta),
    log_resources.


%% update the list containing the remaining elements in timed sequence
update_timed_seq(S, All, S1) :-
    S = [_], !, S1 = All.
update_timed_seq(S, _All, S1) :-
    S = [_|S1].

/*
  until_test(G, T, U) :-
    errornl(until_test(G, T, U)),
    call(T), !, call(U).
until_test(G, T, U) :-
    \+call(G), !, fail.
until_test(G, T, U) :-
    call(U).
*/

do_action(OldA, Actions, CallList) :-
    thread_symbol(Task),
    do_action_changes(OldA, Actions, Controls),
    log_action(Task, CallList, Actions, OldA, Controls),
%    write(do_action_changes(OldA, Actions, Controls)),nl,
%    write_term_list(['\nLast actions were: ', OldA]),
%    write_term_list(['\nControl actions for ',Task,' are:',nl]), 
%    write('<<'),write(Controls),
%    nl,
    controls(Controls),
    !.  % to eliminate choice points

new(TR) :-
   compatible_with_existing_tasks(TR),
   !,
   do_task(TR).
new(TR) :-
   write_term_list(stderr, ['\nError: ' , TR, 
       ' is not compatible with existing tasks - ignored', nl]).
new(TR, Name) :-
   compatible_with_existing_tasks(TR),
   !,
   do_task(TR, Name).
new(TR,_) :-
   write_term_list(stderr, ['\nError: ' , TR, 
       ' is not compatible with existing tasks - ignored', nl]).

%% If you don't give the TR program a name it uses the functor as the name
do_task(TR) :-
    functor(TR, TRF, _N),
    (all_resources_used(yes) ; TR = start ; task_start(TR) ),!,
    thread_fork(TName,  start_task(TR), TRF),
    assert(task_(TName,TR)).
do_task(TR) :-
    write_term_list(stderr, ['Error: ' , TR, 
       ' has not been declared as task_start- ignored', nl]).

    
%% You can supply a name (an atom)
%% You can supply a name (an atom)
do_task(TR, Name) :-
    thread_is_thread(Name),
    !,
    write_term_list(stderr, ['Error: For ' , do_task(TR, Name), ' the name is already in use', nl]).
do_task(TR, Name) :-
    functor(TR, _TRF, _N),
    task_start(TR),!,
    thread_fork(Name,  start_task(TR)),
    assert(task_(Name,TR)).

do_task(TR, _) :-
    write_term_list(stderr, ['Error: ' , TR, ' has not been declared as task_initial', nl]).

%% Start TR running
start_task_aux(TR) :-
    catch(start_task_aux2(TR), ErrorTerm, process_error(ErrorTerm)).

start_task_aux2(TR) :-
    thread_atomic_goal(start_task_atomic(TR, Action, OutStamps, Delay,
                                          TermSeq)),
    (
      '$delayed_tr_call'(_)
    ->
      forall(retract('$delayed_tr_call'(G)), '$call_tr_call'(G)),
      '$update_time_fact'
    ;
      true
    ),
    get_predicate_timestamps([+'__time__'/2], TimeStamp),
    eval(Action, OutStamps, Delay, TermSeq, TimeStamp).

start_task_atomic(TR, Action, Stamps, Delay, TermSeq) :-
    gettimeofday(Now),
    '$update_time_fact',
%    log_resources,
    tr_rule(TR, none, 1, Now, [], TermSeq, [], Depends, [],
            Delay, [], Action, no),
    get_predicate_timestamps(Depends, Stamps).
   
process_error(error(no_matching_rule, Prog, OldTermSeq, OldActions)) :-
    !,
    clean_up,
    do_action(OldActions, [], OldTermSeq),
    write_term_list(stderr, ['No matching rule for ', Prog,
                             ' - task terminated.', nl]),
    thread_exit.
process_error(error(error_timeout, _Prog, OldTermSeq, OldActions)) :-
    !,
    clean_up,
    do_action(OldActions, [], OldTermSeq),
    write_term_list(stderr, ['Error timeout ', OldTermSeq,
                             ' - task terminated.', nl]),
    thread_exit.
process_error(ErrorTerm) :-
    !,
    clean_up,
    write_term_list(stderr, ['Unhandled error ',
                             ErrorTerm,
                             ' - task terminated (WARNING actions not stopped).', nl]),
    thread_exit.
    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% DB management
%%

       
                                                      
mode(_).                                    
 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% logging

?- dynamic(do_log/1).

send_log_message(Stream) :-
    logger_name(Name), !,
    stream_to_string(Stream, String),
    String ->> Name.
log_command(T) :-
    do_log(commands),!,
    open_string(write, Stream),
    write_term_list(Stream, T),
    send_log_message(Stream).
log_command(_).

log_percepts(T) :-
    do_log(percepts),!,
    open_string(write, Stream),
    write_term_list(Stream, T),
    send_log_message(Stream).
log_percepts(_).

log_error(T) :-
    do_log(errors),!,
    open_string(write, Stream),
    write_term_list(Stream, T),
    send_log_message(Stream).
log_error(_).

log_resources :-
    thread_symbol(Th), 
    do_log(resources),
    '__resources__'(dummy,Th,_ThRes), !,
    open_string(write, Stream),
    thread_atomic_goal(
      (
% write(Th),write(' needing '),write(ThRes),errornl(' sees:'),
       write(Stream, 'Running tasks:\n'),
       forall( (  '__running__'(dummy, Task),'__resources__'(dummy, Task, Res)),
                  (write(Stream, Task-Res), nl(Stream))),
       write(Stream, 'Waiting tasks:\n'),
       forall( (  '__waiting__'(dummy, Task),'__resources__'(dummy, Task, Res)),
                  (write(Stream, Task-Res), nl(Stream))),
       nl(Stream)
       )
     ),
    send_log_message(Stream).
    
log_resources.

log_action(actions, _Chain, Action,  OldA, Controls) :-
    do_log(actions),!,
    open_string(write, Stream),
    write_term_list(Stream, [nl,'New actions are: ', w(Action), nl]),
    write_term_list(Stream, ['Last actions were: ', OldA, nl]),
    write_term_list(Stream, ['Control actions are:',nl]), 
    write(Stream, '<<'),write(Stream, Controls),
    nl(Stream),
    send_log_message(Stream).
log_action(actions, _D, _Action, _, _) :- !.
log_action(Task, Chain, Action,  OldA, Controls) :-
    do_log(actions),!,
    open_string(write, Stream),
    write_term_list(Stream, [w(Task), w(' : '), nl]),
    log_action_chain(Chain, Action, D, Stream),
    D1 is D+1,
    write_term_list(Stream, [tab(D1), w(Action), nl]),
    write_term_list(Stream, ['\nLast actions were: ', OldA]),
    write_term_list(Stream, ['\nControl actions for ',Task,' are:',nl]), 
    write(Stream, '<<'),write(Stream, Controls),
    nl(Stream),
    send_log_message(Stream).
log_action(_, _D, _Action, _, _).

log_action_chain([], _A, 1, _).
log_action_chain([timed_seq(_, _, A, _, _, _)|Rest], Action, DN, Stream) :-
    !,
    log_action_chain(Rest, Action, D, Stream),
    DN is D+1,
%    D1 is 2*(D+1)-2,
    write_term_list(Stream, [tab(DN), w(A), nl]).
log_action_chain([rule_choice(A, _, rid(I,_,_,_), _, _, _)|Rest],
                 Action, DN, Stream) :-
    !,
    log_action_chain(Rest, Action, D, Stream),
    DN is D+1,
%    D1 is 2*D-1,
    write_term_list(Stream, [tab(DN), w(A-I), ' ', fired,nl]).
log_action_chain([wait_choice(TR, _)|Rest], Action, DN, Stream) :-
     !,
     log_action_chain(Rest, Action, D, Stream),
     DN is D+1,
%     D1 is 2*D-1,
     write_term_list(Stream, [tab(DN), w(wait_choice(TR)), nl]).
   

log_action_chain([_|Rest], Action, D, Stream) :-
    log_action_chain(Rest, Action, D, Stream).

log(commands) :-
    assert(do_log(commands)).
log(percepts) :-
    assert(do_log(percepts)).
log(actions) :-
    assert(do_log(actions)).
log(beliefs) :-
    assert(do_log(beliefs)).
log(errors) :-
    assert(do_log(errors)).
log(resources) :-
    assert(do_log(resources)).
log(tasks) :-
    assert(do_log(tasks)).
log(all) :-
    assert(do_log(commands)),
    assert(do_log(percepts)),
    assert(do_log(actions)),
    assert(do_log(errors)),
    assert(do_log(resources)),
    assert(do_log(beliefs)),
    assert(do_log(tasks)).

logger(Name, _Kind) :-
    retract(logger_name(OldName)) , !,
    writeL(["Changing logger from ", OldName, " to ", Name, nl_]),
    assert(logger_name(Name)).
logger(Name, Kind) :-
    writeL(["Starting logging to ", Name, nl_]),
    assert(logger_name(Name)),
    log(Kind).
unlog :-
    writeL(["Stoping logging",nl_]),
    retractall(do_log(_)),
    retractall(logger_name(_)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Display beliefs

?- dynamic(do_display_beliefs).
?- dynamic(task/2).

display_beliefs(T) :-
    do_log(beliefs), !,
    open_string(write, Stream),
    thread_atomic_goal(display_all_beliefs(T, Stream)),
    send_log_message(Stream).
display_beliefs(_).

display_all_beliefs(T, Stream) :-
    writeL(Stream, [nl_,
                    "------ Current Percepts and Dynamic Beliefs at Time: ", T,
                    " ------",nl_]),
    get_percepts_belief(B),
    writeL(Stream, [B," "]),
    fail.
display_all_beliefs(_, Stream) :-
    nl(Stream),
    forall( task_(T,Call), (nl(Stream),write(Stream, task_(T,Call)))),
    writeL(Stream, [nl_,"---------------------",nl_]).

get_percepts_belief(G) :-
    '$percept_type_info'(_, _, G),
    call(G).

get_percepts_belief(G) :-
    '$belief_type_info'(_, _, G),
    call(G).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Percept management
%%

infer_and_remember :-
   forall((should_be_remembered_(PFact), \+ can_infer_now_(PFact)),
          forget(PFact)),
   forall(can_infer_now_(PFact), remember(PFact)).     
 

do_initialise :-
    '$agent_info'(_, _, Handle),
    repeat,
    initialise_ ->> Handle,
    writeL(["Waiting for initial percepts from ", Handle, nl_]),
    ipc_peek(_, H, [timeout(5)]),
    same_handle(Handle, _:B@Host),
    same_handle(H, _:B@Host), !,
    writeL(["Received initial percepts from ", Handle, nl_]).

start_percept_handler :-
    do_initialise,
    get_percepts_interface(Ps),
    filter(ground, Ps, GPs),
     thread_atomic_goal(( 
			  handle_percepts_interface(GPs),
			  checkBS_interface,
			  infer_and_remember,
			  '$update_time_fact'
                       )),
    remember('__new_percepts__'),
    belief_store_updated ->> messages,
%    display_beliefs,
    thread_yield,
    forget('__new_percepts__'),
    handle_percepts.       

handle_percepts :-
    (
     \+current_predicate(all_resources_used/1)
    ->
      process_percepts_without_resources
    ;
      all_resources_used(no)
    ->
      process_percepts_with_resources
    ;
      process_percepts_without_resources
    ).

get_or_wait_for_percepts(Percepts) :-
    '$get_percepts'(P),
    get_or_wait_for_percepts_aux(P, Percepts).

get_or_wait_for_percepts_aux(P, Percepts) :-
    ipc_peek(_, _, [timeout(poll)]),
    !,
    '$get_percepts'(P1),
    append(P, P1, P2),
    get_or_wait_for_percepts_aux(P2, Percepts).
get_or_wait_for_percepts_aux(P, P).



'$get_percepts'(Ps) :-
    do_log(percepts),!,
    open_string(write, Stream),
    write(Stream, 'Waiting for new percepts\n'),
    send_log_message(Stream),
    Ps1 <<- _ ,
    (
      list(Ps1)
    ->
      Ps = Ps1
    ;
      Ps = [Ps1]
    ),
    open_string(write, StreamP),
    write(StreamP, 'Received new percepts:\n'),
    write(StreamP, Ps), nl(StreamP),
    send_log_message(StreamP).
'$get_percepts'(Ps) :-
    Ps1 <<- _,
    (
      list(Ps1)
    ->
      Ps = Ps1
    ;
      Ps = [Ps1]
    ).

/*
  handle_percepts_(A) :- 
    ground(A),
    !,
    '$handle_percepts_top'(A).
handle_percepts_(_).

'$handle_percepts_top'(A) :-
    current_predicate(handle_percept/1), !,
    '$handle_percepts_'(A).
'$handle_percepts_top'(A) :-
    '$handle_percepts_default'(A).

'$handle_percepts_'([]).
'$handle_percepts_'([P|Rest]) :-
    type(P, percept_message_type), !,
    handle_percept(P),
    '$handle_percepts_'(Rest).
'$handle_percepts_'([_|Rest]) :-
    '$handle_percepts_'(Rest).

'$handle_percepts_default'([]).
'$handle_percepts_default'([P|Ps]) :-
    \+type(P, percept_message_type), !,
    '$handle_percepts_default'(Ps).
'$handle_percepts_default'([r_(P)|Ps]) :-
    !,
    remember(P),
    '$handle_percepts_default'(Ps).
'$handle_percepts_default'([f_(P)|Ps]) :-
    !,
    forget(P),
    '$handle_percepts_default'(Ps).
*/

process_percepts_with_resources :-
    get_percepts_interface(Ps),
    filter(ground, Ps, GPs),
   thread_atomic_goal(process_percepts_with_resources_atomic(GPs)),
    process_percepts_with_resources.

process_percepts_with_resources_atomic(Ps) :-
    thread_atomic_goal((
                        handle_percepts_interface(Ps),
			checkBS_interface,
                        infer_and_remember,
                        '$update_time_fact'
                       )),
    %display_beliefs, 
    %log_resources,
    findall(T, '__running__'(dummy, T), RunTasks),
    (thread_is_thread(message_handler) ->
        %% agent has a message handler let message handler also update the BS 
        %% using any waiting messages before running tasks get a time slice
        schedule_threads_now([message_handler|RunTasks])
    ;
        schedule_threads_now(RunTasks)
    ),
    %% the following assert will allow waiting tasks to switch to running
    %% tasks if they can and make them all runnable threads
    remember('__new_percepts__'),
    %% find current wait queue including any running tasks
    %% that switched to waiting in their extra time slice
    findall(T, '__waiting__'(dummy, T), WaitTaskQueue),
    %% give each waiting task a time slice in wait queue order
    schedule_threads_now(WaitTaskQueue), 
    forget('__new_percepts__').
    %% waiting tasks must now wait until next percepts update


process_percepts_without_resources :-
    get_percepts_interface(Ps),
    filter(ground, Ps, GPs),
    %errornl(percepts_____(Ps)),
    thread_atomic_goal((
                        handle_percepts_interface(GPs),
			checkBS_interface,
                        infer_and_remember,
                        '$update_time_fact'                      
                       )),
    %display_beliefs, 
    %log_resources,
%    thread_atomic_goal(handle_the_percepts(Ps)),
    process_percepts_without_resources.

/*
  resources_callback :-
    resources_hook_, !.
resources_callback.
*/

'$assert_resources_hook_interface' :-
    current_predicate(resources_hook/0), !,
    assert((resources_callback :-  resources_hook)).
'$assert_resources_hook_interface' :-
    assert(resources_callback).

get_active_resources(Resources) :-
    findall(res(Task, Res),
            ('__running__'(dummy, Task),
                '__resources__'(dummy, Task,Res)),
            Resources).

get_waiting_resources(Resources) :-
    findall(res(Task, SRes),
            ('__waiting__'(dummy, Task),
                '__resources__'(dummy, Task,Res), sort(Res, SRes)),
            Resources).

/*    
handle_percepts(Ps) :-
      log_percepts([w('>>'),w(Ps),nl]),
      process_percepts(Ps).

process_percepts([]).
process_percepts([Msg|MsgList]) :-
    process_percept(Msg), !,
    process_percepts(MsgList).
process_percepts([Msg|MsgList]) :-
    log_error([w('Unknown Percept: '), w(Msg), nl]),
    process_percepts(MsgList).
*/

%%start_agent(MyName, Handle) :-
%%    thread_is_thread(percepts), !,
%%    writeL(stderr, ['Error: For ' , start_agent(Handle),
%%                    ' the agent is already started', nl]).

'$assert_mod_controls_interface' :-
    current_predicate(mod_controls_/2), !,
    assert((mod_controls_interface(A, B) :- mod_controls_(A, B))).
'$assert_mod_controls_interface' :-
    assert(mod_controls_interface(A, A)).

'$assert_send_robotic_message_interface' :-
    current_predicate(send_robotic_message_/1), !,
    assert((send_robotic_message_interface(A) :- send_robotic_message_(A))).
'$assert_send_robotic_message_interface' :-
    assert((send_robotic_message_interface(A) :- send_robotic_message_default(A))).



'$assert_get_percepts_interface' :-
    current_predicate(get_percepts_/1), !,
    assert((get_percepts_interface(P) :- get_percepts_(P))).
'$assert_get_percepts_interface' :-
    assert((get_percepts_interface(P) :- get_or_wait_for_percepts(P))).
    
controls(C) :-
    mod_controls_interface(C, ModC),
    (
      ModC = []
    ->
      true
    ;
      send_robotic_message_interface(controls(ModC))
    ).

send_robotic_message_default(Controls) :-
    '$agent_info'(_, _, Handle),
    Controls ->> Handle.

/*
    controls_aux(ModC, Handle).
controls_aux([], _).
controls_aux([C1|Rest], Handle) :-
    C1 ->> Handle,
    controls_aux(Rest, Handle).
*/

actions(Actions) :-
    findall(A, retract('$current_actions'(A)), OldActions),
    forall((member(SA, Actions), \+'$discrete_action'(SA)),
           assert('$current_actions'(SA))),
    do_action_changes(OldActions, Actions, Controls),
    log_action(actions, [], Actions,  OldActions, Controls),
    controls(Controls).


'$update_interfaces'(Interface) :-
    '$running_teleo', !,
    '$retract_interface'(handle_percepts_interface(_)),
    '$retract_interface'(mod_controls_interface(_, _)),
    '$retract_interface'(handle_message_interface(_, _)),
    '$retract_interface'(checkBS_interface),
    '$retract_interface'(resources_callback),
    '$retract_interface'(get_percepts_interface(_)),
    '$retract_interface'(send_robotic_message_interface(_)),
    '$assert_percepts_interface'(Interface),
    '$assert_mod_controls_interface',
    '$assert_message_interface',
    '$assert_resources_hook_interface',
    '$assert_checkBS_interface',
    '$assert_get_percepts_interface',
    '$assert_send_robotic_message_interface'.
'$update_interfaces'(_).

'$retract_interface'(Call) :-
    retractall(Call).

start_agent(_MyName, _Handle, _Interface) :-
    '$agent_info'(agent, Name, _), !,
    writeL(["Error: Agent with name ", Name, " already running", nl_]),
    fail.

start_agent(MyName, Handle, Interface) :-
    '$running_teleo', Interface = user, !,
    set_default_message_thread(messages),
    retractall('$current_actions'(_)),
    (
     '$user_type_info'(percept_message_type, _, _, _, _)
    ->
     true
    ;
     writeL(["Error: percept_message_type must be declared in user mode",
             nl_]), fail 
    ),
    ensure_pedro_connected(MyName),
    assert('$agent_info'(agent, MyName, Handle)),
    '$update_interfaces'(Interface),
    thread_fork(messages, handle_messages),
    writeL(["Starting messages thread", nl_]),
    (current_predicate(handle_requests/0) ->
        thread_fork(requests, handle_requests) ; true ),
    thread_fork(percepts,start_percept_handler),
     writeL(["Starting percepts thread", nl_]),
     thread_wait_on_goal('__time__'(_, _)).
start_agent(MyName, Handle, Interface) :-
    '$running_teleo', !, 
    set_default_message_thread(messages),
    (
     '$user_type_info'(percept_message_type, _, _, _, _)
    ->
     writeL(["Error: percept_message_type has been declared but should not be when in ", Interface, " mode", nl_]), fail 
    ;
      Interface = updates
    ->
     assert('$user_type_info'(percept_message_type, defined_type,
                              percept_message_type,
                              '$constr_enum_type'([r_(percept), f_(percept)]),
                              []))
    ;
      true
    ),
    ensure_pedro_connected(MyName),
    assert('$agent_info'(agent, MyName, Handle)),
    '$update_interfaces'(Interface),
    thread_fork(messages, handle_messages),
    writeL(['Starting messages thread', nl_]),
    (current_predicate(handle_requests/0) ->
        thread_fork(requests, handle_requests) ; true ),
    thread_fork(percepts,start_percept_handler),
     writeL(['Starting percepts thread', nl_]),
     thread_wait_on_goal('__time__'(_, _)).
start_agent(_MyName, _Handle, _Interface) :-
    writeL(["Error: Attempting to start agent outside of the teleor interface",
	    nl_]),
    fail.

ensure_pedro_connected(MyName) :-
    \+ pedro_is_connected, 
    !,
    (
      pedro_connect, pedro_register(MyName)
    ->
      true
    ;
      writeL(["Error: Cannot connect to Pedro", nl_]), fail
    ).
ensure_pedro_connected(MyName) :-
    thread_handle(_:MyName@_),
    !.
ensure_pedro_connected(MyName) :-
    pedro_deregister,
     (
      pedro_register(MyName)
    ->
      writeL(["Re-registering this process with Pedro using name: ", MyName,nl_])
    ;
      writeL(["Error: Cannot register Pedro using the name ", MyName, nl_]), fail
    ).   

kill_agent :-
    retract('$agent_info'(agent, _, Handle)), !,
    finish ->> Handle,
    retractall('$current_actions'(_)),
    forall(task_(Name, _), kill_task(Name)),
    ( thread_is_thread(messages) -> thread_exit(messages) ; true ),
    thread_exit(percepts),
    pedro_deregister,
    pedro_disconnect.
kill_agent :-
    writeL(["Error: No agent to kill", nl_]),
    fail.



'$assert_percepts_interface'(all) :-
    !,
    assert((handle_percepts_interface(A) :- handle_percepts_all_(A))).
'$assert_percepts_interface'(updates) :-
    !,
    assert((handle_percepts_interface(A) :- handle_percepts_updates_(A))).
'$assert_percepts_interface'(none) :-   
    !,
    assert(handle_percepts_interface(_)).        
'$assert_percepts_interface'(_) :-
    current_predicate(handle_percepts_/1), !,
    assert((handle_percepts_interface(A) :-
            handle_percepts_(A))).
'$assert_percepts_interface'(_) :-   
    !,
    assert(handle_percepts_interface(_)).        


handle_percepts_all_(Ps) :-
    \+ground(Ps), !.
handle_percepts_all_(Ps) :-
    a_percept(P),
    \+member_eq(P, Ps),
    forget(P),
    fail.
handle_percepts_all_(Ps) :-
    handle_percepts_all_aux_(Ps).

handle_percepts_all_aux_([]).
handle_percepts_all_aux_([P|Ps]) :-
    type(P, percept), !,
    remember(P),
    handle_percepts_all_aux_(Ps).
handle_percepts_all_aux_([_P|Ps]) :-
    handle_percepts_all_aux_(Ps).

handle_percepts_updates_(Ps) :-
     \+ground(Ps), !.
handle_percepts_updates_(Ps) :-
    handle_percepts_updates_aux_(Ps).

handle_percepts_updates_aux_([]).
handle_percepts_updates_aux_([r_(P)|Ps]) :-
    type(P, percept), !,
    remember(P),
    handle_percepts_updates_aux_(Ps).
handle_percepts_updates_aux_([f_(P)|Ps]) :-
    type(P, percept), !,
    forget(P),
    handle_percepts_updates_aux_(Ps).
handle_percepts_updates_aux_([_|Ps]) :-
    handle_percepts_updates_aux_(Ps).
    

a_percept(P) :-
    '$percept_type_info'(_, _, P),
    call(P).

start_task(Name, TR) :-
    thread_is_thread(Name), !,
    writeL(stderr, ["Error: For " , start_task(Name, TR),
                    " the name is already in use", nl]).
start_task(Name, TR) :-
    '$task_start'(_),
    TR =.. [F|_],
    \+'$task_start'(F), !, 
    writeL(stderr, ["Error: for " , start_task(Name, TR),
                    " ", TR,
                    " has not been declared as a task start procedure", nl_]).
    
start_task(Name, TR) :-
    thread_wait_on_goal('__time__'(_, _)),
    writeL(['Starting task ', Name, nl_]),
    thread_fork(Name,  start_task_aux(TR)),
    assert(task_(Name,TR)).

kill_task(Name) :-
    thread_is_thread(Name), !,
    thread_exit(Name),
    retract(task_(Name,_TR)).
kill_task(Name) :-
        writeL(stderr, ["Error: For " , kill_task(Name),
                        " the task is not running", nl]).
/*
start :-
    thread_set_symbol(messages),
    thread_fork(percepts,start_percept_handler),
    belief_store_updated <<= _,
    (current_predicate(evaluator) -> thread_fork(evaluator, evaluator) ; true),
    (current_predicate(handle_message/2) -> handle_messages ; true).

run(TR) :-
    %log(all),
    thread_set_symbol(messages),
    thread_fork(percepts,start_percept_handler),
    belief_store_updated <<= _,
    do_task(TR),
    (current_predicate(evaluator) -> thread_fork(evaluator, evaluator) ; true),
    (current_predicate(handle_message/2) -> handle_messages ; true).

start_eval(Task) :-
    do_task(Task).
    %log(all),
    %thread_set_symbol(messages),1
    %thread_fork(percepts,start_percept_handler),
    %belief_store_updated <<= _,
    %errornl(belief_store_updated),
    %(current_predicate(evaluator) -> thread_fork(evaluator, evaluator) ; true),
    %(current_predicate(handle_message/2) ->
    %    errornl('Starting message handler'),handle_messages ; true),
    %errornl(end_of_start).
*/

handle_messages :-
    M <<- A,
    (
      M = quiting
    ->
      thread_exit(thread0)
    ;
     (
      nonvar(M)
     ->
      same_handle(A, _:Ag@Host),
      thread_atomic_goal(handle_message_interface(M, (Ag@Host)))
     ;
      true
     )
    ),
    handle_messages.

'$assert_checkBS_interface' :-
    current_predicate(checkBS_/0), !,
    assert((checkBS_interface :- checkBS_)).
'$assert_checkBS_interface' :-
    assert(checkBS_interface).

'$assert_message_interface' :-
    current_predicate(handle_message_/2), !,
    assert((handle_message_interface(M, A) :- handle_message_(M, A))).
'$assert_message_interface' :-
    assert((handle_message_interface(M, A) :-
	   '$default_handle_message'(M,A))).

'$default_handle_message'(M,A) :-
    ground(M), !,
    '__time__'(dummy, Time),
    remember(message_(M, A, Time)).
'$default_handle_message'(_, _).
	    
%%do_action_changes(OldActions, NewActions, Controls)
%% in theory stops all actions in OldActions and starts all
%% actions in NewActions.
%% In practice a stop and a start can be replaced by a modify


%%do_action_changes(OldActions, NewActions, Controls)
%% in theory stops all actions in OldActions and starts all
%% actions in NewActions.
%% In practice a stop and a start can be replaced by a modify
do_action_changes(Old, rpa(Actions), Controls) :-
    !,
    do_action_changes(Old, [], Controls1),
    process_actions(Actions, Controls2),
    append(Controls1, Controls2, Controls).
do_action_changes(Old, no_action, Controls) :-
    !,
    do_action_changes(Old, [], Controls).
do_action_changes(no_action, Actions, Controls) :-
    do_action_changes([], Actions, Controls).
do_action_changes([], Actions, Controls) :-
    process_actions(Actions, Controls).

%% An old action is discrete - so ignore
do_action_changes([OA|OldActions], Actions, Controls) :-
    '$discrete_action'(OA),
    !,
    do_action_changes(OldActions, Actions, Controls).
%% An old action is to be stopped and the same action started - so ignore
do_action_changes([OA|OldActions], Actions, Controls) :-
    delete(OA, Actions, Actions1),
    !,
    do_action_changes(OldActions, Actions1, Controls).
%% An old action is modified by a new action so send the modify message
do_action_changes([OA|OldActions], Actions, [mod_(MA)|Controls]) :-
    '$modify_match'(OA, MA),
    delete(MA, Actions, DActions),
    !,
    do_action_changes(OldActions, DActions, Controls).
%% The old action is not nil and not updated so stop
do_action_changes([OA|OldActions], Actions, [stop_(OA)|Controls]) :-
    do_action_changes(OldActions, Actions, Controls).

modification_delete([A|Actions], OAF, OAN, A, Actions) :-
    functor(A, OAF, OAN), !.
modification_delete([A|Actions], OAF, OAN, MA, [A|DActions]) :-
    modification_delete(Actions, OAF, OAN, MA, DActions).

process_actions([], []).
process_actions([], []) :- !.
process_actions([no_match|A], B) :-
    !,
    log_error(no_match),
    process_actions(A, B).
process_actions([A|As], [exec_(A)|Controls]) :-
    '$discrete_action'(A),
    process_actions(As, Controls).
process_actions([A|As], [start_(A)|Controls]) :-
    process_actions(As, Controls).

'$discrete_action'(A) :-
    '$user_type_info'(_, discrete, _, A, _).

'$modify_match'(OA, MA) :-
    '$user_type_info'(F, durative, _, OA, _),
    '$user_type_info'(F, durative, _, MA, _).

