%% Copyright 2014 Keith Clark, Peter Robinson
%%
%% Licensed under the Apache License, Version 2.0 (the "License");
%% you may not use this file except in compliance with the License.
%% You may obtain a copy of the License at
%%
%%     http://www.apache.org/licenses/LICENSE-2.0
%%
%% Unless required by applicable law or agreed to in writing, software
%% distributed under the License is distributed on an "AS IS" BASIS,
%% WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
%% See the License for the specific language governing permissions and
%% limitations under the License.


/*
  compiler for TR code

*/
?- op(1023, xfx,  '~~' ).
?- op(600, fx,  'display' ).
?- op(500, fy, '++').
?- op(500, fy, '--').
?-op(500,xfy,'++').
?- op(600,xfx,'..').
?-op(920,xfx,'~>').
?-op(920,xfx,'~~>').
?- op(1028, xfx,  '^^' ). 
?- op(1028, xfx,  'until' ). 

?- op(1025, xfx, '+>').
?- op(1025, xfx, 'or_while'). 

?- op(1025, xfx, '~^'). 
?- op(1025, xfx, 'or_while_until'). 
?- op(1023, fx, 'time'). 
?- op(1023, xfy, 'time').


/*
?- op(600,xfx,'..').

?-op(1200,xfx,'if').
?-op(1100,xfx,'||').
?-op(1050,fx,'||').
?-op(1020,fx,'forall').
?-op(1027,xfx,'then').
?-op(1024,xfx,'where').
?- op(1020, xfx,  '~~' ). 
?- op(1027, xfx,  '--' ). 
?- op(1020, xfx, '=>'). 
?- op(1028, xfx,  '^^' ). 
?- op(1028, xfx,  'until' ). 
?- op(1025, xfx, '+>'). 
?- op(1025, xfx, '++'). 
?- op(1025, xfx, 'or_while'). 
?- op(1025, xfx, '+^'). 
?- op(1025, xfx, 'or_while_until'). 
?- op(1023, fx, 'time'). 
?- op(1023, xfy, 'time').
?- op(1000, xfy,  '&' ).
?- op(900, fy, 'not').
?-op(600,xfx,'in').
?-op(500,xfy,'<>').
?- op(200, fy, 'type').
?- op(200, xfx, ':=').


%% temporary storage while processing program
?- dynamic('$tr_prog'/2).       %% remember TR programs
?- dynamic('$tr_pred'/2).	%% remember dynamicB info
?- dynamic('$tr_clause'/2).     %% for storing clauses
?- dynamic('$runtime_goal'/1).  %% remember goals to exec at runtime
?- dynamic('$task_atomic'/2).
?- dynamic('$task_start'/1).
?-dynamic('$resource_info'/3).
?-dynamic('$resources'/1).
?-dynamic('$durative'/1).
?-dynamic('$discrete'/1).
?-dynamic('$resource_type'/1).
?-dynamic('$defined_type'/2).

?-dynamic('$tmp_resources'/1).
?-dynamic('$tmp_durative'/1).
?-dynamic('$tmp_discrete'/1).
?- dynamic('$tmp_task_atomic'/1).
?- dynamic('$tmp_defined_type'/1).
?- dynamic('$tmp_type'/1).
?- dynamic('$expand_map'/3).
?- dynamic('$resource_type'/1).
?- dynamic('$type_decl'/3).
*/

?- dynamic('$seen_clause'/3). %% To avoid looks of recursive defns
?-dynamic('$uses_named'/3). %% for checking uses of names resources

?- dynamic('$tr_pred'/2).
?- dynamic('$tr_clause'/2).

'$tr_clause_head'(Head) :- '$tr_clause'(Head, _).

'$generate_durative_clauses'([], []).
'$generate_durative_clauses'([D|DurativeDecls],
                             [Clause|DuativeClauses]) :-
    '$resource_info'(Term1, RTypes, D),
    '$resource_info'(Term2, RTypes, D),
    Clause = '$modify_match'(Term1, Term2),
   '$generate_durative_clauses'(DurativeDecls, DuativeClauses).

'$generate_discrete_clauses'([], []).
'$generate_discrete_clauses'([D|DiscreteDecls],
                             [Clause|DiscreteClauses]) :-
    '$resource_info'(Dis, _RTypes, D),
   Clause = '$discrete_action'(Dis),
    '$generate_discrete_clauses'(DiscreteDecls, DiscreteClauses).

%% Check that no times or actions are used between the task_start program
%% and the first task_atomic found in the call chain

%% resources are used so check is done
'$check_task_start' :-
    ( '$resource'(_) ; '$task_atomic'(_,_) ),
    '$task_start'(TR),
    '$resource_info'(TRHead, _, TR),
    freeze_vars(TRHead),
    '$check_call_trees'(TRHead, [], no_times, []),
    fail.

%% resources are not used so only need to check there is at least one
%% task_start
'$check_task_start' :-
    (
      \+'$task_start'(_)
    ->
      write_term_list(stderr,
                      ['Error: there are no task_start declarations',nl]),
      thread_exit
    ;
      true
    ).

%% Checking the resource uses

%% If there are no resource or task_atomic
%% declarations then there is nothing to do
'$check_resource_uses' :-
    (\+'$resource'(_) ; \+'$task_atomic'(_,_) ),
    !.

%% Check if there is a declaration for all the resources
'$check_resource_uses' :-
    (
      '$task_atomic'(_,_)
    ->
      true
    ;
      write_term_list(stderr, ['Error: No task_atomic declaration', nl]),
      thread_exit
    ),
    fail.

%% Check that, if resources are defined, all TR programs have a
%% resource declaration of some kind
'$check_resource_uses' :-
    '$resource'(_),
    '$tr_prog'(TR, _),
    (
      '$resource_info'(TR, _, _)
    ->
      true
    ;
      write_term_list(stderr, ['Error: ', TR,
                               ' has no resources declaration', nl]),
      thread_exit
    ),
    fail.
%% Check that the actions (recursively) of each task_atomic TR program
%% only uses resources that are either named resources or are listed in the call
'$check_resource_uses' :-
    '$task_atomic'(TRF, Named),
    '$type_info'(TRF, _, _, TR, _),
    functor(TR, P1, N),
    '$add_resource_types'(Named, TNamed),
    %% assert that this atomic TR has already been checked and its extra
    %% resources are its named resources
    assert('$uses_named'(P1, N, TNamed)),
    '$resource_info'(TR, Resources,  _),
    '$tr_prog'(TR, TRBody),
    '$check_resource_uses_aux'(TRBody, TR, Resources, TNamed, 1, [], _Out, _),
    fail.
'$check_resource_uses'.

   
'$add_tr_resource_info' :-
    '$tr_prog'(Term, _),
    functor(Term, P, _),
    '$assert_resource_info'(P),
    fail.
'$add_tr_resource_info'.

'$assert_resource_info'(P) :-
    !,
    '$type_info'(P, _, Types, Term, _),
    Term =..[_|Args],
    '$zip_type'(Args, Types, A3),
    '$add_file_info'('$resource_info'(_, _, P)),
    assert('$resource_info'(Term, A3, P)).

'$zip_type'([], [], []).
'$zip_type'([V|Vs], [T|Ts], [(V:T1)|VTs]) :-
    ( T = _(T1) ; T = T1), 
    '$resource'(T1), !,
    '$zip_type'(Vs, Ts, VTs).
'$zip_type'([_V|Vs], [_T|Ts],VTs) :-
    '$zip_type'(Vs, Ts, VTs).


'$add_resource_types'([], []).
'$add_resource_types'([N|Named], [(N:T)|TNamed]) :-
    '$resource'(RL),
    member((N:T), RL),
    '$add_resource_types'(Named, TNamed).
    
   
%% Check each rule individually
%% '$check_resource_uses_aux'(RuleList, TR, Resources, Named, Index,
%%                            In, Out, NRes)
%% RuleList is the list of rules ot TR being checked.
%% Resources is the list of resource info of form [Var1:type1, ...]
%% Named is the list of (typed) named resources as above
%% Index is the index of the rule to be examined
%% In is the input P/N of TRs already being checked
%% Out is the output version of In
%% NRes is output the set of resources found other than those
%%      listed in Resources
%%
%% The dynamic predicate '$uses_named'(P, N, NRes)
%% stores the 'extra' resources found in P/N that are not args to the call

'$check_resource_uses_aux'([], _TR, _Resources, _Named, _I, Out, Out, []).
'$check_resource_uses_aux'([(_LHS ~> RHS)|TRBList], TR, Resources, Named,
                           Index, In, Out, NRes) :-
    '$check_resource_uses_aux2'(RHS, TR, Resources, Named, Index, In, Mid,
                                Seen2),
    Index1 is Index+1,
    '$check_resource_uses_aux'(TRBList, TR, Resources, Named,
                               Index1, Mid, Out, Seen3),
    %% Seen2 are the 'extra' resources of this rule and
    %% Seen3 are the extra resources for the rest of the rules
    union_list(Seen2, Seen3, NRes).


%% '$check_resource_uses_aux2'(Actions, TR, Resources, Named, Index, In, Out, NRes) :- errornl('$check_resource_uses_aux2'(Actions, TR, Resources, Named, Index, In, Out, NRes)),fail.

%% This is like the above predicate but for the action of the given rule
'$check_resource_uses_aux2'((Actions ++ _), TR, Resources, Named, Index,
                            In, Out, NRes) :-
    !,
   '$check_resource_uses_aux2'(Actions, TR, Resources, Named, Index,
                               In, Out, NRes). 
'$check_resource_uses_aux2'('$nil', _TR, _Resources, _Named, _Index,
                            In, Out, []) :-
    !, In = Out.
'$check_resource_uses_aux2'(timed_seq(TS), TR, Resources, Named, Index,
			    In, Out, NRes) :-
    !,
    %% a timed sequence
    %% for a timed sequence we collect all the extra resources from each
    %% entry in the sequence
    '$extract_ts_actions'(TS, As),
    '$check_resource_uses_aux3'(As, TR, Resources, Named, Index, In, Out, NRes),
    diff_list(NRes, Named, Invalid),
    (
      Invalid = []
    ->
      true
    ;
      write_term_list(stderr, [nl, 'Error: Rule number ', Index, ' of ',
			       TR, nl, 
			       'contains invalid resources ', Invalid, nl]),
      thread_exit
    ) .
'$check_resource_uses_aux2'(WR, TR, Resources, Named, Index, In, Out, NRes) :-
    %% a wait-repeat
    %% for a wait repeat check the action
    WR = '$wait_repeat'(Act, _), !,
    '$check_resource_uses_aux2'(Act, TR, Resources, Named, Index,
                                In, Out, NRes).
'$check_resource_uses_aux2'([TR1], _TR, _Resources, _Named,
                            _Index, In, Out, NRes) :-
    '$type_info'(_, tr_prog, _, TR1, _),
    %% a TR program
    %% this TR has already been checked
    functor(TR1, P, N),
    '$uses_named'(P, N, NRes),
    !,
    %% processed in different task atomic with same resources
    In = Out.
'$check_resource_uses_aux2'([RHS], _TR, Resources, _Named,
			    _Index, In, Out, NRes) :-
    %% a TR program
    %% this TR is currently being checked and so can be ignored
    functor(RHS, P, N),
    member(P/N, In), !,
    %% already process in this task_atomic
    '$resource_info'(RHS, Res,  _),
    diff_list(Res, Resources, NRes),
    In = Out.
'$check_resource_uses_aux2'([TR1], TR, _Resources, Named,
                            Index, In, Out, NRes) :-
    '$type_info'(_, tr_prog, _, TR1, _),
    %% a TR program that need checking
    '$resource_info'(TR1, Res,  _), !,
    '$tr_prog'(TR1, BList),
    functor(TR1, P, N),
    %% We check inside for resources not part of its resource args
    '$check_resource_uses_aux'(BList, TR1, Res, Named, 1, [P/N|In], Out, NRes),
    %% at this point NRes is the extra resources used - so save that away
    assert('$uses_named'(P, N, NRes)),
    %% Any resource that is in NRes but not in Named are Invalid
    diff_list(NRes, Named, Invalid),
    (
      Invalid = []
    ->
      true
    ;
      write_term_list(stderr, [nl, 'Error: Rule number ', Index, ' of ',
			       TR, nl, 
			       'contains invalid resources ', Invalid, nl]),
      thread_exit
    ).
'$check_resource_uses_aux2'(ActionList, TR, Resources, Named,
                            Index, In, Out, NRes) :-
    !,
    %append(Resources, Named, AllR),
    '$collect_resources'(ActionList, Seen1),
    %% Seen1 is all the resources mentioned in the primitive actions
    %% and so we subtract out Resources to get the extras in NRes
    diff_list(Seen1, Resources, NRes),
    %% Any resource that is in NRes but not in Named are Invalid
    diff_list(NRes, Named, Invalid),
    (
      Invalid = []
    ->
      In = Out
    ;
      write_term_list(stderr, [nl, 'Error: Rule number ', Index, ' of ',
			       TR, nl, 
			       'contains invalid resources ', Invalid, nl]),
      thread_exit
    ).
    

%% For checking timed sequences - check each element in the list
%% and return the overall collection of extra resources
'$check_resource_uses_aux3'([], _TR, _Resources, _Named, _Index, In, Out, []) :-
    In = Out.
'$check_resource_uses_aux3'([A|As], TR, Resources, Named, Index,
                            In, Out, NRes) :-
    '$check_resource_uses_aux2'(A, TR, Resources, Named, Index, In, Mid, NRes1),
    '$check_resource_uses_aux3'(As, TR, Resources, Named,
                               Index, Mid, Out, NRes2),
    union_list(NRes1, NRes2, NRes).


'$extract_ts_actions'([], []).
'$extract_ts_actions'([TSA|Rest], Actions) :-
    '$extract_ts_actions'(Rest, RestActions),
    (
     TSA = '$time'(_, A, _)
    ->
     Actions = [A|RestActions]
    ;
     Actions = [TSA|RestActions]
    ).
    

/*
'$resources_of_guards'((LHS +^ _), LHSRes) :-
    !,
    goals2list(LHS, Goals),
    '$collect_resources'(Goals, LHSRes).
'$resources_of_guards'((LHS ^^ _), LHSRes) :-
    !,
    goals2list(LHS, Goals),
    '$collect_resources'(Goals, LHSRes).
'$resources_of_guards'((LHS ++ _), LHSRes) :-
    !,
    goals2list(LHS, Goals),
    '$collect_resources'(Goals, LHSRes).
'$resources_of_guards'((LHS ++ _ ^^  _), LHSRes) :-
    !,
    goals2list(LHS, Goals),
    '$collect_resources'(Goals, LHSRes).
'$resources_of_guards'(LHS, LHSRes) :-
    !,
    goals2list(LHS, Goals),
    '$collect_resources'(Goals, LHSRes).
*/

%% Collect all the resources of a list of primitive actions
'$collect_resources'([], []).
'$collect_resources'([G|Goals], LHSRes) :-
    '$collect_resources'(Goals, LHSres1),
    (
      '$resource_info'(G, Res,  _)
    ->
      union_list(Res, LHSres1, LHSRes)
    ;
      LHSRes = LHSres1
    ).

    
%% check the possible call stack for time/primitive action errors

%% Already checked
'$check_call_trees'(TR, _Resources, _Times, Done) :-
    functor(TR, P, N),
    \+ \+ member(P/N, Done), !.

%% At a task_atomic call
'$check_call_trees'(TR, Resources, Times, Done) :-
    '$task_atomic'(TR,_),
    !,
    '$resource_info'(TR, Res,  _),
    (
      Resources = [], Times = has_times
    ->
      %% resources are used but a time has been used earlier
      write_term_list(stderr, ['Error: ' , TR,
                               ' has a first use of resources with non-empty timeouts', nl]),
      thread_exit
    ;
      '$tr_prog'(TR, Body),
      '$check_first_nil'(TR, Body),
      functor(TR, P, N),
      %% If no resources has been used so far then use the task atomic resources
      (
        Resources = []
      ->
        '$check_call_trees_body'(Body, TR, Res, Times, [(P/N)|Done])
      ;
        '$check_call_trees_body'(Body, TR, Resources, Times, [(P/N)|Done])
      )
    ).
%% Action is not task atomic and so is a set of primitive actions
'$check_call_trees'(TR, Resources, Times, Done) :-
    '$tr_prog'(TR, Body),
    functor(TR, P, N),
    '$check_call_trees_body'(Body, TR, Resources, Times, [(P/N)|Done]).

'$check_call_trees_body'(Body, TR, Resources, Times, Done) :-
    '$index_member'((Head -> RHS), Index, Body),
    '$times_of'(Head, RHS, Times, OTimes),
    '$check_call_trees_RHS'(RHS, Index, TR, Resources, OTimes, Done),
    fail.

'$check_call_trees_RHS'((RHS ~~ _), Index, TR, Resources, Times, Done) :-
    !,
    '$check_call_trees_RHS'(RHS, Index, TR, Resources, Times, Done).
'$check_call_trees_RHS'({}, _TR, _Index, _Resources, _Times, _Done) :-
    !.
'$check_call_trees_RHS'({}(_A), Index, TR, Resources, _Times, _) :-
    !,
    (
      Resources = []
    ->
      '$get_map'((TR .. _), (TR .. {}(Body))),
      '$index_member'(Rule, Index, Body),
      write_term_list(stderr, ['Error: The action of the rule rule ', nl,
			       Rule, nl,
			       'of ' , TR,
                               ' is primitive and is called outside a task_atomic', nl]),
      thread_exit
    ;
      true
    ).
'$check_call_trees_RHS'(RHS, Index, TR, Resources, _Times, Done) :-
    RHS = [_|_],
    !,
    member((X:_), RHS),
    '$check_call_trees_RHS'(X, Index, TR, Resources, has_times, Done),
    fail.
'$check_call_trees_RHS'(RHS, _Index, _TR, Resources, Times, Done) :-
    '$check_call_trees'(RHS, Resources, Times, Done).

/*
%% determine if a given rule mentions times
'$times_of'(_, _, has_times, has_times) :- !.
'$times_of'((_ ++ time(_)), _RHS, _Times, has_times) :- !.
'$times_of'((_ ++ time(_,_)), _RHS, _Times, has_times) :- !.
'$times_of'((_ ++ time(_) ^^ _), _RHS, _Times, has_times) :- !.
'$times_of'((_ ++ time(_,_) ^^ _), _RHS, _Times, has_times) :- !.
'$times_of'((_ ^^ time(_)), _RHS, _Times, has_times) :- !.
'$times_of'((_ ^^ time(_,_)), _RHS, _Times, has_times) :- !.
'$times_of'(_, _, T, T).
*/

'$check_first_nil'(_, ((_ -> {}), _)) :-
    !.
'$check_first_nil'(TR, _Rules) :-
    '$get_map'((TR.._), (TR.._)),
    write_term_list(stderr, ['Error: The first rule for ' , TR,
                             ' is not the nil action', nl]),
    thread_exit.
    
'$resource_subset'(_Res, all) :- !.
'$resource_subset'(all, _Res2) :-
    !,
    fail.
'$resource_subset'(Res1, Res2) :-
    forall(member(X, Res1), member(X, Res2)).

cons_member(A, (A,_)).
cons_member(A, (_, B)) :-
    !,
    cons_member(A, B).
cons_member(A, A).




/*
  Collect all the goals in the body of all clauses in a predicate
  into a single 'clause'
  E.G.
  p :- a,b
  p :- c
  p :- d -> e;f
  is first stored as '$tr_clause'(p, (a,b)), ...
  and then replaced by a single fact
  '$tr_clause'(p, ((a,b),c, (d->e;f)))
  This is used to see what dynamicB facts depends on
*/
'$tr_build_clause_info' :-
    '$tr_prog'(_, _), !,
    findall(Head,
            ('$user_type_info'(_, relation, _, Head, _),
                \+'$percept'(Head), \+'$belief'(Head)),
           Heads),
    forall(member(Head, Heads), '$tr_build_clause_info_aux'(Head)).
'$tr_build_clause_info'.

'$tr_build_clause_info_aux'(Head) :-
    findall(Body, clause(Head, Body), Clauses),
    functor(Head, P, N),
    (
      Clauses = []
    ->
      assert('$tr_clause'(P/N, true))
    ;
      list2goals(Clauses, Goals),
      assert('$tr_clause'(P/N, Goals))
    ).

list2goals([X], X) :- !.
list2goals([X|Y], (X, Ys)) :-
    list2goals(Y, Ys).


flip_sign(-, +).
flip_sign(+, -).

/*
  construct the list of all dynamicB preds the goal depends on
  '$tr_find_clause_depend'(Goal, InPreds, OutPreds)
*/

'$tr_find_clause_depend'(G, P, Preds) :-
    '$tr_find_clause_depend_aux'(G, [], P1, +),
    append(P, P1, Preds).

%% recursive go through goals
'$tr_find_clause_depend_aux'(V, _P1, _Preds, _Sign) :-
    var(V), !,
    write_term_list(stderr, [nl, 'Error: variable in TR guard.', nl]).
'$tr_find_clause_depend_aux'((B1, B2), P1, Preds, Sign) :-
    !,
    '$tr_find_clause_depend_aux'(B1, P1, P2, Sign),
    '$tr_find_clause_depend_aux'(B2, P2, Preds, Sign).
/*
  '$tr_find_clause_depend_aux'((B1 -> B2;B3), P1, Preds, Sign) :-
    !,
    '$tr_find_clause_depend_aux'(B1, P1, P2, Sign),
    '$tr_find_clause_depend_aux'(B2, P2, P3, Sign), 
    '$tr_find_clause_depend_aux'(B3, P3, Preds, Sign).
'$tr_find_clause_depend_aux'((B1; B2), P1, Preds, Sign) :-
    !,
    '$tr_find_clause_depend_aux'(B1, P1, P2, Sign),
    '$tr_find_clause_depend_aux'(B2, P2, Preds, Sign).
*/
'$tr_find_clause_depend_aux'(forall(B1, B2), P1, Preds, Sign) :-
    !,
    flip_sign(Sign, OSign),
    '$tr_find_clause_depend_aux'(B1, P1, P2, OSign),
    '$tr_find_clause_depend_aux'(B2, P2, Preds, Sign).
'$tr_find_clause_depend_aux'(\+(B1), P1, Preds, Sign) :-
    !,
    flip_sign(Sign, OSign),
    '$tr_find_clause_depend_aux'(B1, P1, Preds, OSign).
/*
  '$tr_find_clause_depend_aux'(try(B1), P1, Preds, Sign) :-
    !,
    '$tr_find_clause_depend_aux'(B1, P1, Preds, Sign).
'$tr_find_clause_depend_aux'(try_not(B1), P1, Preds, Sign) :-
    !,
    flip_sign(Sign, OSign),
    '$tr_find_clause_depend_aux'(B1, P1, Preds, OSign).
*/
'$tr_find_clause_depend_aux'(not(B1), P1, Preds, Sign) :-
    !,
    flip_sign(Sign, OSign),
    '$tr_find_clause_depend_aux'(B1, P1, Preds, OSign).
'$tr_find_clause_depend_aux'('$time'(_), P1, Preds, _Sign) :-
    !,
    Preds = P1.
'$tr_find_clause_depend_aux'(true, P1, Preds, _Sign) :-
    !,
    Preds = P1.
'$tr_find_clause_depend_aux'('$time'(P, _), P1, Preds, Sign) :-
    !,
    '$tr_find_clause_depend_aux'(P, P1, Preds, Sign).
%% if B is a dynamicB then add to preds
'$tr_find_clause_depend_aux'(B, P1, Preds, Sign) :-
    ( '$percept'(B) ; '$belief'(B) ), !,
    functor(B,P,N),
    append([Sign(P)/N], P1, Preds).
%% if B is a normal goal that is defined as a pred
%% then recursively find dependencies
'$tr_find_clause_depend_aux'(B, P1, Preds, Sign) :-
    functor(B,P,N),nonvar(P),
    '$tr_clause'(P/N, Code),!,
    (
      '$seen_clause'(Sign(P), N, P2)
    ->
      append(P1, P2, Preds)
    ;
      flip_sign(Sign, OSign), '$seen_clause'(OSign(P), N, OPreds)
    ->
      flip_all_signs(OPreds, P2),
      assert('$seen_clause'(Sign(P), N, P2)),
      append(P1, P2, Preds)
    ;
      assert('$seen_clause'(Sign(P), N, [])),
      '$tr_find_clause_depend_aux'(Code, P1, Preds, Sign),
      retract('$seen_clause'(Sign(P), N, [])),
      assert('$seen_clause'(Sign(P), N, Preds))
    ).
'$tr_find_clause_depend_aux'(fail, _, _, _) :-
    write_term_list(stderr, ['Error: fail appears in a predicate used in a TR guard', nl]),
    thread_exit.

%% otherwise ignore 
'$tr_find_clause_depend_aux'(_B, P1, P1, _Sign).

flip_all_signs([], []).
flip_all_signs([Sign(P)/N|Preds], [OSign(P)/N|OPreds]) :-
    flip_sign(Sign, OSign),
    flip_all_signs(Preds, OPreds).

'$strip_types'([], []).
'$strip_types'([(X:_)|Xs], [X|Ys]) :-
    '$strip_types'(Xs, Ys).

'$set_tr_prog_vars'(Head) :-
    '$type_info'(_, tr_prog, Types, Head, _),
    Head =.. [_|Args],
    '$set_tr_prog_vars_aux'(Args, Types, Vars),
    ip_set('$tr_prog_vars', Vars).

'$set_tr_prog_vars_aux'([], [], []).
'$set_tr_prog_vars_aux'([A|Args], [T|Types], [A|Vars]) :-
    var(A), T = '~>'(_), !,
    '$set_tr_prog_vars_aux'(Args, Types, Vars).
'$set_tr_prog_vars_aux'([_A|Args], [_T|Types], Vars) :-
    '$set_tr_prog_vars_aux'(Args, Types, Vars).

%% expand a TR program

'$expand_a_TR'(Head, Body) :-
    TR_prog = '$tr_prog'(Head, Body),
    '$saved_input_with_vars'(TR_prog, Vars, _),
    '$get_underscores'(TR_prog, Vars, Unders),
    ip_set('$underscore_vars', Unders),
    freeze_term(Body),
    '$set_tr_prog_vars'(Head),
    '$translate_TR_body'(Body, TransBody),
    %thaw_term(TransBody),
    '$expand_one_TR'(Head, TransBody, 1, [], ClausesTR,
                     ClausesRetry, ClausesOverlap),
    %%XXXXXX
    (
      functor(Head, HeadF, _), '$task_atomic'(HeadF, _)
    ->
      '$resource_info'(Head, TypesResources, _),
      '$strip_types'(TypesResources, Resources),
      TR = Head,
      AtomicClause =
    (
      (
        tr_rule(TR, AT, _, _, InTerms, OutTerms, InDepend, OutDepends,
                _InTime, OutDelay, OldActions, Actions, 
                ResourcesUsed) :-
      ResourcesUsed = no,
        thread_symbol(Task),
        \+'__can_run__'(Task, Resources, AT), !,
        AT \= wait,
        Actions = [],
        OutTerms = [wait_choice(TR, InDepend)|InTerms],
        do_action(OldActions, Actions, OutTerms),
        merge([+'__new_percepts__'/0], InDepend, OutDepends), 
        OutTerms = [wait_choice(TR, InDepend)|InTerms],
        OutDelay = -1
      )
    ),
      ClausesTR = [NilClause|RestClausesTR],
      FullClausesTR = [NilClause, AtomicClause|RestClausesTR]
    ;
      FullClausesTR = ClausesTR
    ),
    functor(Head, HeadF, HeadN),
    functor(Template, HeadF, HeadN),
    '$add_file_info'(tr_rule(Template, _, _, _, _, _, _, _, _, _, _, _, _)),
    '$add_file_info'(retry_from(_, Template, _, _, _, _)),
    '$add_file_info'(overlap_at(Template, _, _, _)),
    forall(member(C, FullClausesTR), assert(C)),
    forall(member(C, ClausesRetry), assert(C)),
    forall(member(C, ClausesOverlap), assert(C)).

'$translate_TR_body'(Body, Trans) :-
    '$update_whilstuntils'(Body, TransBody),
    '$translate_TR_rules'(TransBody, Trans).

'$translate_TR_rules'([], []).
'$translate_TR_rules'([R|Rs], [RT|RsT]) :-
    '$translate_a_tr_rule'(R, RT),
    '$translate_TR_rules'(Rs, RsT).

'$translate_a_tr_rule'((LHS ~> RHS), (TLHS ~> TRHS)) :-
    LHS = whilstuntil(G, W, U),
    '$expand_funs'(f(W, U, RHS), f(TW,TU,TRHS), ExtraCalls),
    '$translate_goal'(G, TransG),
    '$translate_timed_goal'(TW, TransW),
    '$translate_timed_goal'(TU, TransU),
    (
      ExtraCalls = []
    ->
      TLHS = whilstuntil(TransG, TransW, TransU)
    ;
      '$list2tuple'(ExtraCalls, EC),
      TLHS = whilstuntil((TransG, EC), TransW, TransU)
    ).

'$translate_timed_goal'('$time'(_, G, T), '$time'(TransG, T)) :-
    !,
    '$translate_goal'(G, TransG).
'$translate_timed_goal'('$time'(_, T), '$time'(T)) :- !.
'$translate_timed_goal'(G, TransG) :-
    '$translate_goal'(G, TransG).

%% expand all tr programs into clauses
/*
'$expand_TR'([], [], [], []).
'$expand_TR'([tr_prog(Head, Body)|Rest], AllClauses,
             RetryClauses, OverlapClauses) :-
     write_term_list(['\nExpanding: ',tr_prog(Head,Body),nl]),
    '$expand_TR'(Rest, Clauses, Retry, Overlap), 
     write_term_list(['\nDoing while untils: ',Body,nl]),    
    '$update_whilstuntils'(Body, WBody),
   write_term_list(['\nDoing expand one: ',Head,nl,WBody,nl]),
     (
       '$expand_one_TR'(Head, WBody, 1, [], ClausesTR,
			ClausesRetry, ClausesOverlap)
     ->
       true
     ;
       writeL(stderr, ['Cannot compile ', nl_, '$tr_prog'(Head, Body)]),
       thread_exit
     ),
    append(ClausesRetry, Retry, RetryClauses),
    append(ClausesOverlap, Overlap, OverlapClauses),
    (
      '$task_atomic'(Head,_)
    ->
      '$resource_info'(Head, TypesResources, _),
      '$strip_types'(TypesResources, Resources),
      TR = Head,
      AtomicClause =
    (
      (
        tr_rule(TR, AT, _, _, InTerms, OutTerms, InDepend, OutDepends
                _InTime, OutDelay, OldActions, Actions, 
                ResourcesUsed) :-
      ResourcesUsed = no,
        thread_symbol(Task),
        \+'__can_run__'(Task, Resources, AT), !,
        AT \= wait,
        Actions = [],
        OutTerms = [wait_choice(TR, InDepend)|InTerms],
        do_action(OldActions, Actions, OutTerms),
        merge([+'__new_percepts__'/0], InDepend, OutDepends), 
        OutTerms = [wait_choice(TR, InDepend)|InTerms],
        OutDelay = -1
      )
    ),
      ClausesTR = [NilClause|RestClausesTR],
      append([NilClause, AtomicClause|RestClausesTR], Clauses, AllClauses)
    ;
      append(ClausesTR, Clauses, AllClauses)
    ).
'$expand_TR'([tr_prog(Head, Body)|_Rest], [], [], []) :-
    write_term_list(['Cannot compile ', tr_prog(Head, Body)]).
  */
%%%%%
'$update_whilstuntils'([(true ~> RBody)], WBody) :-
    !,
    '$update_whilstuntils_aux'(true, WHead),
    '$add_action_wrapper'(RBody, WRBody),
    WBody = [(WHead ~> WRBody)].
'$update_whilstuntils'([(RHead ~> RBody)], WBody) :-
    !,
    '$update_whilstuntils_aux'(RHead, WHead),
    '$update_whilstuntils_aux'(true, TrueHead),
    '$add_action_wrapper'(RBody, WRBody),
    WBody = [(WHead ~> WRBody), (TrueHead ~> rb(pa([error]), true))].
'$update_whilstuntils'([(RHead ~> RBody) | BRest], WBody) :-
    !,
    '$update_whilstuntils_aux'(RHead, WHead),
    '$add_action_wrapper'(RBody, WRBody),
    '$update_whilstuntils'(BRest, WBody1),
    WBody = [(WHead ~> WRBody)|WBody1].

'$update_whilstuntils_aux'('$until'(_, '$while'(_, G, W), U),
                           whilstuntil(G, W, U)) :- !.
'$update_whilstuntils_aux'('$until'(_, G, U),
                           whilstuntil(G, true, U)) :- !.
'$update_whilstuntils_aux'('$while'(_, G, W),
                           whilstuntil(G, W, true)) :- !.
'$update_whilstuntils_aux'('$while_until'(_, G, '$time'(_, WU, T)),
                           whilstuntil(G,'$time'(_, \+WU, T),
				       '$time'(_, WU, T))) :- !.
'$update_whilstuntils_aux'('$while_until'(_, G, WU),
                           whilstuntil(G, \+WU, WU)) :- !.
'$update_whilstuntils_aux'('$time'(_, G, T),
                           whilstuntil(G, '$time'(_, T), '$time'(_, T))) :- !.
'$update_whilstuntils_aux'(G, whilstuntil(G, true, true)) :- !.

/*'$update_whilstuntils_aux'((G ~^ H), whilstuntil(G, \+H, H)) :- !.
'$update_whilstuntils_aux'((G time T), whilstuntil(G, time(T), time(T))) :- !.
'$update_whilstuntils_aux'((G ^^ H), whilstuntil(G1, T, H)) :-
    (
     G=(G1 ~~ T)
    ->
      true
    ;
      T = true,
      G = G1
    ), !.
'$update_whilstuntils_aux'((G ~~ T), whilstuntil(G, T, true)) :-
    !.
'$update_whilstuntils_aux'(RHead, whilstuntil(RHead, true, true)).
*/

'$add_action_wrapper'(Body, WBody) :-
    Body = '++'(G,A),!,
    '$add_action_wrapper'(G, rb(WG, _)),
    WBody = rb(WG,A).

'$add_action_wrapper'(Body, WBody) :-
    Body = ['$nil'], !,
    WBody = rb(pa([]), true).
'$add_action_wrapper'(Body, WBody) :-
    Body = [TR], '$type_info'(_, tr_prog, _, TR, _), !,
    WBody = rb(TR, true).
'$add_action_wrapper'(Body, WBody) :-
    Body = [TR], functor(TR, F, _), var(F),  ip_lookup('$tr_prog_vars', PV),
    member_eq(F, PV), !,
    WBody = rb(TR, true).
'$add_action_wrapper'(Body, WBody) :-
    Body = [ActionGoal],
    '$type_info'(_, action, _, ActionGoal, _), !,
    WBody = rb(pa([]), ActionGoal).
'$add_action_wrapper'(Body, WBody) :-
    Body = timed_seq(Seq), !,
    '$add_action_wrapper_ts'(Seq, WSeq),
    WBody = rb(timed_seq(WSeq), true).
'$add_action_wrapper'(Body, WBody) :-
    Body = '$wait_repeat'(A, '$repeat'(_, T, N)), !,
    WBody = rb(wait_repeat(A, T, N), true).
'$add_action_wrapper'(Body, WBody) :-
    Body = '$wait_repeat'(A, T), !,
    WBody = rb(wait_repeat(A, T, 1), true).

'$add_action_wrapper'(Body, rb(pa(Body), true)).

'$add_action_wrapper_ts'([], []).
'$add_action_wrapper_ts'([S|Seq], [WS|WSeq]) :-
    (
      S = '$time'(_, B,T)
    ->
      '$add_action_wrapper'(B, rb(BW, _)),
      WS = (BW : T)
    ;
      '$add_action_wrapper'(S, rb(BW, _)),
      WS = (BW : -1)
    ),
    '$add_action_wrapper_ts'(Seq, WSeq).
 
'$process_vars_of_rule'(TR, Index, RHead, Whilst, Until, A, RIDVars) :-
    collect_vars(TR, TRVars),
    collect_vars(Whilst, WhilstVarsU),
    collect_vars(Until, UntilVarsU),
    collect_vars(A, AVars),
    collect_vars((TR, RHead), HTRVars),
    ip_lookup('$underscore_vars', Unders),
    diff_list(WhilstVarsU, Unders, WhilstVars),
    diff_list(UntilVarsU, Unders, UntilVars),
    (
      diff_list(WhilstVars, HTRVars, Diff), Diff \= []
    ->
      errornl(xxxxxxxxxxx),
      '$get_map'((TR .. _), (TR .. {}(Body))),
      '$index_member'(Rule, Index, Body),
      write_term_list(stderr, [nl, 'Error: Rule number ', Index, ' of ',
			       (TR .. {}(Body)), nl, 
			       'contains non-global vars not in rule guard ',nl]),
      thread_exit
      ;
      true
    ),
    (
      diff_list(UntilVars, HTRVars, Diff),Diff\=[]
    ->
      '$get_map'((TR .. _), (TR .. {}(Body))),
      '$index_member'(Rule, Index, Body),
      write_term_list(stderr, [nl, 'Error: Rule number ', Index, ' of ',
			       (TR .. {}(Body)), nl, 
			       'contains non-global vars not in rule guard ',nl]),
      thread_exit
     ;
     true
    ),
    (
      diff_list(AVars, HTRVars, Diff),Diff\=[]
    ->
      '$get_map'((TR .. _), (TR .. {}(Body))),
      '$index_member'(Rule, Index, Body),
      write_term_list(stderr, [nl, 'Error: Rule number ', Index, ' of ',
			       (TR .. {}(Body)), nl, 
			       'contains non-global vars not in rule guard ',nl]),
      thread_exit
      ;
      true
    ),
    collect_vars((Whilst, Until, A), BVars),
    diff_list(BVars, TRVars, RIDVars).
    
generate_indexed_TR(TR, Index, TR_I) :-
    functor(TR, F, N),
    concat_atom([F, N, Index], '_', TR_I).

%% expand a single TR program
%% the first call on this has Depends = [] which means the
%% first guarded action has no previous depedencies
%% subsequently Depends contains all the dependencies of
%% all the previous guards in the TR program
'$expand_one_TR'(_, [], _, _, [], [], []) :- !.
'$expand_one_TR'(Head, [(whilstuntil(RHead, Whilst, Until) ~> RBody)|BRest],
                 Index, Depends, AllClausesTR,
                 AllClausesRetry, AllClausesOverlap) :-
    %% generate the code for the action
    RBody = CRBody, %%'$tr_compute_rule_body0'(RBody, CRBody),
    !,
    '$build_rule'(Head, Index, RHead, Whilst, Until, CRBody,
                  Depends, OutDepends,
                  RuleList, RetryList, OverlapList),
    !,
    Index1 is Index+1,
    append(RuleList, ClausesTR, AllClausesTR),
    append(RetryList, ClausesRetry, AllClausesRetry),
    append(OverlapList, ClausesOverlap, AllClausesOverlap),
    '$expand_one_TR'(Head, BRest, Index1, OutDepends, ClausesTR,
                     ClausesRetry, ClausesOverlap).

'$repeat'(0, _RPA, []) :- !.
'$repeat'(N, RPA, [RPA|Rest]) :- N1 is N-1, '$repeat'(N1, RPA, Rest).

'$build_rule'(TR, _Index, _RHead, _Whilst, _Until, CRBody, _InDepends,
              _OutDepends,
              RuleList, _RetryList, _OverlapList) :-
    CRBody = rb(A, _G),
    A = pa([error]), !,
    RuleHead = tr_rule(TR, _, _, _, InTerms, _, _, _,
                       _, _, OldActions, _, _),
    RuleBody = (throw(error(no_matching_rule, TR, InTerms, OldActions))),
    RuleList = [(RuleHead :- RuleBody)].
    
'$build_rule'(TR, Index, RHead, Whilst, Until, CRBody, InDepends, ROutDepends,
              RuleList, RetryList, OverlapList) :-
    CRBody = rb(A, G),
    '$task_atomic'(_,_),
    A = pa([]), !,
    A1 = [],
    generate_indexed_TR(TR, Index, TR_I),
    '$process_vars_of_rule'(TR, Index, RHead, Whilst, Until, A, RIDVars),
    RuleHead = tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms,
		       RInDepends, OutDepends,
                       InTimes, OutDelay, OldActions, Actions, 
		       ResourcesUsed),
    (
      G == true
    ->
      Goal_and_Action = do_action(OldActions, Actions, OutTerms)
    ;
      Goal_and_Action = ('$delay_tr_call'(G), do_action(OldActions, Actions, OutTerms))
    ),
    RuleBody = (
		 StartIndex =< Index, 
                 RHead, !,
                 A1V = A1, RID = rid(Index, RIDVars, EW0, EU0),
                 AT \=RID,
                 EW0 = EW,
                 EU0 = EU,
                 check_if_should_clean_up(ResourcesUsed),
                 Term = rule_choice(TR, TR_I, RID,
				    InDepends, InTimes, ResourcesUsed), 
                 Actions = A1V,
                 OutTerms = [Term|InTerms],
                 merge(SRuleDepends, RInDepends, OutDepends),
		 Goal_and_Action, 
		 %gettimeofday(Now),
		 get_delay(Now, OutTimes, OutDelay)

                 ),
    Rule = (RuleHead :- RuleBody),
    '$tr_find_clause_depend'(RHead, InDepends, RHDepend),
    sort(RHDepend, ROutDepends),
    '$build_overlap_clauses'(TR, RHead, InDepends, Index, OverlapList),
    '$build_depends'(Whilst, Until, RHead, InDepends, RIDVars, 
                     SRuleDepends, RetryDepends),
    '$build_whilst_until_rules'(Whilst, Until, TR_I, TR, RHead, RIDVars, Index,
                                RetryDepends, Now, InTimes, OutTimes,
                                Rule, RetryList1, RuleList, EW, EU),
    '$filter_retry_list'(RetryList1, RetryList).
'$build_rule'(TR, Index, RHead, Whilst, Until, CRBody, InDepends, ROutDepends,
              RuleList, RetryList, OverlapList) :-
    CRBody = rb(A, G),
    A = pa(A1), !,
    generate_indexed_TR(TR, Index, TR_I),
    '$process_vars_of_rule'(TR, Index, RHead, Whilst, Until, A, RIDVars),
    RuleHead = tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms,
		       RInDepends, OutDepends,
                       InTimes, OutDelay, OldActions, Actions, 
		       _ResourcesUsed),
    (
      G == true
    ->
      Goal_and_Action = do_action(OldActions, Actions, OutTerms)
    ;
      Goal_and_Action = ('$delay_tr_call'(G), do_action(OldActions, Actions, OutTerms))
    ),
    RuleBody = (
		 StartIndex =< Index, 
                 RHead, !,
                 A1V = A1, RID = rid(Index, RIDVars, EW0, EU0),
                 AT \= RID,
                 EW0 = EW,
                 EU0 = EU,
                 Term = rule_choice(TR, TR_I, RID,
				    RInDepends, InTimes, no), 
                 Actions = A1V,
                 OutTerms = [Term|InTerms],
                 merge(SRuleDepends, RInDepends, OutDepends),
		 Goal_and_Action, 
		 %gettimeofday(Now),
		 get_delay(Now, OutTimes, OutDelay)

                 ),
    Rule = (RuleHead :- RuleBody),
    '$tr_find_clause_depend'(RHead, InDepends, RHDepend),
    sort(RHDepend, ROutDepends),
    '$build_overlap_clauses'(TR, RHead, InDepends, Index, OverlapList),
    '$build_depends'(Whilst, Until, RHead, InDepends, RIDVars, 
                     SRuleDepends, RetryDepends),
    '$build_whilst_until_rules'(Whilst, Until, TR_I, TR, RHead, RIDVars, Index,
                                RetryDepends, Now, InTimes, OutTimes,
                                Rule, RetryList1, RuleList, EW, EU),
    '$filter_retry_list'(RetryList1, RetryList).

'$build_rule'(TR, Index, RHead, Whilst, Until, CRBody, InDepends, ROutDepends,
              RuleList, RetryList, OverlapList) :-
    CRBody = rb(A, G),
    A = wait_repeat(ActList, T1, N), !,
    %goals2list(Act, ActList),
    N1 is N-1,
    '$repeat'(N1, RPA, RPAList),
    append([(pa(ActList):T1)|RPAList], [(pa([error]):(-1))], AV0),
    generate_indexed_TR(TR, Index, TR_I),
    '$process_vars_of_rule'(TR, Index, RHead, Whilst, Until, A, RIDVars),
    (
      functor(TR, TRF, _), '$task_atomic'(TRF,_)
    ->
      OutResourcesUsed = yes
    ;
      OutResourcesUsed = ResourcesUsed
    ),
    (
      G == true
    ->
      Goal = true
    ;
      Goal = '$delay_tr_call'(G)
    ),
    RuleHead = tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms,
		       RInDepends, OutDepends,
                       InTimes, OutDelay, OldActions, Actions, 
		       ResourcesUsed),


    TimeGoal = (NowTS = Now+T1, OutTimes = [NowTS|MidTimes]),
    RuleBody = (
		 StartIndex =< Index, 
                 RHead, !, AV = AV0, RID = rid(Index, RIDVars, EW0, EU0),
                 AT \= RID,
		 RPA = (rpa(ActList):T1),
                 EW0 = EW,
                 EU0 = EU,
                 Term = rule_choice(TR, TR_I, RID,
				    RInDepends, InTimes, ResourcesUsed),
                 TimeGoal,
                 TSTerm = timed_seq(NowTS, AV, AV, OutDepends,
                                    MidTimes, OutResourcesUsed),
                 Actions = ActList,
                 
                 OutTerms = [TSTerm, Term|InTerms],
                 merge(SRuleDepends, RInDepends, OutDepends),
		 Goal, 
		 do_action(OldActions, rpa(Actions), OutTerms),
		 get_delay(Now, OutTimes, OutDelay)
               ),
    
    Rule = (RuleHead :- RuleBody),
    '$tr_find_clause_depend'(RHead, InDepends, RHDepend),
    sort(RHDepend, ROutDepends),
    '$build_overlap_clauses'(TR, RHead, InDepends, Index, OverlapList),
    '$build_depends'(Whilst, Until, RHead, InDepends, RIDVars, 
                     SRuleDepends, RetryDepends),
    '$build_whilst_until_rules'(Whilst, Until, TR_I, TR, RHead, RIDVars, Index,
                                 RetryDepends, Now, InTimes, MidTimes,
                                Rule, RetryList1, RuleList, EW, EU),
    '$filter_retry_list'(RetryList1, RetryList).
'$build_rule'(TR, Index, RHead, Whilst, Until, CRBody, InDepends, ROutDepends,
              RuleList, RetryList, OverlapList) :-
    CRBody = rb(A, G),
    A = timed_seq([pa(A1):T1|_]), !,
    generate_indexed_TR(TR, Index, TR_I),
    '$process_vars_of_rule'(TR, Index, RHead, Whilst, Until, A, RIDVars),
    (
      functor(TR, TRF, _), '$task_atomic'(TRF,_)
    ->
      OutResourcesUsed = yes
    ;
      OutResourcesUsed = ResourcesUsed
    ),
    (
      G == true
    ->
      Goal_and_Action = do_action(OldActions, Actions, OutTerms)
    ;
      Goal_and_Action = ('$delay_tr_call'(G), do_action(OldActions, Actions, OutTerms))
    ),

    RuleHead = tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms,
		       RInDepends, OutDepends,
                       InTimes, OutDelay, OldActions, Actions, 
		       ResourcesUsed),


    TimeGoal = (NowTS = Now+T1, OutTimes = [NowTS|MidTimes]),
    RuleBody = (
		 StartIndex =< Index, 
                 RHead, !, A = timed_seq(AV),
                 RID = rid(Index, RIDVars, EW0, EU0),
                 AT \= RID,
                 EW0 = EW,
                 EU0 = EU,
                 Term = rule_choice(TR, TR_I, RID,
				    RInDepends, InTimes, ResourcesUsed),
                 TimeGoal,
                 TSTerm = timed_seq(NowTS, AV, AV, OutDepends,
                                    MidTimes, OutResourcesUsed),
                 Actions = A1,
                 
                 OutTerms = [TSTerm, Term|InTerms],
                 merge(SRuleDepends, RInDepends, OutDepends),
		 Goal_and_Action,
		 get_delay(Now, OutTimes, OutDelay)
               ),
    
    Rule = (RuleHead :- RuleBody),
    '$tr_find_clause_depend'(RHead, InDepends, RHDepend),
    sort(RHDepend, ROutDepends),
    '$build_overlap_clauses'(TR, RHead, InDepends, Index, OverlapList),
    '$build_depends'(Whilst, Until, RHead, InDepends, RIDVars, 
                     SRuleDepends, RetryDepends),
    '$build_whilst_until_rules'(Whilst, Until, TR_I, TR, RHead, RIDVars, Index,
                                 RetryDepends, Now, InTimes, MidTimes,
                                Rule, RetryList1, RuleList, EW, EU),
    '$filter_retry_list'(RetryList1, RetryList).

'$build_rule'(TR, Index, RHead, Whilst, Until, CRBody, InDepends, ROutDepends,
              RuleList, RetryList, OverlapList) :-
    CRBody = rb(A, G),
    A = timed_seq([TR1:T1|_]), !,
    generate_indexed_TR(TR, Index, TR_I),
    '$process_vars_of_rule'(TR, Index, RHead, Whilst, Until, A, RIDVars),
    (
      functor(TR, TRF, _), '$task_atomic'(TRF,_)
    ->
      OutResourcesUsed = yes
    ;
      OutResourcesUsed = ResourcesUsed
    ),
    (
      G == true
    ->
      Goal = true
    ;
      Goal = '$delay_tr_call'(G)
    ),
    RuleHead = tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms,
		       RInDepends, Stamps,
                       InTimes, OutDelay, OldActions, Actions, 
		       ResourcesUsed),


    TimeGoal = (NowTS = Now+T1,  OutTimes = [NowTS|MidTimes]),
    RuleBody = (
		 StartIndex =< Index, 
		RHead, !, A = timed_seq(AV),
                 RID = rid(Index, RIDVars, EW0, EU0),
		AT \= RID,
                 EW0 = EW,
                 EU0 = EU,
                 TimeGoal,
                 Term = rule_choice(TR, TR_I, RID,
				    RInDepends, InTimes, ResourcesUsed), 
		TSTerm = timed_seq(NowTS, AV, AV, OutDepends,
				   MidTimes, ResourcesUsed),
		MidTerms = [TSTerm,Term|InTerms],
                 merge(SRuleDepends, RInDepends, OutDepends),
                 Goal,
		tr_rule(TR1, none, 1, Now, MidTerms, OutTerms,
			OutDepends, Stamps,
			OutTimes, OutDelay, OldActions, Actions, 
			OutResourcesUsed)
	      ),
    Rule = (RuleHead :- RuleBody),
    '$tr_find_clause_depend'(RHead, InDepends, RHDepend),
    sort(RHDepend, ROutDepends),
    '$build_overlap_clauses'(TR, RHead, InDepends, Index, OverlapList),
    '$build_depends'(Whilst, Until, RHead, InDepends, RIDVars, 
                     SRuleDepends, RetryDepends),
    '$build_whilst_until_rules'(Whilst, Until, TR_I, TR, RHead, RIDVars, Index,
                                RetryDepends, Now, InTimes, MidTimes,
                                Rule, RetryList1, RuleList, EW, EU),
    '$filter_retry_list'(RetryList1, RetryList).
    


'$build_rule'(TR, Index, RHead, Whilst, Until, CRBody, InDepends, ROutDepends,
              RuleList, RetryList, OverlapList) :-
    CRBody = rb(A, G),
    %A = TR1,
    generate_indexed_TR(TR, Index, TR_I),
    '$process_vars_of_rule'(TR, Index, RHead, Whilst, Until, A, RIDVars),
    (
      functor(TR, TRF, _), '$task_atomic'(TRF,_)
    ->
      OutResourcesUsed = yes
    ;
      OutResourcesUsed = ResourcesUsed
    ),
    (
      G == true
    ->
      Goal = true
    ;
      Goal = '$delay_tr_call'(G)
    ),
    RuleHead = tr_rule(TR, AT, StartIndex, Now, InTerms, OutTerms,
		       RInDepends, Stamps,
                       InTimes, OutDelay, OldAction, Actions, 
		       ResourcesUsed),
    RuleBody = ( 
		 StartIndex =< Index,
                 RHead,!, AV = A, RID = rid(Index, RIDVars, EW0, EU0),
                 AT \= RID,
                 EW0 = EW,
                 EU0 = EU,
                 Term = rule_choice(TR, TR_I, RID,
				    RInDepends, InTimes, ResourcesUsed), 
                 MidTerms = [Term|InTerms],
                 merge(SRuleDepends, RInDepends, OutDepends),
                 Goal,
                 tr_rule(AV, none, 1, Now, MidTerms, OutTerms,
                         OutDepends, Stamps,
                         OutTimes, OutDelay, OldAction, Actions, 
                         OutResourcesUsed)
               ),
    Rule = (RuleHead :- RuleBody),
    '$tr_find_clause_depend'(RHead, InDepends, RHDepend),
    sort(RHDepend, ROutDepends),
    '$build_overlap_clauses'(TR, RHead, InDepends, Index, OverlapList),
    '$build_depends'(Whilst, Until, RHead, InDepends, RIDVars, 
                     SRuleDepends, RetryDepends),
    '$build_whilst_until_rules'(Whilst, Until, TR_I, TR, RHead, RIDVars, Index,
                                RetryDepends, Now, InTimes, OutTimes,
                                Rule, RetryList1, RuleList, EW, EU),
    '$filter_retry_list'(RetryList1, RetryList).

'$build_overlap_clauses'(TR, RHead, InDepends, Index, OverlapClauses) :-
    '$tr_find_clause_depend'(RHead, [], LODepend),
    diff_list(LODepend, InDepends, OverlapDepend),
    sort(OverlapDepend, SOverlapDepend),
    (
      SOverlapDepend = []
    ->
      OverlapClauses = []
    ;
      OverlapClauses = [(overlap_at(TR, Changed, CurrentIndex, I) :-
                               CurrentIndex > Index,
                               overlap(Changed, SOverlapDepend), !,
                               I = Index)]
    ).

%%% No until part
'$build_depends'('$time'(_), true, RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'(\+(RHead), InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead)), InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'('$time'(Whilst, _), true, RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'(Whilst, true, RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.

%%% Until is (partly) timed
'$build_depends'('$time'(_), '$time'(_), RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'(\+(RHead), InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead)), InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'('$time'(_), '$time'(Until,_), RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), Until), InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), Until), InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'('$time'(Whilst, _), '$time'(_), RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'('$time'(Whilst, _), '$time'(Until,_), RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'(Whilst, '$time'(_), RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst)),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst)),
                               [], RD)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst)),
                               [], RD)
    ),
    sort(RD, RetryDepends).
'$build_depends'(Whilst, '$time'(Until, _), RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.

%% Until part has no time - Until is part of depends
'$build_depends'('$time'(_), Until, RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), Until), InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), Until),
			       InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'('$time'(Whilst, _), Until, RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'(Whilst, true, RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    RetryDepends = RuleDepends.
'$build_depends'(Whilst, Until, RHead, InDepends, RIDVars,
                 RuleDepends, RetryDepends) :-
    !,
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst), Until),
                               InDepends, AllRHeadDepend)
    ),
    sort(AllRHeadDepend, RuleDepends),
    (
      RIDVars = []
    ->      
      '$tr_find_clause_depend'((\+(RHead), \+(Whilst), Until),
                               InDepends, RD)
      ;
      '$tr_find_clause_depend'((RHead, \+(RHead), \+(Whilst), Until),
                               InDepends, RD)
    ),
    sort(RD, RetryDepends).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% both have the same time
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
'$build_whilst_until_rules'('$time'(Whilst, WT), '$time'(Until, UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    WT == UT, !,
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule, WRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EW = Now + WT,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ET, _), Now1, Changed, _) :-
                Now1 =< ET, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Whilst, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, Changed, 1))
        ].
'$build_whilst_until_rules'('$time'(WT), '$time'(Until, UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    WT == UT, !,
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    EW = Now + WT,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ET, _), Now1, Changed, _) :-
                Now1 =< ET, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, Changed, 1))
        ].

'$build_whilst_until_rules'('$time'(Whilst, WT), '$time'(UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    WT == UT, !,
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule, WRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EW = Now + WT,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ET, _), Now1, Changed, _) :-
                Now1 =< ET, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Whilst, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, Changed, 1))
        ].

'$build_whilst_until_rules'('$time'(WT), '$time'(UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    WT == UT, !,
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    EW = Now + WT,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ET, _), Now1, Changed, _) :-
                Now1 =< ET, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, Changed, 1))
        ].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% both have times but not know to be the same
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'$build_whilst_until_rules'('$time'(Whilst, WT), '$time'(Until, UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, EU) :-
    !,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    OutTimes = [(Now + WT), (Now + UT)|InTimes],
    RuleList = [Rule, TWRule, WRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EW = Now + WT,
    EU = Now + UT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ETW, ETU), Now1, _, _) :-
                Now1 =< min(ETW, ETU), !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, I) :-
                ETU < ETW, Now1 =< ETW, Until, !, I = 1),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, _) :-
                ETU < ETW, Now1 =< ETW, !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, _) :-
                ETU > ETW, Now1 =< ETU, Whilst, !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, _) :-
                ETU > ETW, Now1 =< ETU, RHead, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                Whilst, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].
'$build_whilst_until_rules'('$time'(WT), '$time'(Until, UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, EU) :-
    !,
    OutTimes = [(Now + WT), (Now + UT)|InTimes],
    RuleList = [Rule, TWRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    EW = Now + WT,
    EU = Now + UT,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ETW, ETU), Now1, _, _) :-
                Now1 =< min(ETW, ETU), !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, I) :-
                ETU < ETW, Now1 =< ETW, Until, !, I = 1),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, _) :-
                ETU < ETW, Now1 =< ETW, !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, _) :-
                ETU > ETW, Now1 =< ETU, RHead, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, RID, Now1, _, I) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].
'$build_whilst_until_rules'('$time'(Whilst, WT), '$time'(UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, EU) :-
    !,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    OutTimes = [(Now + WT), (Now + UT)|InTimes],
    RuleList = [Rule, TWRule, WRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EW = Now + WT,
    EU = Now + UT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ETW, ETU), Now1, _, _) :-
                Now1 =< min(ETW, ETU), !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, I) :-
                ETU < Now1, !, I = 1),
            (retry_from(TR_I, TR, RID , Now1, _, _) :-
                Whilst, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].

'$build_whilst_until_rules'('$time'(WT), '$time'(UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, EU) :-
    !,
    OutTimes = [(Now + WT), (Now + UT)|InTimes],
    RuleList = [Rule, TWRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    EW = Now + WT,
    EU = Now + UT,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, ETW, ETU), Now1, _, _) :-
                Now1 =< min(ETW, ETU), !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, ETW, ETU), Now1, _, I) :-
                ETU < Now1, !, I = 1),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% until time, not whilst time
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
'$build_whilst_until_rules'(true, '$time'(Until, UT), TR_I, TR,
                            RHead, RIDVars, _Index,
                            RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, EU) :-
    !,
    OutTimes = [(Now + UT)|InTimes],
    RuleList = [Rule],
    EU = Now + UT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, RIDVars, _, ETU), Now1, Changed, _) :-
                Now1 =< ETU, RHead, !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, _, ETU), Now1, Changed, I) :-
                Now1 > ETU, Until, !, I = 1),
            (retry_from(TR_I, TR, rid(_, _, _, ETU), Now1, Changed, _) :-
                Now1 > ETU, no_overlap(Changed, RetryDepends), !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, _, ETU), Now1, Changed, _) :-
                Now1 > ETU, RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].
'$build_whilst_until_rules'(true, '$time'(UT), TR_I, TR,
                            RHead, RIDVars, _Index,
                            RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, EU) :-
    !,
    OutTimes = [(Now + UT)|InTimes],
    RuleList = [Rule],
    EU = Now + UT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, _, ETU), Now1, Changed, _) :-
                Now1 =< ETU, no_overlap(Changed, RetryDepends), !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, _, ETU), Now1, Changed, _) :-
                Now1 =< ETU, RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].
'$build_whilst_until_rules'(Whilst, '$time'(Until, UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, EU) :-
    !,
    (
      RIDVars = []
    ->
      RID = rid(_, [], _, ETU)
    ;
      RID = rid(_, RIDVars, _, ETU)
    ),
    OutTimes = [(Now + UT)|InTimes],
    RuleList = [Rule, WRule],
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EU = Now + UT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Now1 =< ETU, RHead, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Now1 =< ETU, Whilst, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, I) :-
                Now1 > ETU, Until, !, I = 1),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Now1 > ETU, no_overlap(Changed, RetryDepends), !, fail),
            (retry_from(TR_I, TR, rid(_, RIDVars, _, ETU), Now1, Changed, _) :-
                Now1 > ETU, RHead, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Now1 > ETU, Whilst, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].

'$build_whilst_until_rules'(Whilst, '$time'(UT), TR_I, TR,
                            RHead, RIDVars, Index,
                            RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, EU) :-
    !,
    (
      RIDVars = []
    ->
      RID = rid(_, [], _, ETU)
    ;
      RID = rid(_, RIDVars, _, ETU)
    ),
    OutTimes = [(Now + UT)|InTimes],
    RuleList = [Rule, WRule],
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EU = Now + UT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, rid(_, _, _, ETU), Now1, Changed, _) :-
                Now1 =< ETU, no_overlap(Changed, RetryDepends), !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Now1 =< ETU, RHead, !, fail),
            (retry_from(TR_I, TR, RID, Now1, Changed, _) :-
                Now1 =< ETU, Whilst, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% whilst time, not until time
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

'$build_whilst_until_rules'('$time'(Whilst, WT), true, TR_I, TR,
                            _RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    !,
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule, WRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EW = Now + WT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, _RID, Now, Changed, I) :-
                overlap_at(TR, Changed, Index, I), !),
            (retry_from(TR_I, TR, rid(_, _, ETW, _), Now1, _, _) :-
                Now1 =< ETW, !, fail),
            (retry_from(TR_I, TR, _, Now1, _, Index))
         ].
'$build_whilst_until_rules'('$time'(WT), true, TR_I, TR,
                            _RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    !,
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule],
    EW = Now + WT,
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    RetryClauses =
        [
            (retry_from(TR_I, TR, _, Now, Changed, I) :-
                overlap_at(TR, Changed, Index, I), !),
            (retry_from(TR_I, TR, rid(_, _, ETW, _), Now1, _, _) :-
                Now1 =< ETW, !, fail),
            (retry_from(TR_I, TR, _, Now1, _, Index))
         ].
'$build_whilst_until_rules'('$time'(Whilst, WT), Until, TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    !,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule, WRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    EW = Now + WT,
    RetryClauses =
        [
            (retry_from(TR_I, TR, RID, Now1, _, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, rid(_, _, ETW, _), Now1, _, _) :-
                Now1 =< ETW, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                Whilst, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _, Now1, _, 1))
         ].
'$build_whilst_until_rules'('$time'(WT), Until, TR_I, TR,
                            RHead, RIDVars, Index,
                            _RetryDepends, Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, EW, _EU) :-
    !,
    OutTimes = [(Now + WT)|InTimes],
    RuleList = [Rule, TWRule],
    TWRule = (tr_rule(TR, AT, StartIndex, Now1, _, _, _, _, _, _, _, _, _) :-
                 StartIndex =< Index, 
                AT = rid(Index, RIDVars, ETW, _),
                Now1 =< ETW, !, fail),
    EW = Now + WT,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, RID, Now1, _, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, rid(_, _, ETW, _), Now1, _, _) :-
                Now1 =< ETW, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, I) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _, Now1, _, 1))
         ].




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% no times
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
'$build_whilst_until_rules'(true, true, TR_I, TR, _RHead, _RIDVars, Index,
                            RetryDepends, _Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, _EU) :-
    OutTimes = InTimes,
    RuleList = [Rule],
    RetryClauses =
        [
            (retry_from(TR_I, TR, _, _, Changed, I) :-
                overlap_at(TR, Changed, Index, I), !),
            (retry_from(TR_I, TR, _, _, Changed, I) :-
                overlap(Changed, RetryDepends), I = Index)
        ].

'$build_whilst_until_rules'(Whilst, true, TR_I, TR,
                            _RHead, RIDVars, Index,
                            RetryDepends, _Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, _EU) :-
    !,
    OutTimes = InTimes,
    RuleList = [Rule, WRule],
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    RetryClauses =
        [
            (retry_from(TR_I, TR, _, _, Changed, I) :-
                overlap_at(TR, Changed, Index, I), !),
            (retry_from(TR_I, TR, _, _, Changed, I) :-
                overlap(Changed, RetryDepends), I = Index)
         ].

'$build_whilst_until_rules'(true, Until, TR_I, TR,
                            RHead, RIDVars, _Index,
                            RetryDepends, _Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, _EU) :-
    !,
    OutTimes = InTimes,
    RuleList = [Rule],
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    RetryClauses =
        [
            (retry_from(TR_I, TR, _, _, Changed, _) :-
                no_overlap(Changed, RetryDepends), !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _, _, _, 1))
         ].

'$build_whilst_until_rules'(Whilst, Until, TR_I, TR,
                            RHead, RIDVars, Index,
                            RetryDepends, _Now, InTimes, OutTimes,
                            Rule, RetryClauses, RuleList, _EW, _EU) :-
    !,
    (
      RIDVars = []
    ->
      true
    ;
      RID = rid(_, RIDVars, _, _)
    ),
    OutTimes = InTimes,
    RuleList = [Rule, WRule],
    WRule = (tr_rule(TR, AT, _, _, _, _, _, _, _, _, _, _, _) :-
                AT = rid(Index, RIDVars, _, _),
                Whilst, !, fail),
    RetryClauses =
        [
            (retry_from(TR_I, TR, _, Now1, Changed, _) :-
                no_overlap(Changed, RetryDepends), !, fail),
            (retry_from(TR_I, TR, RID , Now1, _, I) :-
                Until, !, I = 1),
            (retry_from(TR_I, TR, RID , Now1, _, _) :-
                Whilst, !, fail),
            (retry_from(TR_I, TR, RID, Now1, _, _) :-
                RHead, !, fail),
            (retry_from(TR_I, TR, _RID, Now1, _, 1))
         ].







'$filter_retry_list'([], []).
'$filter_retry_list'([Clause|Rest], RetryList) :-
    Clause = (_ :- overlap_at(_, _, 1, _), _), !,
    '$filter_retry_list'(Rest, RetryList).
'$filter_retry_list'([Clause|Rest], RetryList) :-
    Clause = (_ :- overlap(_, []), _), !,
    '$filter_retry_list'(Rest, RetryList).
'$filter_retry_list'([Clause|Rest], [Clause|RetryList]) :-
    '$filter_retry_list'(Rest, RetryList).


'$tr_compute_rule_body0'(A, A) :- errornl('FIX'), !.

'$tr_compute_rule_body0'((RBody++G), rb(CRBody, G)) :-
    !,
    '$tr_compute_rule_body'(RBody, CRBody, true).
/*    (
      CRBody = pa(_)
    ->
      true
    ;
      write_term_list(['Error ', (RBody~~G),
                       'has a goal but is not a primitive action.', nl]),
      thread_exit
    ).
*/

'$tr_compute_rule_body0'(RBody, rb(CRBody, _)) :-
    !,
    '$tr_compute_rule_body'(RBody, CRBody, true).


%% A conjuction of primitive actions
%% is converted to a list of actions and wrapped in a pa(...)
'$tr_compute_rule_body'({}, pa([]), _) :- !.
'$tr_compute_rule_body'({}(RBody), pa(CBody), _) :-
    !,
    goals2list(RBody, CBody).

'$tr_compute_rule_body'(wait_repeat(A, T, N), wait_repeat(A, T, N), _) :- !.
'$tr_compute_rule_body'(RBody, RBody, _NoSeq) :-
    functor(RBody, P, N),
    functor(Copy, P, N),
    '$tr_prog'(Copy, _), !.
'$tr_compute_rule_body'(RBody, CRBody, NoSeq) :-
    '$tr_compute_cyclic_rule_body'(RBody, CRBody, NoSeq),!.
% a TR as an atom
'$tr_compute_rule_body'(RBody, RBody, _).

'$tr_compute_cyclic_rule_body'(_, _, fail) :-
    write_term_list(stderr, ['Error - timed sequence inside a timed sequence is not allowed', nl]),
    thread_exit.

'$tr_compute_cyclic_rule_body'([], [], _NoSeq).
 
'$tr_compute_cyclic_rule_body'([A:T|Rest], [CA:T|CRest], NoSeq) :-
    !,
    '$tr_compute_rule_body'(A, CA, fail),
    '$tr_compute_cyclic_rule_body'(Rest, CRest, NoSeq).
'$tr_compute_cyclic_rule_body'([A], R, _NoSeq) :-
    '$tr_compute_rule_body'(A, CA, fail),
    R = [(CA:(-1))].    

xxx :-
    listing(tr_rule/13),
    listing(retry_from/6),
    listing(overlap_at/4).